from .variable import Variable

# ====================================================================
#
# CellMeasure object
#
# ====================================================================

class CellMeasure(Variable):
    '''
    
A CF cell measure construct containing information that is needed
about the size, shape or location of the field's cells.

It is a variable which contains a data array and metadata comprising
properties to describe the physical nature of the data.


**Attributes**

==========  =======  =================================================
Attribute   Type     Description
==========  =======  =================================================
`!measure`  ``str``  The spatial measure being represented. Either
                     ``'area'`` or ``'volume'`` are allowed by the CF
                     conventions.
==========  =======  =================================================

'''
    def dump(self, domain=None, key=None, level=0):
        '''

Return a string containing a full description of the cell measure.

:Parameters:

    id: str, optional
       Set the common prefix of variable component names. By default
       the instance's class name is used.

:Returns:

    out : str
        A string containing the description.

**Examples**

>>> x = c.dump()
>>> print c.dump()
>>> print c.dump(id='area_measure')

'''
        indent1 = '    ' * level
        indent2 = '    ' * (level+1)

        if hasattr(self, 'measure'):
            string = ['%sCell measure: %s' % (indent1, self.measure)]
        elif hasattr(self.Units, 'units'):
            string = ['%sCell measure: %s' % (indent1, self.units)]
        else:
            string = ['%sCell measure: %s' % (indent1, self.name(default=''))]

        if self._hasData:
            if domain:
                x = ['%s(%d)' % (domain.axis_name(dim),
                                 domain.dimension_sizes[dim])
                     for dim in domain.dimensions[key]]
                string.append('%sData(%s) = %s' % (indent2, ', '.join(x), str(self.Data)))
            else:
                x = [str(s) for s in self.shape]
                string.append('%sData(%s) = %s' % (indent2, ', '.join(x), str(self.Data)))
        #--- End: if

        if self._simple_properties():
            string.append(self.dump_simple_properties(level=level+1))
          
        return '\n'.join(string)
    #--- End: def

    def identity(self, default=None):
        '''

Return the cell measure's identity.

The idendity is the value of the `!measure` attribute, or if that
doesn't exist the `standard_name` property or, if that does not exist,
the `!id` attribute.

:Parameters:

    default : optional
        If none of `measure`, `standard_name` and `!id` exist then
        return *default*. By default, *default* is None.

:Returns:

    out :
        The identity.

**Examples**

'''
        return getattr(self, 'measure',
                       self.getprop('standard_name',
                                    getattr(self, 'id', default)))
    #--- End: def

#    def name(self, long_name=None, ncvar=None, default=None):
#        '''
#Return a name for the variable.
#
#Returns the the first found of:
#
#* The `!measure` attribute.
#
#* The identity, as returned by the `identity` method.
#
#* The `!long_name` CF property, preceeded by the string
#  ``'long_name:``.
#
#* The `!ncvar` attribute, preceeded by the string ``'ncvar:``.
#
#If none of these are found then the value of the *default* parameter
#is returned.
#
#:Parameters:
#
#    default : str, optional
#        If no name can be found then return the value of the *default*
#        parameter. By default the default is None.
#
#:Returns:
#
#    out : str
#        A  name for the cell measure.
#
#**Examples**
#
#'''  
#        if hasattr(self, 'measure'):
#            return self.measure
#        
#        if hasattr(self, 'standard_name'):
#            return self.standard_name
#        
#        if long_name is not None and hasattr(self, 'long_name'):
#            return self.long_name
#        
#        if ncvar is not None and hasattr(self, 'ncvar'):
#            return self.ncvar
#        
#        return default
#    #--- End: def

#--- End: class
