from numpy import amax          as numpy_amax
from numpy import amin          as numpy_amin
from numpy import array         as numpy_array
from numpy import cos           as numpy_cos
from numpy import dtype         as numpy_dtype
from numpy import empty         as numpy_empty
from numpy import ndarray       as numpy_ndarray
from numpy import ndim          as numpy_ndim
from numpy import ndindex       as numpy_ndindex
from numpy import ones          as numpy_ones
from numpy import result_type   as numpy_result_type
from numpy import shape         as numpy_shape
from numpy import sin           as numpy_sin
from numpy import size          as numpy_size
from numpy import tan           as numpy_tan
from numpy import tile          as numpy_tile
from numpy import unique        as numpy_unique
from numpy import unravel_index as numpy_unravel_index
from numpy import where         as numpy_where
from numpy import vectorize     as numpy_vectorize
from numpy import zeros         as numpy_zeros

from numpy.ma import array        as numpy_ma_array
from numpy.ma import empty        as numpy_ma_empty
from numpy.ma import getmask      as numpy_ma_getmask
from numpy.ma import getmaskarray as numpy_ma_getmaskarray
from numpy.ma import is_masked    as numpy_ma_is_masked
from numpy.ma import isMA         as numpy_ma_isMA
from numpy.ma import masked       as numpy_ma_masked
from numpy.ma import masked_where as numpy_ma_masked_where
from numpy.ma import MaskedArray  as numpy_ma_MaskedArray
from numpy.ma import nomask       as numpy_ma_nomask
from numpy.ma import ones         as numpy_ma_ones
from numpy.ma import where        as numpy_ma_where
from numpy.ma import zeros        as numpy_ma_zeros

import operator

from operator  import mul as operator_mul
from math      import ceil as math_ceil
from sys       import byteorder, getrefcount
from itertools import izip
from itertools import product as itertools_product

from ..netcdf.filearray import NetCDFFileArray
from ..pp.filearray     import PPFileArray

from ..cfdatetime import dt2rt, st2dt, rt2dt
from ..units      import Units
from ..constants  import masked as cf_masked
from ..functions  import (CHUNKSIZE, FM_THRESHOLD, RTOL, ATOL, FREE_MEMORY,
                          parse_indices, _allclose, abspath,
                          pathjoin, hash_array, get_subspace, set_subspace)

from .filearray       import FileArray
from .partition       import Partition
from .partitionmatrix import PartitionMatrix

_arse = {0: []}
_arse4 = {0: True}
#for i in xrange(1, 2):
#    dimensions = [str(n) for n in xrange(i)]
#    arse[i] = dimensions
#    arse4[i] = dict.fromkeys(dimensions, True)
##--- End: for

_xxx = numpy_empty((), dtype=object)

_units_None    = Units()
_units_1       = Units('1')
_units_radians = Units('radians')

# ====================================================================
#
# Data object
#
# ====================================================================

class Data(object):
    '''

An N-dimensional data array with units and masked values.

* Contains an N-dimensional, indexable and broadcastable array with
  many similarities to a numpy array.

* Contains the units of the array elements.

* Supports masked arrays, regardless of whether or not it was
  initialized with a masked array.

* Uses Large Amounts of Massive Arrays (LAMA) functionality to store
  and operate on aggregated (master) data arrays and arrays which are
  larger then the available memory.

**Indexing**

A data array is indexable in a similar way to numpy array

>>> d.shape
(12, 19, 73, 96)
>>> d[...].shape
(12, 19, 73, 96)
>>> d[slice(0, 12), 10:0:-2, :, :].shape
(12, 5, 73, 96)

There are two important extensions to the numpy indexing
functionality:

* Size 1 dimensions are never removed.

  An integer index i takes the i-th element but does not reduce the
  rank of the output array by one:

  >>> d.shape
  (12, 19, 73, 96)
  >>> d[0, ...].shape
  (1, 19, 73, 96)
  >>> d[:, 3, slice(10, 0, -2), 95].shape
  (12, 1, 5, 1)

* When advanced indexing is used on more than one dimension, the
  advanced indices work independently.

  When more than one dimension's slice is a 1-d boolean sequence or
  1-d sequence of integers, then these indices work independently
  along each dimension (similar to the way vector subscripts work in
  Fortran), rather than by their elements:

  >>> d.shape
  (12, 19, 73, 96)
  >>> d[0, :, [0,1], [0,1,2]].shape
  (1, 19, 2, 3)

**Miscellaneous**

Data objects are picklable.

Data objects are hashable, but note that, since Data objects are
mutable, their hash values may change if created at different times.

'''  
    def __init__(self, data=None, units=None, fill_value=None, hardmask=True,
                 chunk=True, loadd=None, dt=False):
        '''

**Initialization**

:Parameters:

    data : array-like, optional
        The data for the array.

    units : str or Units, optional
        The units of the data. By default the array elements are
        dimensionless.

    fill_value : optional 
        The fill value of the data. By default, or if None, the numpy
        fill value appropriate to the array's data type will be used.

    hardmask : bool, optional
        If False then the mask is soft. By default the mask is hard.

    chunk : bool, optional
        If True (the default) then the data array will be partitioned
        if it is larger than the chunk size. If False then the data
        array will be stored in a single partition.

**Examples**

>>> d = cf.Data(5)
>>> d = cf.Data([1,2,3], units='K')
>>> import numpy   
>>> d = cf.Data(numpy.arange(10).reshape(2,5), units=cf.Units('m/s'), fill_value=-999)
>>> d = cf.Data(tuple('fly'))

'''      
        self._hardmask = hardmask
        self._dtype    = None
        self._isdt     = False

        if loadd is not None:
            self.loadd(loadd)
            return
        #--- End: if

        units       = Units(units)
        self._Units = units

        self._fill_value = fill_value
 
        if data is None:
            return

        if not isinstance(data, FileArray):
            if not isinstance(data, numpy_ndarray):
                data = numpy_array(data)

            if (data.dtype.kind == 'O' and not dt and 
                hasattr(data.item((0,)*data.ndim), 'timetuple')):
                # We've been given one or more datetime-like objects
                dt = True
        #--- End: if

#        if not isinstance(data, numpy_ndarray):
#            if not isinstance(data, FileArray):
#                data = numpy_array(data)
#                if (data.dtype.kind == 'O' and not dt and 
#                    hasattr(data.item((0,)*data.ndim), 'timetuple')):
#                    dt = True
#        elif (data.dtype.kind == 'O' and not dt and 
#              hasattr(data.item((0,)*data.ndim), 'timetuple')):
#            dt = True


#        if not (isinstance(data, numpy_ndarray) or isinstance(data, FileArray)):
#            data = numpy_array(data)
#            if hasattr(data, 'timetuple'):
#            dt = True

#        # Set units if they're not already a Units object
#        units = Units(units)
            
        if dt or units.isreftime:
            self._isdt = data.dtype.kind in 'SO'

            if self._isdt and not units:
                units = Units('days since 1970-1-1', units._calendar)
                self._Units = units

            if not units:
                raise ValueError(
                    "Can't initialize a numeric reference time array with %s" %
                    repr(units))

            if not units.isreftime:
                raise ValueError(
                    "Can't initialize a datetime array with %s" %
                    repr(units))

            if data.dtype.kind == 'S':
                # Convert datetime strings to datetime objects
                data = st2dt(data, units)
        #--- End: if

#        self._Units      = units
#        self._fill_value = fill_value
 
#        if data is None:
#            return
#
#        if isinstance(data, (list, int, float, tuple, bool, long, basestring)):
#            data = numpy_array(data)
        
        shape = data.shape
        ndim  = data.ndim

#        try:
#            dimensions = _arse[ndim]
#            directions = _arse4[ndim]
#        except KeyError:            
#            dimensions = [str(i) for i in xrange(ndim)]
#            directions = dict.fromkeys(dimensions, True)
#            _arse[ndim] = dimensions
#            _arse4[ndim] = directions
#        #--- End: try

        dimensions = [str(i) for i in xrange(ndim)]

        if not ndim:
            # Scalar data
            dimensions = []
            directions = True
        else:
            # N (>0) dimensional data
            dimensions = [str(i) for i in xrange(ndim)]
            directions = dict.fromkeys(dimensions, True)
        #--- End: if

        matrix = _xxx.copy()

        matrix[()] = Partition(location    = [(0, n) for n in shape],
                               shape       = list(shape),
                               _dimensions = dimensions,
                               _directions = directions,
                               Units       = units,
                               subarray    = data,
                               part        = [])
        self.partitions = PartitionMatrix(matrix, [])

        self.dtype = data.dtype

        self._ndim  = ndim
        self._shape = shape
        self._size  = data.size

        self.dimensions = dimensions
        self.directions = directions        

        if chunk:
            self.chunk()            
    #--- End: def

    def __contains__(self, value):
        '''

Membership test operators

x.__contains__(y) <==> y in x

Returns True if the value is contained anywhere in the data array. The
value may be a `cf.Data` object.

**Examples**

>>> d = cf.Data([[0.0, 1,  2], [3, 4, 5]], units='m')
>>> 4 in d
True
>>> cf.Data(3) in d
True
>>> value = d[1, 2]
>>> value.Units *= 2
>>> value.squeeze(0)
>>> value
<CF Data: [2.5] 2 m>
>>> value in d
True
>>> numpy.array([[[2]]]) in d
True

'''
        if isinstance(value, self.__class__):
            if value.Units.equivalent(self.Units):
                if not value.Units.equals(self.Units):                    
                    value = value.copy()                    
                    value.Units = self.Units
            elif value.Units:
                return False

            value = value.unsafe_array
        #--- End: if

        conform_args = self.conform_args(revert_to_file=True,
                                         readonly=True)

        for partition in self.partitions.matrix.flat:
            array = partition.dataarray(**conform_args)
            partition.close()
            if value in array:
                return True
        #--- End: for

        return False
    #--- End: def

    def __deepcopy__(self, memo):
        '''

Used if copy.deepcopy is called

''' 
        return self.copy()
    #--- End: def

#    def __getstate__(self):
#        '''
#
#Called when pickling.
#
#:Parameters:
#
#    None
#
#:Returns:
#
#    out : dict
#        A dictionary of the instance's attributes
#
#**Examples**
#
#'''
#        return dict([(attr, getattr(self, attr))
#                     for attr in self.__slots__ if hasattr(self, attr)])
#    #--- End: def        

#    def __setstate__(self, odict):
#        '''
#
#Called when unpickling.
#
#:Parameters:
#
#    odict : dict
#        The output from the instance's `__getstate__` method.
#
#:Returns:
#
#    None
#
#'''
#        for attr, value in odict.iteritems():
#            setattr(self, attr, value)
#    #--- End: def

    def __hash__(self):
        '''

The built-in function `hash`

Generating the hash temporarily realizes the entire array in memory,
which may not be possible for large arrays.

The hash value is dependent on the data type, shape of the data
array. If the array is a masked array then the hash value is
independent of the fill value and of data array values underlying any
masked elements.

The hash value is not guaranteed to be portable across versions of
Python, numpy and cf.

.. note:: The hash value may be different if regenerated.

:Returns:

    out : int
        The hash value.

**Examples**

>>> print d.array
[[0 1 2 3]]
>>> d.hash()
-8125230271916303273
>>> d[1, 0] = numpy.ma.masked
>>> print d.array
[[0 -- 2 3]]
>>> hash(d)
791917586613573563
>>> d.hardmask = False
>>> d[0, 1] = 999
>>> d[0, 1] = numpy.ma.masked
>>> d.hash()
791917586613573563
>>> d.squeeze()
>>> print d.array
[0 -- 2 3]
>>> hash(d)
-7007538450787927902
>>> d.dtype = float
>>> print d.array
[0.0 -- 2.0 3.0]
>>> hash(d)
-4816859207969696442

'''
        return hash_array(self.unsafe_array)
    #--- End: def

    def __iter__(self):
        '''

Efficient iteration

x.__iter__() <==> iter(x)

**Examples**

>>> d = cf.Data([1, 2, 3], 'metres')
>>> for e in d:
...    print repr(e)
...
1
2
3

>>> d = cf.Data([[1, 2, 3], [4, 5, 6]], 'metres')
>>> for e in d:
...    print repr(e)
...
<CF Data: [1, ..., 3] metres>
<CF Data: [4, ..., 6] metres>

>>> d = cf.Data(34, 'metres')
>>> for e in d:
...     print repr(e)
...
TypeError: Iteration over 0-d Data

'''
        ndim = self._ndim

        if not ndim:
            raise TypeError("Iteration over 0-d %s" % self.__class__.__name__)
            
        elif ndim == 1:
             i = iter(self.unsafe_array)
             while 1:
                 yield i.next()

        elif ndim > 1:
            for i in xrange(self.shape[0]):
                d = self[i, ...]
                d.squeeze(0)
                yield d
    #--- End: def

    def __len__(self):
        '''

The built-in function `len`

x.__len__() <==> len(x)

**Examples**

>>> len(cf.Data([1, 2, 3]))
3

>>> len(cf.Data([[1, 2, 3]]))
1

>>> len(cf.Data([[1, 2, 3], [4, 5, 6]])
2

>>> len(cf.Data(1))
TypeError: len() of scalar Data

'''
        shape = self._shape
        if shape:
            return shape[0]

        raise TypeError("len() of scalar Data")
    #--- End: def

    def __nonzero__(self):
        '''

Truth value testing and the built-in operation `bool`

x.__nonzero__() <==> x != 0

**Examples**

>>> bool(cf.Data(1))
True
>>> bool(cf.Data([[False]]))
False
>>> bool(cf.Data([1, 2]))
ValueError: The truth value of Data with more than one element is ambiguous. Use d.any() or d.all()

'''
        if self._size == 1:
            return bool(self.unsafe_array)

        raise ValueError(
"The truth value of Data with more than one element is ambiguous. Use d.any() or d.all()")
    #--- End: def

    def __repr__(self):
        '''

The built-in function `repr`

x.__repr__() <==> repr(x)

'''
        return '<CF %s: %s>' % (self.__class__.__name__, str(self))
    #--- End: def

    def __str__(self):
        '''

The built-in function `str`

x.__str__() <==> str(x)

'''

        self_units = self.Units
        isreftime = self_units.isreftime

        if not self_units or self_units.equals(_units_1):
            units = ''            
        elif not isreftime:
            units = self_units.units
        else:
            units = getattr(self_units, 'calendar', '')
                     
        size = self._size
        ndim = self._ndim
        open_brackets  = '[' * ndim
        close_brackets = ']' * ndim
  
        first = self.datum(0)

        if size == 1:
            if isreftime and not self._isdt:
                # Convert reference time to datetime
                first = rt2dt(first, self_units).item()

            out = '%s%s%s %s' % (open_brackets, first, close_brackets,
                                 units)
            
        else:
            last = self.datum(-1)
            if isreftime and not self._isdt:
                # Convert reference times to datetimes
                first, last = rt2dt((first, last), self_units)

            if size >= 3:
                out = '%s%s, ..., %s%s %s' % (open_brackets, first,
                                              last, close_brackets,
                                              units)
            else:
                out = '%s%s, %s%s %s' % (open_brackets, first,
                                         last, close_brackets,
                                         units)
        #--- End: if
            
        return out
    #--- End: def

    def __getitem__(self, indices):
        '''

Evaluation of self[indices]

x.__getitem__(indices) <==> x[indices]

'''
        if self._size == 1 or indices is Ellipsis:
            return self.copy()

#        dim2position = dict([(dim, i) for i, dim in enumerate(dimensions)])

        indices, reverse_dimensions = _parse_indices(self, indices)

        # If it's a scalar, then just copy it. ??? dch trapped in _parse_indices??
#        if not self._ndim and not indices:
#            return self.copy()
        
        dimensions = self.dimensions
        directions = self.directions

        new_shape = tuple(map(_size_of_index, indices, self._shape))
        new_size  = long(reduce(operator_mul, new_shape, 1))

        new = Data.__new__(Data)

        new._shape      = new_shape
        new._size       = new_size
        new._ndim       = self._ndim
        new._fill_value = self._fill_value
        new._hardmask   = self._hardmask
        new._isdt       = self._isdt
        new.dimensions  = dimensions #[:]
        new.directions  = directions #self._copy_directions(directions)
        new.Units       = self.Units #.copy()
        new.dtype       = self.dtype

        dim2position = {}
        for i, dim in enumerate(dimensions):
            dim2position[dim] = i

        partitions = self.partitions

        new_partitions = PartitionMatrix(_overlapping_partitions(partitions, 
                                                                 indices, 
                                                                 dim2position,
                                                                 directions),
                                         partitions.dimensions[:])

#        new_shape = tuple(map(_size_of_index, indices, self._shape))
#        new_size  = long(reduce(mul, new_shape, 1)) #new._shape, 1))

        if new_size != self._size:
            new_partitions.set_location_map(dimensions)

        new.partitions = new_partitions

#        new._shape     = new_shape #tuple(map(_size_of_index, indices, self._shape))
#        new._size      = new_size
#        new._ndim      = self._ndim
#        new.dimensions = dimensions[:]
#        new.directions = self._copy_directions(directions)
#        new.Units      = self.Units.copy()
#        new._FillValue = self._FillValue
#        new.dtype      = self.dtype
#        new._hardmask  = self._hardmask

        # Reverse dimensions with flipped slices
        if reverse_dimensions:
            new.flip(reverse_dimensions)

        # Remove size 1 partition dimensions
        new_partitions.squeeze()

        return new
    #--- End: def

    def __setitem__(self, indices, value):
        '''

Assignment to self[indices]

x.__setitem__(indices, y) <==> x[indices]=y

Elements of a data array may be changed by assigning values to a
subspace. See `__getitem__` for details on how to define subspace of
the data array.

**Missing data**

The treatment of missing data elements during assignment to a subspace
depends on the value of the `hardmask` attribute. If it is True then
masked elements will notbe unmasked, otherwise masked elements may be
set to any value.

In either case, unmasked elements may be set, (including missing
data).

Unmasked elements may be set to missing data by assignment to the
`cf.masked` constant or by assignment to a value which contains masked
elements.

.. seealso:: `cf.masked`, `hardmask`, `setdata`

**Examples**

'''            
        def _mirror_index(index, size):
            '''

    Note that it assumed that the step of the input slice is 1.
    
    :Parameters:
    
        index : slice
    
        size : int
    
    :Returns:
    
        out : slice
    
    **Examples**
    
    >>> s = slice(2, 6)
    >>> t = _mirror_index(slice(2, 6), 8)
    >>> t
    slice(5, 1, -1)
    >>> range(8)[s]
    [2, 3, 4, 5]
    >>> range(8)[t]
    [5, 4, 3, 2]
    >>> range(7, -1, -1)[t]
    [2, 3, 4, 5]
    
    '''
            start, stop, step = index.indices(size)
            size -= 1
            start = size - start
            stop  = size - stop
            if stop < 0:
                stop = None
                
            return slice(start, stop, -1)
        #--- End: def

        conform_args = self.conform_args()

        # ------------------------------------------------------------
        # parse the indices
        # ------------------------------------------------------------
        indices, reverse_dimensions = _parse_indices(self, indices)

        scalar_value = False
        if value is cf_masked:
            scalar_value  = True
        else:
            copied = False
            if not isinstance(value, Data):
                # Convert to the value to a Data object
                value = type(self)(value, self.Units)
            else:
                if value.Units.equivalent(self.Units):
                    if not value.Units.equals(self.Units):                    
                        value = value.copy()                    
                        value.Units = self.Units
                        copied = True
                elif not value.Units:                    
                    value = value.copy()                    
                    value.override_units(self.Units)
                    copied = True
                else:
                    raise ValueError(" Some bad units %s, %s" %
                                     (repr(self.Units), repr(value.Units)))
#                value = value.copy()
#                if value.Units and value.Units.equivalent(self.Units):
#                    value.Units = self.Units
#                elif not value.Units:
#                    value.override_units(self.Units)
#                else:
#                    raise ValueError(" Some bad units %s, %s" %
#                                     (repr(self.Units), repr(value.Units)))
            #--- End: if

            value2dt = self._isdt and not value._isdt
#            value2rt = False
#            if self._isdt:
#                if not value._isdt:
#                    value2dt = True
#            elif self.Units.isreftime and value._isdt:
#                value2rt = True

            if value._size == 1:
                scalar_value = True
                
                if value2dt:
                    if not copied:
                        value = value.copy()
                    value.asdatetime()
#                elif rt:
#                    if not copied:
#                        value = value.copy()
#                    value.asreftime()

                value = value.datum(0) #first_datum
        #--- End: if
# if value.Units.equivalent(self.Units):
#                        if not value.Units.equals(self.Units):                    
#                            value = value.copy()                    
#                            value.Units = self.Units
#                    elif value.Units:
#                        raise ValueError(" Some bad units %s, %s" %
#                                         (repr(self.Units), repr(value.Units)))
#                #--- End: if

        if scalar_value:
            # --------------------------------------------------------
            # The value is logically scalar
            # --------------------------------------------------------           
            for partition in self.partitions.matrix.flat:
                p_indices, shape = partition.overlaps(indices)
                if p_indices is None:
                    # This partition does not overlap the indices
                    continue

                array = partition.dataarray(**conform_args)

                if value is cf_masked and not partition.masked:
                    array = array.view(numpy_ma_MaskedArray)
                    partition.subarray = array

#                array[p_indices] = value
                set_subspace(array, p_indices, value)
                partition.close()
            #--- End: for

            return
        #--- End: if

        # ------------------------------------------------------------
        # Still here? Then the value is not logically scalar.
        # ------------------------------------------------------------
        self_shape  = self._shape
        value_shape = value._shape
        
        shape0 = map(_size_of_index, indices, self_shape)
        self_ndim  = self._ndim
        value_ndim = value._ndim
        align_offset = self_ndim - value_ndim
        if align_offset >= 0:
            # self has more dimensions than other
            shape0   = shape0[align_offset:]
            shape1   = value_shape
            ellipsis = None 
            
            reverse_dimensions = [i - align_offset
                                  for i in reverse_dimensions
                                  if i >= align_offset]                
        else:
            # value has more dimensions than self
            v_align_offset = -align_offset
            if value_shape[:v_align_offset] != [1] * v_align_offset:
                # Can only allow value to have more dimensions then
                # self if the extra dimensions all have size 1.
                raise ValueError("Can't broadcast shape %s across shape %s" %
                                 (repr(value_shape), repr(self_shape)))
            
            shape1       = value_shape[v_align_offset:]
            ellipsis     = Ellipsis
            align_offset = 0
        #--- End: if

        # Find out which of the dimensions of value are to be
        # broadcast, and those which are not. Note that, as in numpy,
        # it is not allowed for a dimension in value to be larger than
        # a size 1 dimension in self
        base_value_indices       = []
        non_broadcast_dimensions = []

        for i, (a, b) in enumerate(izip(shape0, shape1)):
            if b == 1:
                base_value_indices.append(slice(None))
            elif a == b and b != 1:
                base_value_indices.append(None)
                non_broadcast_dimensions.append(i)
            else:
                raise ValueError("Can't broadcast shape %s across shape %s" %
                                 (repr(shape1), repr(shape0)))
        #--- End: for

        previous_location = ((-1,),) * self_ndim
        start             = [0]      * value_ndim

        save = conform_args['save']

        value.to_memory()

        for partition in self.partitions.matrix.flat:
            p_indices, shape = partition.overlaps(indices)        

            if p_indices is None:
                # This partition does not overlap the indices          
                continue

            # --------------------------------------------------------
            # Find which elements of value apply to this partition
            # --------------------------------------------------------
            value_indices = base_value_indices[:]
       
            for i in non_broadcast_dimensions:
                j                  = i + align_offset
                location           = partition.location[j][0]
                reference_location = previous_location[j][0]

                if location > reference_location:
                    stop             = start[i] + shape[j]
                    value_indices[i] = slice(start[i], stop)
                    start[i]         = stop

                elif location == reference_location:
                    value_indices[i] = previous_slice[i]

                elif location < reference_location:
                    stop             = shape[j]
                    value_indices[i] = slice(0, stop)
                    start[i]         = stop
            #--- End: for

            previous_location = partition.location
            previous_slice    = value_indices[:]
          
            for i in reverse_dimensions:
                value_indices[i] = _mirror_index(value_indices[i], shape1[i])

            if ellipsis:
                value_indices.insert(0, ellipsis)

            # --------------------------------------------------------
            # 
            # --------------------------------------------------------
#            v = value[tuple(value_indices)].varray
            if value2dt:
                v = value[tuple(value_indices)].dtarray
#            elif rt:
#                v = value[tuple(value_indices)].rtarray
            else:
                v = value[tuple(value_indices)].varray

            if not save:
                v = v.copy()

            # Update the partition's data
            array = partition.dataarray(**conform_args)
#            if numpy_ma_isMA(array):
            if partition.masked:
                if not conform_args['hardmask']:
                    # Soft mask
#                    array[p_indices] = v
                    set_subspace(array, p_indices, v)
                else:
                    # Hard mask
                    array.soften_mask()
                
#                    array = array[p_indices]
                    array = get_subspace(array, p_indices)
                
                    if array.mask is numpy_ma_nomask:
                        mask = True
                    else:
                        mask = ~numpy_ma_getmaskarray(array)
                
                    both_unmasked = mask & ~numpy_ma_getmaskarray(v)
                
#                    partition.subarray[p_indices] = numpy_ma_where(both_unmasked,
#                                                                   v, array)
                    set_subspace(partition.subarray, p_indices, 
                                 numpy_ma_where(both_unmasked, v, array))
                
                    partition.subarray.harden_mask()
            else:
#                array[p_indices] = v
                set_subspace(array, p_indices, v)
            #--- End: if

            partition.close()
        #--- End: For
            
    #--- End: def
    
    def dumpd(self, dimensions=None):
        '''

Serialize the data array as a dictionary following loosely the CFA
framework.

:Parameters:

    dimensions : sequence of str, optional

:Returns:

    out : dict

**Examples**

>>> x = d.dumpd()

>>> x = d.dumpd(dimensions=['time', 'latitude', 'longitutude'])

''' 
        if dimensions is not None:
            dimensions = list(dimensions)

            dim_name_map = {}
            for dim, ndim in izip(self.dimensions, dimensions):
                dim_name_map[dim] = ndim
                                
            xdata = self.copy()
            xdata.change_axis_names(dim_name_map)
 
        else:
            dimensions = self.dimensions[:]
            dim_name_map = None
            xdata = self
        #--- End: if

        data_isscalar = not xdata.ndim

        directions = xdata.directions
        cfa_data = {
            'dimensions'  : dimensions,
            'directions'  : xdata._copy_directions(directions),
            '_FillValue'  : xdata._fill_value,
            'dtype'       : xdata.dtype,
            'shape'       : xdata._shape,
            'Units'       : xdata.Units,        
            'pmshape'     : xdata.pmshape,
            'pmdimensions': xdata.pmdimensions,
            'Partitions'  : [],
            }

        for index, partition in xdata.partitions.ndenumerate():

            attrs = {}

            p_dimensions = partition._dimensions
            p_directions = partition._directions
            p_part       = partition.part                       
            p_Units      = partition.Units
            p_dtype      = partition.subarray.dtype
            
            # Location in partition matrix
            if index:
                attrs['index'] = index

            # Sub-array location
            attrs['location'] = partition.location
            
            # Sub-array part
            if p_part:
                attrs['part'] = p_part

            # Sub-array dimensions
            if p_dimensions != dimensions:
                attrs['dimensions'] = p_dimensions              

            # Sub-array units
            if p_Units != xdata.Units:
                attrs['Units'] = p_Units

            # Sub-array directions
            if p_directions != directions:
                if not p_dimensions or data_isscalar:
                    # The sub-array is a scalar or the data array is scalar
                    attrs['directions'] = p_directions                                  
                elif (not set(p_directions).issubset(directions) or
                      not all(p_directions[dim] == directions[dim]
                              for dim in p_dimensions)):
                    attrs['directions'] = p_directions
            #--- End: if

            # --------------------------------------------------------
            # File format specific stuff
            # --------------------------------------------------------
            if isinstance(partition.subarray, NetCDFFileArray):
                # ----------------------------------------------------
                # NetCDF File Array
                # ----------------------------------------------------
                attrs['format'] = 'netCDF'

                d = {}
                for attr in ('file', 'shape', 'ncvar'):
                    d[attr] = getattr(partition.subarray, attr)
                attrs['subarray'] = d         

                if p_dtype != xdata.dtype:
                    attrs['subarray']['dtype'] = p_dtype

                varid = getattr(partition.subarray, 'varid', None)
                if varid is not None:
                    attrs['subarray']['varid'] = varid
                      
            elif isinstance(partition.subarray, PPFileArray):
                # ----------------------------------------------------
                # PP File Array
                # ----------------------------------------------------
                attrs['format'] = 'PP'
            
                d = {}
#                for attr in ('file', 'shape', 'file_offset', 'lbpack',
#                             'scale_factor', 'add_offset'):
#                    value = getattr(partition.subarray, attr) #, None)
#                    if value is not None:
#                        d[attr] = value
                for attr in ('file', 'shape', 'file_offset'):
                    value = getattr(partition.subarray, attr)
                    if value is not None:
                        d[attr] = value
                #--- End: for

                for attr in ('lbpack', 'add_offset'):
                    value = getattr(partition.subarray, attr)
                    if value:
                        d[attr] = value
               #--- End: for

                value = partition.subarray.scale_factor
                if value != 1:
                    d['scale_factor'] = value

                value = partition.subarray.lblrec
                if value != partition.subarray.size:
                    d['lblrec'] = value

                attrs['subarray'] = d         


#                attrs['subarray'] = dict([(attr, getattr(partition.subarray, attr)) 
#                                          for attr in ('file', 'shape', 
#                                                       'file_offset', 'lbpack',
#                                                       'scale_factor', 'add_offset')
#                                          if hasattr(partition.subarray, attr)]) 
                
                if p_dtype != xdata.dtype:
                    attrs['subarray']['dtype'] = p_dtype
                      
                if (partition.subarray.dtype.byteorder == '<' or
                    partition.subarray.dtype.byteorder == '=' and
                    byteorder == 'little'):
                    attrs['subarray']['endian'] = 'little'

                fill_value = getattr(partition.subarray, '_FillValue', None)
                if fill_value is not None:
                    attrs['subarray']['_FillValue'] = fill_value

                      
            else:
                attrs['subarray'] = partition.subarray   
            #--- End: if
                 
            cfa_data['Partitions'].append(attrs)            
        #--- End: for

        return cfa_data
    #--- End: def
        
    def loadd(self, x):
        '''

Deserialize the dictionary, updating the master data array in place.

:Parameters:

:Returns:

    None

**Examples**

'''
        dimensions = list(x['dimensions'])
        directions = self._copy_directions(x.get('directions', True))
        units      = x['Units'] #.copy()
        shape      = tuple(x['shape'])
        dtype      = x['dtype']

        self.dimensions  = dimensions
        self.directions  = directions
        self.Units       = units
        self._fill_value = x['_FillValue']
        self._dtype      = dtype
        self._shape      = shape

        self._ndim = len(shape)
        self._size = long(reduce(operator_mul, shape, 1))
        
        filename = x.get('file', None)
        if filename is not None:
            filename = abspath(filename)

        base = x.get('base', None)
        if base is not None:
            base = abspath(base)

        # ------------------------------------------------------------
        # Initialize an empty partition array
        # ------------------------------------------------------------
        partition_matrix = PartitionMatrix(
            numpy_empty(x.get('pmshape', ()), dtype=object),
            x.get('pmdimensions', [])
            )
        
        # ------------------------------------------------------------
        # Fill the partition array with partitions
        # ------------------------------------------------------------
        for attrs in x['Partitions']:

            index = attrs.get('index', ())
            if len(index) == 1:
                index = index[0]
            else:
                index = tuple(index)

            partition = attrs.get('partition', None)
            if partition:
                partition_matrix[index] = partition.copy()
                continue

            partition = Partition(
                _dimensions = attrs.get('dimensions', dimensions),
                location    = attrs['location'],
                Units       = attrs.get('Units', units),
                part        = attrs.get('part', []),
                )
            
            format = attrs['format']

            # Set the 'directions' attribute
            d = attrs.get('directions', None)
            if d is not None:
                partition._directions = d
            else:
                # Take the default directions
                if not partition._dimensions:
                    if not dimensions:
                        partition._directions = directions
                    else:
                        partition._directions = True
                else:
                    d = {}
                    for dim in partition._dimensions:
                        d[dim] = directions.get(dim, True)
                    partition._directions = d
            #--- End: if

            # Set the 'subarray' attribute 
            kwargs = attrs['subarray'].copy()
            kwargs['shape'] = tuple(kwargs['shape'])
            if format in ('netCDF', 'PP'):
                kwargs['ndim'] = len(kwargs['shape'])
                kwargs['size'] = long(reduce(operator_mul, kwargs['shape'], 1))

                kwargs.setdefault('dtype', dtype)

                if 'file' in kwargs:
                    f = kwargs['file']
                    if f == '':
                        kwargs['file'] = filename
                    else:
                        if base is not None:
                            f = pathjoin(base, f)
                        kwargs['file'] = abspath(f)
                else:
                    kwargs['file'] = filename

                if format == 'netCDF':                
                    partition.subarray = NetCDFFileArray(**kwargs)
                elif format == 'PP':
                                        
                    endian = kwargs.get('endian', 'big')
                    kwargs['dtype'] = kwargs['dtype'].newbyteorder(endian)
                    kwargs['binary_mask'] = False

                    kwargs.setdefault('add_offset', 0)
                    kwargs.setdefault('scale_factor', 1)
                    kwargs.setdefault('lblrec', kwargs['size']) 

                    partition.subarray = PPFileArray(**kwargs)
            else:
                raise TypeError(
                    "Don't know how to load sub-array from file format '%s'" %
                    format)
            #--- End: if

            # Put the partition into the partition array 
            partition_matrix[index] = partition
        #--- End: for

        # Save the partition array
        self.partitions = partition_matrix

        self.chunk()
    #--- End: def

    def chunk(self, chunksize=None):
        '''

Partition the data array

:Parameters:

    chunksize : int, optional
        The 

:Returns:

    None

**Examples**

>>> d.chunk()
>>> d.chunk(100000)

'''
        if not chunksize:
            # Set the default chunk size
            chunksize = CHUNKSIZE()

        # Define the factor which, when multiplied by the size of a
        # partition, determines how many chunks are in the partition.
        factor = (self.dtype.itemsize + 1.0)/chunksize

        if self._size*factor <= 1:
            # Don't do any partitioning if the data array is already
            # smaller than the chunk size
            return

        # Initialize the dictionary relating each dimension to new
        # partition boundaries for that dimension
#        d = dict([(dim, set()) for dim in self.dimensions])
        dimensions = self.dimensions
        d = {} #dict_fromkeys(dimensions, set())

        # ------------------------------------------------------------
        # Find any new partition boundaries for each dimension
        # ------------------------------------------------------------
        for partition in self.partitions.matrix.flat:
            
            n_chunks = int(partition.size*factor + 0.5)
            
            if n_chunks <= 1:
                continue
            
            extra_boundaries = []
            chunk_dims       = []
            
            for size, location, dim in izip(partition.shape, 
                                            partition.location,
                                            dimensions):
                if size == 1 or n_chunks <= 1:
                    continue
                
                if size <= n_chunks:
                    d[dim] = range(location[0]+1, location[1])

                    n_chunks = int(math_ceil(n_chunks/float(size)))

                else:
                    step = int(size/n_chunks)
                    d[dim] = range(location[0]+step, location[1], step)
                    
                    break
                #--- End: if
            #--- End: for                 
        #--- End: for

        # ------------------------------------------------------------
        # Create any new partition boundaries for each dimension
        # ------------------------------------------------------------
        for dim in dimensions[::-1]: #, extra_bounds in d.iteritems():
            extra_bounds = d.get(dim, None)
#            extra_bounds = d[dim]

            if not extra_bounds:
                continue
            
            if dim not in self.partitions.dimensions:
                # Create a new partition dimension
                self.partitions.expand_dims(dim)

            # Create the new partitions
            self.add_partitions(sorted(extra_bounds), dim)
        #--- End: for       
    #--- End: def
    
    def _combined_units(self, other, method, inplace):
        '''

:Parameters:

    y : cf.Data

    method : str
        The 

:Returns:

    out : cf.Units

**Examples**

>>> d._combined_units(e, '__sub__')
>>> d._combined_units(e, '__imul__')
>>> d._combined_units(e, '__irdiv__')
>>> d._combined_units(e, '__lt__')
>>> d._combined_units(e, '__rlshift__')
>>> d._combined_units(e, '__iand__')


'''

        method_type  = method[-5:-2]

        x = self

        self_units  = self.Units
        other_units = other.Units

        if not self_units and not other_units:
            return x, other, self_units
        
        if (self_units.isreftime and other_units.isreftime and
            not self_units.equivalent(other_units)):
            if not other_units._calendar and self_units._calendar:
                other = other.copy()
                other.override_calendar(self_units._calendar)
                other.Units = self_units
            elif other_units._calendar and not self_units._calendar:
                if not inplace:
                    x = self.copy()
                x.override_calendar(other_units._calendar) 
                x.Units = other_units
        #--- End: if

        if method_type in ('_eq', '_ne', '_lt', '_le', '_gt', '_ge'):
            #---------------------------------------------------------
            # Operator is ==, !=, >=, >, <= or <
            #---------------------------------------------------------
            if self_units.equivalent(other_units): 
                if not self_units.equals(other_units):
                    other = other.copy()
                    other.Units = self_units
                return x, other, _units_None
            elif not other_units or not self_units:
                return x, other, _units_None           
            else: 
                raise ValueError("Can't compare %s to %s" %
                                 (repr(self_units), repr(other_units)))
        #--- End: if

        if method_type in ('and', '_or', 'ior', 'ror', 'xor', 'ift'):
            #---------------------------------------------------------
            # Operation is &, |, ^, >> or <<
            #---------------------------------------------------------
            if self_units.equivalent(other_units):
                if not self_units.equals(other_units):
                    other = other.copy()
                    other.Units = self_units
                return x, other, self_units
            elif not other_units:
                return x, other, self_units
            elif not self_units:
                return x, other, other_units

        if self_units.isreftime:
            #---------------------------------------------------------
            # self is reference time
            #---------------------------------------------------------
            if method_type == 'sub':
                if other_units.isreftime:
                    if self_units.equivalent(other_units):
                        # If (reference_time-reference_time) then the
                        # output units are time 
                        if not self_units.equals(other_units):
                            other = other.copy()
                            other.Units = self_units
                        return x, other, Units(_ut_unit=self_units._ut_unit)
                    else:
                        return x, other, self_units - other_units
                elif other_units.istime:
                    # If (reference_time-time) then the output units
                    # are reference_time
                    self_time = Units(_ut_unit=self_units._ut_unit)
                    if not other_units.equals(self_time):
                        other = other.copy()
                        other.Units = self_time
                    return x, other, self_units
                elif not other_units:
                    # If (reference_time-no_units) then the output
                    # units are reference_time
                    return x, other, self_units
                else:
                    return x, other, self_units - other_units

            elif method_type == 'add':
                if other_units.istime:
                    # If (reference_time+time) then the output units
                    # are reference_time ...
                    self_time = Units(_ut_unit=self_units._ut_unit)
                    if not other_units.equals(self_time):
                        other = other.copy()
                        other.Units = Units(_ut_unit=self_units._ut_unit)
                    return x, other, self_units
                elif not other_units:
                    # If (reference_time+no_units) then the output
                    # units are reference_time
                    return x, other, self_units
                else:
                    return x, other, self_units + other_units

            else:
                return x, other, getattr(self_units, method)(other_units)

        elif other_units.isreftime:
            #---------------------------------------------------------
            # other is reference time
            #---------------------------------------------------------
            if method_type == 'add':
                if self_units.istime:
                    # If (time+reference_time) then the output units are
                    # reference_time
                    self_time = Units(_ut_unit=self_units._ut_unit)
                    if not other_units.equals(self_time):
                        other = other.copy()
                        other.Units = self_time
                    return x, other, other_units
                elif not self_units:
                    # If (no_units+reference_time) then the output units are
                    # reference_time
                    return x, other, other_units
                else:
                    return x, other, self_units + other_units
        #--- End: if

        if method_type in ('mul', 'div'):
            #---------------------------------------------------------
            # Method is *, / or //
            #---------------------------------------------------------
            if not other_units:
                return x, other, getattr(self_units, method)(_units_1)
            elif not self_units:
                return x, other, getattr(_units_1, method)(other_units)
            else:
                return x, other, getattr(self_units, method)(other_units)

        if method_type in ('sub', 'add'):
            #---------------------------------------------------------
            # Operator is + or -
            #---------------------------------------------------------
            if self_units.equivalent(other_units):
                if not self_units.equals(other_units):
                    other = other.copy()
                    other.Units = self_units
                return x, other, self_units
            elif not other_units:
                return x, other, self_units
            elif not self_units:
                return x, other, other_units
            else:
                return x, other, getattr(self_units, method)(other_units)
        #--- End: if

        if method_type == 'pow':
            #---------------------------------------------------------
            # Operator is **
            #---------------------------------------------------------
            if other._size > 1:
                raise ValueError(
"Can't raise %s to the power of an array with more than one element" %
self.__class__.__name__)

            if not other_units:
                return x, other, self_units ** other.datum(0)
            elif other_units.isdimensionless:
                if not self_units.equals(_units_1):
                    other = other.copy()
                    other.Units = _units_1
                return x, other, self_units ** other.datum(0)
            else:
                return x, other, self_units ** other_units
        #--- End: if

        raise ValueError("Can't operate with %s on data with %s to %s" %
                         (method, repr(self_units), repr(other_units)))
    #--- End: def

    def _binary_operation(self, other, method):
        '''

Implement binary arithmetic and comparison operations with the numpy
broadcasting rules.

It is called by the binary arithmetic and comparison
methods, such as `__sub__`, `__imul__`, `__rdiv__`, `__lt__`, etc.

:Parameters:

    other : object
        The object on the right hand side of the operator.

    method : str
        The binary arithmetic or comparison method name (such as
        "__imul__" or "__ge__").

:Returns:

    new : Data
        A new Data object, or if the operation was in place, the same
        Data object.

.. seealso:: `_unary_operation`

**Examples**

>>> d = cf.Data([0, 1, 2, 3])
>>> e = cf.Data([1, 1, 3, 4])

>>> f = d._binary_operation(e, '__add__')
>>> print f.array
[1 2 5 7]

>>> e = d._binary_operation(e, '__lt__')
>>> print e.array
[ True False  True  True]

>>> d._binary_operation(2, '__imul__')
>>> print d.array
[0 2 4 6]

'''
        inplace      = (method[2] == 'i')
        method_type  = method[-5:-2]

        # ------------------------------------------------------------
        # Ensure that other is an independent Data object
        # ------------------------------------------------------------    
        if getattr(other, '_NotImplemented_RHS_Data_op', False):
            # Make sure that 
            return NotImplemented

        elif not isinstance(other, self.__class__):
            other = type(self)(other)

            if other._isdt and self.Units.isreftime:
                # Make sure that an array of datetime-like objects has
                # the right calendar.
                other.override_units(self.Units)
        #--- End: if
                
        xself, other, new_Units = self._combined_units(other, method, inplace)

#        if isinstance(other, self.__class__):
#            # --------------------------------------------------------
#            # other is Data object
#            # --------------------------------------------------------
#            other = other.copy()            
#            result_Units = self._combined_units(other, method)
#
#
#        elif isinstance(other, (float, int, long)):
#            # --------------------------------------------------------
#            # other is a number
#            # --------------------------------------------------------
#            if method_type == 'pow':
#                result_Units = self.Units ** other
#            elif method_type in ('_eq', '_ne', '_lt', '_le', '_gt', '_ge'):
#                result_Units = _units_None
#            else:
#                result_Units = self.Units #.copy()
#
#            other = type(self)(other)
#            
#        elif isinstance(other, numpy_ndarray):
#            # --------------------------------------------------------
#            # other is a numpy array
#            # --------------------------------------------------------
#            if method_type == 'pow':
#                if other.size > 1:
#                    raise ValueError(
#"Can't raise %s to the power of an array with more than one element" %
#self.__class__.__name__)
#
#                result_Units = self.Units ** other.item()
#            elif method_type in ('_eq', '_ne', '_lt', '_le', '_gt', '_ge'):
#                result_Units = _units_None
#            else:
#                result_Units = self.Units #.copy()
#
#            other = type(self)(other)
#   
#        elif isinstance(other, (list, tuple)):
#            # --------------------------------------------------------
#            #
#            # --------------------------------------------------------
#            if method_type == 'pow':
#                if len(other) > 1:
#                    raise ValueError(
#"Can't raise %s to the power of a sequence with more than one element" %
#self.__class__.__name__)
#
#                result_Units = self.Units ** other.varray.item()
#            elif method_type in ('_eq', '_ne', '_lt', '_le', '_gt', '_ge'):
#                result_Units = _units_None
#            else:
#                result_Units = self.Units #.copy()
#
#            other = type(self)(other)
#   
#        else:
#            # --------------------------------------------------------
#            # other is something else
#            # --------------------------------------------------------
#            return NotImplemented
#        #--- End: if

        # ------------------------------------------------------------
        # Bring other into memory, if appropriate.
        # ------------------------------------------------------------
        other.to_memory()

        copy_directions = xself._copy_directions

        # ------------------------------------------------------------
        # Find which dimensions need to be broadcast in one or other
        # of the arrays.
        #
        # Method:
        #
        #   For each common dimension, the 'broadcast_indices' list
        #   will have a value of None if there is no broadcasting
        #   required (i.e. the two arrays have the same size along
        #   that dimension) or a value of slice(None) if broadcasting
        #   is required (i.e. the two arrays have the different sizes
        #   along that dimension and one of the sizes is 1).
        #   
        #   Example:
        #
        #     If c.shape is (7,1,6,1,5) and d.shape is (6,4,1) then
        #     broadcast_indices will be
        #     [None,slice(None),slice(None)].
        #   
        #     The indices to d which correspond to a partition of c,
        #     are the relevant subset of partition.indices updated
        #     with the non None elements of the broadcast_indices
        #     list.
        #   
        #     In this example, if a partition of c were to have a
        #     partition.indices value of (slice(0,3), slice(0,1),
        #     slice(2,4), slice(0,1), slice(0,5)), then the relevant
        #     subset of these is partition.indices[2:] and the
        #     corresponding indices to d are (slice(2,4), slice(None),
        #     slice(None))
        #
        # ------------------------------------------------------------
        self_shape  = xself._shape
        other_shape = other._shape
        if self_shape == other_shape:
            # self and other have the same shapes
            align_offset = 0
            ellipsis     = None

            new_shape      = self_shape
            new_ndim       = xself._ndim
            new_dimensions = xself.dimensions
            new_directions = xself.directions
            new_size       = xself._size

            broadcasting = False
        else:
            # self and other have different shapes
            self_directions = xself.directions
            self_ndim       = xself._ndim
            other_ndim      = other._ndim

            align_offset = self_ndim - other_ndim
            if align_offset >= 0:
                # self has the same amount or more axes than other
                shape0 = self_shape[align_offset:]
                shape1 = other_shape

                new_shape      = self_shape[:align_offset]
                new_ndim       = self_ndim
                new_dimensions = xself.dimensions
                new_directions = self_directions #copy_directions(self.directions)
            else:
                # other has more axes than self
                align_offset = -align_offset
                shape0       = self_shape
                shape1       = other_shape[align_offset:]

                new_shape      = other_shape[:align_offset]
                new_ndim       = other_ndim
                if not self_ndim:
                    new_dimensions = other.dimensions
                else:
                    new_dimensions = []
                    existing_axes = self.all_axis_names()
                    for n in new_shape:
                        axis = xself.new_axis_identifier(existing_axes)
                        existing_axes.add(axis)
                        new_dimensions.append(axis)
                    #--- End: for
                    new_dimensions += xself.dimensions
                #--- End: for
            
                if not self_shape:
                    new_directions = other.directions.copy()
                else:
                    new_directions = {}
                    other_dimensions = other.dimensions
                    other_directions = other.directions
                    for axis, x in izip(new_dimensions[:align_offset],
                                        other_dimensions):
                        new_directions[axis] = other_directions[x]
                        
                    for axis, x in izip(new_dimensions[align_offset:],
                                        other_dimensions):
                        new_directions[axis] = self_directions[x]
                #--- End: if
            
                align_offset = 0
            #--- End: if
            
            broadcast_indices = []
            for a, b in izip(shape0, shape1):
                if a == b:
                    new_shape += (a,)
                    broadcast_indices.append(None)
                    continue
                
                # Still here?
                if a > 1 and b == 1:
                    new_shape += (a,)
                elif b > 1 and a == 1:
                    new_shape += (b,)
                else:
                    raise ValueError("Can't broadcast shape %s against shape %s" %
                                     (other_shape, self_shape))
                
                broadcast_indices.append(slice(None))
            #--- End: for

            new_size = long(reduce(operator_mul, new_shape, 1))

            broadcasting   = True
            dummy_location = [None] * new_ndim
        #---End: if

        # ------------------------------------------------------------
        # Create a Data object which just contains the metadata for
        # the result. If we're doing a binary arithmetic operation
        # then result will get filled with data and returned. If we're
        # an augmented arithmetic assignment then we'll update self
        # with this new metadata.
        # ------------------------------------------------------------

        #if new_shape != self_shape:
        #    set_location_map = True
        #    new_size = self._size
        #    dummy_location   = [None] * new_ndim
        #else:
        #    set_location_map = False
        #    new_size = long(reduce(mul, new_shape, 1))

#        if not set_location_map:
#            new_size = long(reduce(mul, new_shape, 1))
#        else:
#            new_size = self._size

        result            = xself.copy()
        result._shape     = new_shape
        result._ndim      = new_ndim 
        result._size      = new_size 
        result.dimensions = new_dimensions
        result.directions = new_directions

        # Is the result an array of datetime objects?
        new_isdt = xself._isdt and new_Units.isreftime

        # ------------------------------------------------------------
        # Set the data type of the result
        # ------------------------------------------------------------
        if method_type in ('_eq', '_ne', '_lt', '_le', '_gt', '_ge'):
            new_dtype = numpy_dtype(bool)
        elif not new_isdt:
            new_dtype = numpy_result_type(xself.dtype, other.dtype) 
        else:
            new_dtype = numpy_dtype(object)

        # ------------------------------------------------------------
        # Set flags to control whether or not the data of result and
        # self should be kept in memory
        # ------------------------------------------------------------
        save_result = result.save_to_disk(new_dtype.itemsize)
        if not inplace:
            # When doing a binary arithmetic operation we need to
            # decide whether or not to keep self's data in memory
            revert_to_file = True
            save_self      = xself.save_to_disk(xself.dtype.itemsize)
        else:
            # When doing an augmented arithmetic assignment we don't
            # need to keep self's original data in memory
            revert_to_file = False
            save_self      = False
        #--- End: if

#        dimensions     = self.dimensions
#        direction = self.direction
#        units     = self.Units

        conform_args = xself.conform_args(save=save_self,
                                          revert_to_file=revert_to_file)


        if xself._isdt:
            # If self is a datatime array then make sure that it gets
            # converted to a reference time array prior to the
            # operation
            conform_args['func']   = dt2rt
            conform_args['update'] = False
        #--- End: if                    

# Think about dtype, here.

        for partition_r, partition_s in izip(result.partitions.matrix.flat,
                                             xself.partitions.matrix.flat):

            indices = partition_s.indices
            array0  = partition_s.dataarray(**conform_args)

            if broadcasting:
                indices = tuple([
                        (index if not broadcast_index else broadcast_index) 
                        for index, broadcast_index in izip(
                            indices[align_offset:], broadcast_indices)
                        ])
                indices = (Ellipsis,) + indices
            #--- End: if

#            if other._isdt:
#                # If other is a datatime array then make sure that it
#                # gets converted to a reference time array prior to
#                # the operation
#                array1 = other[indices].rtarray
#            else:
            array1 = other[indices].array #unsafe_array

            # ISSUE: array1 could be much larger than the chunk size.

            if not inplace:
                partition = partition_r
                partition.update_inplace_from(partition_s)
            else:
                partition = partition_s

            # --------------------------------------------------------
            # Do the binary operation on this partition's data
            # --------------------------------------------------------
            array0 = getattr(array0, method)(array1)

            if new_isdt:
                # Convert result array to a datetime object array
                array0 = rt2dt(array0, xself.Units)
 
            if array0 is NotImplemented:
                array0 = numpy_zeros(partition.shape, dtype=bool)
                
            partition.subarray    = array0
            partition.Units       = new_Units
            partition._dimensions = new_dimensions
            partition._directions = new_directions

            if broadcasting:
                partition.location = dummy_location
                partition.shape    = list(array0.shape)
            #--- End: if

            partition.close(save_result)            

            if not inplace:
                partition_s.close() 
        #--- End: for

        if not inplace:
            result._isdt = new_isdt
            result.Units = new_Units
            result.dtype = new_dtype

            if broadcasting:
                result.partitions.set_location_map(result.dimensions)

            return result
        else:
            # Update the metadata for the new master array in place
            xself._isdt      = new_isdt
            xself._shape     = new_shape
            xself._ndim      = new_ndim
            xself._size      = new_size
            xself.Units      = new_Units
            xself.dtype      = new_dtype
            xself.dimensions = new_dimensions
            xself.directions = new_directions
                
            if broadcasting:
                xself.partitions.set_location_map(new_dimensions)

            self.__dict__ = xself.__dict__

            return self
    #--- End: def

    def _extend_pm(self, other, reversed=False):
        '''
     
:Parameters:

    other : cf.Data

    reversed : bool

:Returns:

    None

**Examples**

'''
        partitions = self.partitions
        matrix = partitions.matrix

        partitions2 = other.partitions
        matrix2 = partitions2.matrix

        if not matrix.ndim:
            new_shape = [1]
            shape0 = 1
        else:
            new_shape = list(matrix.shape)
            shape0 = new_shape[0]
       
        if matrix2.size == 1:
            new_shape[0] += 1
        else:
            new_shape[0] += matrix2.shape[0]

        new_matrix = numpy_empty(new_shape, dtype=object)

        new_matrix[:shape0] = matrix
        if not reversed:
            new_matrix[shape0:] = matrix2
        else:
            new_matrix[:shape0-1:-1] = matrix2
         
        partitions.matrix = new_matrix
    #--- End: def

    def _prepend_pm(self, other, reversed=False):
        '''

:Parameters:

:Returns:

    None

**Examples**

'''
        partitions = self.partitions
        matrix = partitions.matrix

        partitions2 = other.partitions
        matrix2 = partitions2.matrix

        if not matrix.ndim:
            new_shape = [1]
        else:
            new_shape  = list(matrix.shape)

        if matrix2.size == 1:
            other_shape0 = 1
        else:
            other_shape0 = matrix2.shape[0]

        new_shape[0] += other_shape0
        
        new_matrix = numpy_empty(new_shape, dtype=object)

        if not reversed:
            new_matrix[:other_shape0] = matrix2
        else:
            new_matrix[other_shape0-1::-1] = matrix2

        new_matrix[other_shape0:] = matrix
           
        partitions.matrix = new_matrix
    #--- End: def

    def _insert_data(self, other, extend, reversed,
                     PDim, PDim_direction, dim_name_map):
        '''

'''
        data0 = self
        data1 = other
            
        # ------------------------------------------------------------
        # 1. Change the dimension names in data1 to match those in the
        #    the domain of data0.
        # ------------------------------------------------------------
        data1.change_axis_names(dim_name_map)

        # ------------------------------------------------------------
        # 2. Insert (size 1) dimensions into data0 which are present
        #    in data1 but not in data0
        #
        # Note that the dimensions to be inserted into data0 will
        # always have size 1 in data1 (because the two domains are
        # aggregatable).
        #
        # Note that after this process data1.dimensions will be a
        # subset of data0.dimensions (because the two domains are
        # aggregatable and so have the same rank).
        # ------------------------------------------------------------
        dimensions0 = data0.dimensions
        directions1 = data1.directions
        for dim1 in data1.dimensions[::-1]:
            if dim1 not in dimensions0:
                data0.expand_dims(0, dim1, directions1[dim1])
        #--- End: for
    
        # ------------------------------------------------------------
        # 3. If the Aggregating Dimension is missing from data0 then
        #    insert it with size 1 in the slowest varying position.
        # 
        # Note that after this process data1.dimensions will be a
        # subset of data0.dimensions
        # ------------------------------------------------------------
        if PDim not in data0.dimensions:
            data0.expand_dims(0, PDim, PDim_direction)

        # ------------------------------------------------------------
        # 4. Permute the dimensions of data1 to match data0.
        #
        # For example: If data0.dimensions=['dim0','dim1','dim2'] and
        #              data1.dimensions=['dim2','dim0'] to start with
        #              then data1.dimensions will become
        #              ['dim0','dim2']
        #
        # Note that by now, data1.dimensions is a subset of
        # data0.dimensions (this was was ensured by steps 2 and 3).
        # ------------------------------------------------------------
        dimensions1 = data1.dimensions
        axes   = [dimensions1.index(dim)
                  for dim in data0.dimensions if dim in dimensions1] 
        data1.transpose(axes)

        # ------------------------------------------------------------
        # 5. Insert (size 1) dimensions into data1 which are present
        #    in data0 but not in data1.
        #
        # For example: If data0.dimensions=['dim0','dim1','dim2'] and
        #              data1.dimensions=['dim0','dim2'] to start with
        #              then data1.dimensions will become
        #              ['dim0','dim1','dim2']
        #
        # Note that after this process data0 and data1 will span the
        # same dimensions.
        #
        # Note that the orders of dimensions in data0 and data1 have
        # to be the same (so that their partitions' locations
        # conform).
        #
        # Note that after this process the Aggregating Dimension will
        # definitely be in both data0 and data1.
        # ------------------------------------------------------------
        directions0 = data0.directions
        dimensions1 = data1.dimensions
        for i, dim in enumerate(data0.dimensions):
            if dim not in dimensions1:
                data1.expand_dims(i, dim, directions0)
        #--- End: for

        # ------------------------------------------------------------
        # 6. Make sure that the partition dimensions of data0 are the
        #    same as data1 (although, for now, they may have different
        #    orders).
        #
        # Note that this may involve adding new partition dimensions
        # to either or both of data0 and data1.
        #            
        # Note that if the Aggregating Dimension needs to be added it
        # is inserted as the outer (slowest varying) dimension to
        # reduce the likelihood of having to transpose (expensively)
        # the partition array.
        # ------------------------------------------------------------
        for f, g in izip((data0, data1), 
                         (data1, data0)):

            g_pmdimensions = g.partitions.dimensions[:]
            try:
                g_pmdimensions.remove(PDim)
            except ValueError:
                pass
    
            f_pmdimensions = f.partitions.dimensions[:]
            for pdim in g_pmdimensions[::-1]:
                if pdim not in f_pmdimensions:
                    f.expand_partitions_dims(pdim)

            if PDim not in f.partitions.dimensions:
                f.expand_partition_dims(PDim)
        #--- End: for

        # ------------------------------------------------------------
        # 7. Permute the partition dimensions of the partition array
        #    of data0 so that the Aggregating Dimension is the
        #    outermost (slowest varying) dimension.
        # ------------------------------------------------------------
        axis = data0.partitions.dimensions.index(PDim)
        data0.partitions.rollaxis(axis, 0)
            
        # ------------------------------------------------------------
        # 8. Permute the partition dimensions of the partition array
        #    of data1 to match data0.
        # ------------------------------------------------------------
        pmdimensions1 = data1.partitions.dimensions
        axes = [pmdimensions1.index(dim) for dim in data0.partitions.dimensions]
        data1.partitions.transpose(axes)
            
        # ------------------------------------------------------------
        # 9. Create new partition boundaries in the partition arrays
        #    of data0 and data1 so that their partition arrays may be
        #    considered as different slices of a common, larger
        #    hyperrectangular partition array.
        #
        # Note that there is no need to add any boundaries across the
        # Aggregating Dimension.
        # ------------------------------------------------------------
        boundaries0 = data0.partition_boundaries()        
        boundaries1 = data1.partition_boundaries()

        for dim in data0.partitions.dimensions[1:]:
            
            # Still here? Then see if there are any partition
            # boundaries to be created for this partition dimension
            bounds0 = boundaries0[dim]
            bounds1 = boundaries1[dim]

            symmetric_diff = set(bounds0).symmetric_difference(bounds1)            
            if not symmetric_diff:
                # The partition boundaries for this partition
                # dimension are already the same in data0 and data1
                continue
            
            # Still here? Then there are some partition boundaries to
            # be created for this partition dimension in data0 and/or
            # data1.
            for f, g, bf, bg in izip((data0, data1, bounds0, bounds1), 
                                     (data1, data0, bounds1, bounds0)):                
                extra_bounds = [i
                                for i in bg
                                if i in symmetric_diff]
                
                f.add_partitions(extra_bounds, dim)
            #--- End: for
        #--- End: for
                
        # ------------------------------------------------------------
        # 10. Merge data1 into data0 along the Aggregating Dimension.
#        #
#        # If no indices have been provided then, along the Aggregating
#        # Dimension, concatenate the partitions of data1 onto the end
#        # of the partitions of data0.
        # ------------------------------------------------------------
#        if indices is None:
#            npartitions0 = data0.npartitions
#            indices = xrange(npartitions0, npartitions0+data1.npartitions)
#        #--- End: if
#
#        partitions0 = data0.partitions
#        partitions1 = data1.partitions
#        for i, index in enumerate(indices):
#            partitions0.insert(index, partitions1[i])
                                          
        if extend:
#            data0.partitions._extend(data1.partitions, reversed=reversed)
            data0._extend_pm(data1, reversed=reversed)
        else:
            data0._prepend_pm(data1, reversed=reversed)
 
        # Update the size of data0
        data0._size += long(data1._size)

        # Update the shape of data0
        index = data1.dimensions.index(PDim)
        shape = list(data0._shape)
        shape[index] += data1._shape[index]
        data0._shape = tuple(shape)

        # Update the location map of the partition array of data0
        data0.partitions.set_location_map(data0.dimensions)

        # Set the new dtype
        dtype0 = data0.dtype
        dtype1 = data1.dtype
        if dtype0 != dtype1:
            data0.dtype = numpy_result_type(dtype0, dtype1)        
        
        # ------------------------------------------------------------
        # 11. Done
        # ------------------------------------------------------------
    #--- End: def

    def _copy_directions(self, directions):
        '''
'''
        if isinstance(directions, dict):
            return directions.copy()

        return directions
    #--- End: def

    def _parse_axes(self, axes, method):
        '''
        
:Parameters:

    axes : (sequence of) int or str
        The axes of the data array. May be one of, or a sequence of
        any combination of zero or more of:

            * The integer position of a dimension in the data array
              (negative indices allowed).
            * The internal name a dimension.

    method : str

:Returns:

    out : list

**Examples**

'''
        data_axes = self.dimensions
        ndim      = self._ndim

        if isinstance(axes, (str, int, long)):
            axes = (axes,)
            
        axes2 = []
        for axis in axes:
            if axis in data_axes:
                axes2.append(data_axes.index(axis))
            elif axis in xrange(ndim):
                axes2.append(axis)
            elif axis in xrange(-ndim, 0):
                axes2.append(axis + ndim)
            else:
                raise ValueError(
                    "Can't %s: Invalid axis: %s" % (method, repr(axis)))
        #--- End: for
            
        # Check for duplicate axes
        if len(axes2) != len(set(axes2)):
            raise ValueError("Can't %s: Duplicate axis: %s" %
                             (method, repr(axes2)))            

#        if dims:
#            # Convert axes from integer positions to axis identifiers
#            data_axes = self.dimensions
#            return axes, [data_axes[i] for i in axes]
#        else:
#            return axes

        return axes2
#    #--- End: def

    def _unary_operation(self, operation):
        '''

Implement unary arithmetic operations.

It is called by the unary arithmetic methods, such as
__abs__().

:Parameters:

    operation : str
        The unary arithmetic method name (such as "__invert__").

:Returns:

    new : Data
        A new Data array.

.. seealso:: `_binary_operation`

**Examples**

>>> d = cf.Data([1, 2, -3, -4, -5])

>>> e = d._unary_operation('__abs__')
>>> print e.array
[1 2 3 4 5]]

>>> e = d.__abs__()
>>> print e.array
[[1 2 3 4 5]]

>>> e = abs(d)
>>> print e.array
[[1 2 3 4 5]]

'''
        self.to_memory()

        new = self.copy()

        conform_args = new.conform_args()

        for partition in new.partitions.matrix.flat:
            array = partition.dataarray(**conform_args)
#            partition.subarray = getattr(array, operation)()
            partition.subarray = getattr(operator, operation)(array)            
            partition.close()
        #--- End: for

        return new
    #--- End: def

    def __add__(self, other):
        '''

The binary arithmetic operation ``+``

x.__add__(y) <==> x+y

'''
        return self._binary_operation(other, '__add__')
    #--- End: def

    def __iadd__(self, other):
        '''

The augmented arithmetic assignment ``+=``

x.__iadd__(y) <==> x+=y

'''
        return self._binary_operation(other, '__iadd__')
    #--- End: def

    def __radd__(self, other):
        '''

The binary arithmetic operation ``+`` with reflected operands

x.__radd__(y) <==> y+x

'''
        return self._binary_operation(other, '__radd__')
    #--- End: def

    def __sub__(self, other):
        '''

The binary arithmetic operation ``-``

x.__sub__(y) <==> x-y

'''
        return self._binary_operation(other, '__sub__')
    #--- End: def

    def __isub__(self, other):
        '''

The augmented arithmetic assignment ``-=``

x.__isub__(y) <==> x-=y

'''
        return self._binary_operation(other, '__isub__')
    #--- End: def

    def __rsub__(self, other):
        '''

The binary arithmetic operation ``-`` with reflected operands

x.__rsub__(y) <==> y-x

'''
        return self._binary_operation(other, '__rsub__')
    #--- End: def

    def __mul__(self, other):
        '''

The binary arithmetic operation ``*``

x.__mul__(y) <==> x*y

'''
        return self._binary_operation(other, '__mul__')
    #--- End: def

    def __imul__(self, other):
        '''

The augmented arithmetic assignment ``*=``

x.__imul__(y) <==> x*=y

'''
        return self._binary_operation(other, '__imul__')
    #--- End: def

    def __rmul__(self, other):
        '''

The binary arithmetic operation ``*`` with reflected operands

x.__rmul__(y) <==> y*x

'''
        return self._binary_operation(other, '__rmul__')
    #--- End: def

    def __div__(self, other):
        '''

The binary arithmetic operation ``/``

x.__div__(y) <==> x/y

'''
        return self._binary_operation(other, '__div__')
    #--- End: def

    def __idiv__(self, other):
        '''

The augmented arithmetic assignment ``/=``

x.__idiv__(y) <==> x/=y

'''
        return self._binary_operation(other, '__idiv__')
    #--- End: def

    def __rdiv__(self, other):
        '''

The binary arithmetic operation ``/`` with reflected operands

x.__rdiv__(y) <==> y/x

'''
        return self._binary_operation(other, '__rdiv__')
    #--- End: def

    def __floordiv__(self, other):
        '''

The binary arithmetic operation ``//``

x.__floordiv__(y) <==> x//y

'''
        return self._binary_operation(other, '__floordiv__')
    #--- End: def

    def __ifloordiv__(self, other):
        '''

The augmented arithmetic assignment ``//=``

x.__ifloordiv__(y) <==> x//=y

'''
        return self._binary_operation(other, '__ifloordiv__')
    #--- End: def

    def __rfloordiv__(self, other):
        '''

The binary arithmetic operation ``//`` with reflected operands

x.__rfloordiv__(y) <==> y//x

'''
        return self._binary_operation(other, '__rfloordiv__')
    #--- End: def

    def __truediv__(self, other):
        '''

The binary arithmetic operation ``/`` (true division)

x.__truediv__(y) <==> x/y

'''
        return self._binary_operation(other, '__truediv__')
    #--- End: def

    def __itruediv__(self, other):
        '''

The augmented arithmetic assignment ``/=`` (true division)

x.__itruediv__(y) <==> x/=y

'''
        return self._binary_operation(other, '__itruediv__')
   #--- End: def

    def __rtruediv__(self, other):
        '''

The binary arithmetic operation ``/`` (true division) with reflected
operands

x.__rtruediv__(y) <==> y/x

'''
        return self._binary_operation(other, '__rtruediv__')
    #--- End: def

    def __pow__(self, other, modulo=None):
        '''

The binary arithmetic operations ``**`` and ``pow``

x.__pow__(y) <==> x**y

'''  
        if modulo is not None:
            raise NotImplementedError("3-argument power not supported for '%s'" %
                                      self.__class__.__name__)

        return self._binary_operation(other, '__pow__')
    #--- End: def

    def __ipow__(self, other, modulo=None):
        '''

The augmented arithmetic assignment ``**=``

x.__ipow__(y) <==> x**=y

'''  
        if modulo is not None:
            raise NotImplementedError("3-argument power not supported for '%s'" %
                                      self.__class__.__name__)

        return self._binary_operation(other, '__ipow__')
    #--- End: def

    def __rpow__(self, other, modulo=None):
        '''

The binary arithmetic operations ``**`` and ``pow`` with reflected
operands

x.__rpow__(y) <==> y**x

'''  
        if modulo is not None:
            raise NotImplementedError("3-argument power not supported for '%s'" %
                                      self.__class__.__name__)

        return self._binary_operation(other, '__rpow__')
    #--- End: def

    def __eq__(self, other):
        '''

The rich comparison operator ``==``

x.__eq__(y) <==> x==y

'''
        return self._binary_operation(other, '__eq__')
    #--- End: def

    def __ne__(self, other):
        '''

The rich comparison operator ``!=``

x.__ne__(y) <==> x!=y

'''
        return self._binary_operation(other, '__ne__')
    #--- End: def

    def __ge__(self, other):
        '''

The rich comparison operator ``>=``

x.__ge__(y) <==> x>=y

'''
        return self._binary_operation(other, '__ge__')
    #--- End: def

    def __gt__(self, other):
        '''

The rich comparison operator ``>``

x.__gt__(y) <==> x>y

'''
        return self._binary_operation(other, '__gt__')
    #--- End: def

    def __le__(self, other):
        '''

The rich comparison operator ``<=``

x.__le__(y) <==> x<=y

'''
        return self._binary_operation(other, '__le__')
    #--- End: def

    def __lt__(self, other):
        '''

The rich comparison operator ``<``

x.__lt__(y) <==> x<y

'''
        return self._binary_operation(other, '__lt__')
    #--- End: def

    def __and__(self, other):
        '''

The binary bitwise operation ``&``

x.__and__(y) <==> x&y

'''
        return self._binary_operation(other, '__and__')
    #--- End: def

    def __iand__(self, other):
        '''

The augmented bitwise assignment ``&=``

x.__iand__(y) <==> x&=y

'''
        return self._binary_operation(other, '__iand__')
    #--- End: def

    def __rand__(self, other):
        '''

The binary bitwise operation ``&`` with reflected operands

x.__rand__(y) <==> y&x

'''
        return self._binary_operation(other, '__rand__')
    #--- End: def

    def __or__(self, other):
        '''

The binary bitwise operation ``|``

x.__or__(y) <==> x|y

'''
        return self._binary_operation(other, '__or__')
    #--- End: def

    def __ior__(self, other):
        '''

The augmented bitwise assignment ``|=``

x.__ior__(y) <==> x|=y

'''
        return self._binary_operation(other, '__ior__')
    #--- End: def

    def __ror__(self, other):
        '''

The binary bitwise operation ``|`` with reflected operands

x.__ror__(y) <==> y|x

'''
        return self._binary_operation(other, '__ror__')
    #--- End: def

    def __xor__(self, other):
        '''

The binary bitwise operation ``^``

x.__xor__(y) <==> x^y

'''
        return self._binary_operation(other, '__xor__')
    #--- End: def

    def __ixor__(self, other):
        '''

The augmented bitwise assignment ``^=``

x.__ixor__(y) <==> x^=y

'''
        return self._binary_operation(other, '__ixor__')
    #--- End: def

    def __rxor__(self, other):
        '''

The binary bitwise operation ``^`` with reflected operands

x.__rxor__(y) <==> y^x

'''
        return self._binary_operation(other, '__rxor__')
    #--- End: def

    def __lshift__(self, y):
        '''

The binary bitwise operation ``<<``

x.__lshift__(y) <==> x<<y

'''
        return self._binary_operation(y, '__lshift__')
    #--- End: def

    def __ilshift__(self, y):
        '''

The augmented bitwise assignment ``<<=``

x.__ilshift__(y) <==> x<<=y

'''
        return self._binary_operation(y, '__ilshift__')
    #--- End: def

    def __rlshift__(self, y):
        '''

The binary bitwise operation ``<<`` with reflected operands

x.__rlshift__(y) <==> y<<x

'''
        return self._binary_operation(y, '__rlshift__')
    #--- End: def

    def __rshift__(self, y):
        '''

The binary bitwise operation ``>>``

x.__lshift__(y) <==> x>>y

'''
        return self._binary_operation(y, '__rshift__')
    #--- End: def

    def __irshift__(self, y):
        '''

The augmented bitwise assignment ``>>=``

x.__irshift__(y) <==> x>>=y

'''
        return self._binary_operation(y, '__irshift__')
    #--- End: def

    def __rrshift__(self, y):
        '''

The binary bitwise operation ``>>`` with reflected operands

x.__rrshift__(y) <==> y>>x

'''
        return self._binary_operation(y, '__rrshift__')
    #--- End: def

    def __abs__(self):
        '''

The unary arithmetic operation ``abs``

x.__abs__() <==> abs(x)

'''
        return self._unary_operation('__abs__')
    #--- End: def

    def __neg__(self):
        '''

The unary arithmetic operation ``-``

x.__neg__() <==> -x

'''
        return self._unary_operation('__neg__')
    #--- End: def

    def __invert__(self):
        '''

The unary bitwise operation ``~``

x.__invert__() <==> ~x

'''
        return self._unary_operation('__invert__')
    #--- End: def

    def __pos__(self):
        '''

The unary arithmetic operation ``+``

x.__pos__() <==> +x

'''
        return self._unary_operation('__pos__')
    #--- End: def

#    def _set(self, attr, value):
#        '''
#
#Set an attribute which actually belongs to the partition matrix.
#'''
#        setattr(self.partitions, attr, value)  
#    #--- End: def
#
#    def _get(self, attr):
#        '''
#
#Get an attribute which actually belongs to the partition matrix.
#
#'''
#        return getattr(self.partitions, attr)
#    #--- End: def
#
#    def _del(self, attr):
#        '''
#
#Delete an attribute which actually belongs to the partition
#array. This is used by the decorated properties,
#
#'''
#        try:
#            delattr(self.partitions, attr)
#        except AttributeError:
#            raise AttributeError("Can't delete '%s' attribute '%s'" %
#                                 (self.__class__.__name__, attr))
#    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: Units
    # ----------------------------------------------------------------
    @property
    def Units(self):
        '''

The `Units` object containing the units of the data array.

Deleting this attribute is equivalent to setting it to an undefined
units object, so this attribute is guaranteed to always exist.

**Examples**

>>> d.Units = cf.Units('m')
>>> d.Units
<CF Units: m>
>>> del d.Units
>>> d.Units
<CF Units: >

'''
        return self._Units
    #--- End: def

    @Units.setter
    def Units(self, value): self._Units = value
    @Units.deleter
    def Units(self)       : self._Units = _units_None

#    # ----------------------------------------------------------------
#    # Attribute: direction
#    # ----------------------------------------------------------------
#    @property
#    def direction(self):
#        '''
#
#
#
#**Examples**
#
#>>> d.dimensions
#['dim2', 'dim0', 'dim3']
#>>> d.direction
#{'dim0': True, 'dim1': False, 'dim2': True}
#
#>>> d.dimensions
#[]
#>>> d.direction
#True
#
#>>> d.dimensions
#[]
#>>> d.direction
#False
#
#'''
#        return self._get('direction')
#    @direction.setter
#    def direction(self, value): self._set('direction', value)
#    @direction.deleter
#    def direction(self)       : self._del('direction')

    # ----------------------------------------------------------------
    # Attribute: dtype (read only)
    # ----------------------------------------------------------------
    @property
    def data(self):
        '''

The data array object as an object identity.

**Examples**

>>> d.data is d
True

'''
        return self
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: dtype
    # ----------------------------------------------------------------
    @property
    def dtype(self):
        '''

The numpy data type of the data array.

By default this is the data type with the smallest size and smallest
scalar kind to which all data array partitions may be safely cast
without loss of information. For example, if the partitions have data
types 'int64' and 'float32' then the data array's data type will be
'float64'; or if the partitions have data types 'int64' and 'int32'
then the data array's data type will be 'int64'.

Setting the data type to a `numpy.dtype` object, or any object
convertible to a `numpy.dtype` object, will cause the data array
elements to be recast to the specified type at the time that they are
next accessed. This does not immediately change the data array
elements, so, for example, reinstating the original data type prior to
data access results in no loss of information.

Deleting the data type reinstates the default behaviour. Note that if
the data type of any partitions have been changed after `dtype` has
been set (which could occur if the data array is accessed) then the
reinstated default data type may be different to the data type prior
to `dtype` being set.

Setting the data type to None is equivalent to deleting the `dtype`
attribute.

**Examples**

>>> print d.array
[0.5 1.5 2.5]
>>> d.dtype
dtype(float64')
>>> type(d.dtype)
<type 'numpy.dtype'>

>>> print d.array
[0.5 1.5 2.5]
>>> import numpy
>>> d.dtype = numpy.dtype(int)
>>> print d.array
[0 1 2]
>>> d.dtype = bool
>>> d.dtype = int
>>> print d.array
[0 1 2]
>>> d.dtype = bool
>>> print d.array
[False True True]
>>> d.dtype = 'float64'
>>> print d.array
[ 0.  1.  1.]

'''
        datatype = self._dtype
        if datatype is None:
            return numpy_result_type(
                *(partition.subarray
                  for partition in self.partitions.matrix.flat))

        return datatype
    #--- End: def
    @dtype.setter
    def dtype(self, value):
        self._dtype = numpy_dtype(value)
    #--- End: def
    @dtype.deleter
    def dtype(self):
        self._dtype = None
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: fill_value
    # ----------------------------------------------------------------
    @property
    def fill_value(self):
        '''

The fill_value CF attribute.

If set to None then the default numpy fill value appropriate to the
data array's data type will be used.

Deleting this attribute is equivalent to setting it to None, so this
attribute is guaranteed to always exist.

**Examples**

>>> d.fill_value = 9999.0
>>> d.fill_value
9999.0
>>> del d.fill_value
>>> d.fill_value
None

'''
        return self._fill_value #_get('_FillValue')
    #--- End: def
    @fill_value.setter
    def fill_value(self, value): self._fill_value = value #self._set('_FillValue', value)
    @fill_value.deleter
    def fill_value(self)       : self._fill_value = None #self._FillValue = None

    # ----------------------------------------------------------------
    # Attribute: hardmask
    # ----------------------------------------------------------------
    @property
    def hardmask(self):
        '''

Whether the mask is hard (True) or soft (False).

When the mask is hard, masked entries of the data array can not be
unmasked by assignment, but unmasked entries may still be masked.

By default, the mask is hard.

**Examples**

>>> d.hardmask = False
>>> d.hardmask
False
>>> d.hardmask = True
>>> d.hardmask
True

'''
        return self._hardmask
    @hardmask.setter
    def hardmask(self, value):
        self._hardmask = value
    @hardmask.deleter
    def hardmask(self):
        raise AttributeError("Won't delete %s attribute 'hardmask'" %
                             self.__class__.__name__)
    #--- End: def

    @property
    def inspect(self):
        '''

Inpsect the object for debugging.

**Examples**

>>> d.inspect

'''
        name = repr(self)
        print name
        print ''.ljust(len(name), '-')

        for key, value in sorted(self.__dict__.items()):
            print '%s: %s' % (key, repr(value))
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: ismasked (read only)
    # ----------------------------------------------------------------
    @property
    def ismasked(self):
        '''

True if the data array has any masked values.

**Examples**

>>> d[0, ...] = cf.masked
>>> d.ismasked
True

'''
        conform_args = self.conform_args(revert_to_file=True, readonly=True)

        for partition in self.partitions.matrix.flat:
            array = partition.dataarray(**conform_args)

            if partition.masked:
                partition.close()
                return True
            
            partition.close()
        #--- End: for

        return False
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: isscalar (read only)
    # ----------------------------------------------------------------
    @property
    def isscalar(self):
        '''

True if the data array is a 0-d scalar array.

**Examples**

>>> d.ndim
0
>>> d.isscalar
True

>>> d.ndim >= 1
True
>>> d.isscalar
False

'''
        return not self._ndim
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: ndim (read only)
    # ----------------------------------------------------------------
    @property
    def ndim(self):
        '''

Number of dimensions in the data array.

**Examples**

>>> d.shape
(73, 96)
>>> d.ndim
2

>>> d.shape
()
>>> d.ndim
0

'''
        return self._ndim
    #--- End: def

#    # ----------------------------------------------------------------
#    # Attribute: dimensions
#    # ----------------------------------------------------------------
#    @property
#    def dimensions(self):
#        '''
#
#'''
#        return self._get('dimensions')
#    @dimensions.setter
#    def dimensions(self, value): self._set('dimensions', value)
#    @dimensions.deleter
#    def dimensions(self)       : self._del('dimensions')

    # ----------------------------------------------------------------
    # Attribute: pmdimensions (read only)
    # ----------------------------------------------------------------
    @property
    def pmdimensions(self):
        '''
'''
        return self.partitions.dimensions
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: pmndim (read only)
    # ----------------------------------------------------------------
    @property
    def pmndim(self):
        '''

Number of dimensions in the partition matrix.

>>> d.pmshape
(4, 7)
>>> d.pmndim
2

>>> d.pmshape
()
>>> d.pmndim
0

'''
        return self.partitions.ndim
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: pmsize (read only)
    # ----------------------------------------------------------------
    @property
    def pmsize(self):
        '''

Number of partitions in the partition matrix.

**Examples**

>>> d.pmshape
(4, 7)
>>> d.pmsize
28

'''
        return self.partitions.size
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: pmshape
    # ----------------------------------------------------------------
    @property
    def pmshape(self):
        '''

Tuple of the partition matrix's dimension sizes.

**Examples**

>>> d.pmshape
(4, 7)

>>> d.pmndim
0
>>> d.pmshape
()

'''
        return self.partitions.shape
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: shape (read only)
    # ----------------------------------------------------------------
    @property
    def shape(self):
        '''

Tuple of the data array's dimension sizes.

**Examples**

>>> d.shape
(73, 96)

>>> d.ndim
0
>>> d.shape
()

'''
        return self._shape
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: size (read only)
    # ----------------------------------------------------------------
    @property
    def size(self):
        '''

Number of elements in the data array.

**Examples**

>>> d.shape
(73, 96)
>>> d.size
7008

>>> d.shape
()
>>> d.size
0

'''
        return self._size
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: array (read only)
    # ----------------------------------------------------------------
    @property
    def array(self):
        '''

A numpy array copy the data array.

If the data array is stored as datetime-like objects then a numpy
array of numeric reference times will be returned. In this case, a
numpy array of datetime-like objects may be created with the `dtarray`
attribute.

.. seealso:: `dtarray`, `dtvarray`, `varray`

**Examples**

>>> a = d.array
>>> isinstance(a, numpy.ndarray)
True

'''
        conform_args = self.conform_args(revert_to_file=True)

        data_type = self.dtype

        if getattr(self, '_dtarray', False):
            del self._dtarray
            if not self._isdt:
                data_type = numpy_dtype(object)
                conform_args['func']   = rt2dt
                # Turn off data type checking and partition updating
                conform_args['dtype']  = None
                conform_args['update'] = False
        elif self._isdt:
            data_type = numpy_dtype(float)
            conform_args['func']   = dt2rt
            # Turn off data type checking and partition updating
            conform_args['dtype']  = None
            conform_args['update'] = False
        #--- End: if
        
        partitions = self.partitions

        if partitions.size == 1:
            # If there is only one partition we can speed things up
            partition = partitions.matrix.item()
            array_out = partition.dataarray(copy_regardless=True, **conform_args)
            partition.close()
            return array_out
        #--- End: if

        # Still here?
        array_out = numpy_empty(self._shape, dtype=data_type)
        masked = False

        for partition in partitions.matrix.flat:
            p_array = partition.dataarray(readonly=True, **conform_args)
# copy okect?
            if not masked and partition.masked:
                array_out = array_out.view(numpy_ma_MaskedArray)
                array_out.set_fill_value(self._fill_value)
                masked = True

            array_out[partition.indices] = p_array
            partition.close()
        #--- End: for

        if masked and self._hardmask:
            # Harden the mask of the output array
            array_out.harden_mask()

        self.dtype = data_type

        return array_out
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: dtarray (read only)
    # ----------------------------------------------------------------
    @property
    def dtarray(self):
        '''

The returned object is of the same type as is stored internally.

.. seealso:: `array`, `dvtarray`, `varray`

**Examples**

'''
        if not self.Units.isreftime:
            raise ValueError("w--343---------5")
        
        self._dtarray = True
        return self.array
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: dtvarray (read only)
    # ----------------------------------------------------------------
    @property
    def dtvarray(self):
        '''

.. seealso:: `array`, `dtarray`, `varray`

'''
        if not self.Units.isreftime:
            raise ValueError("w--343---------5")
        
        self._dtarray = True
        return self.varray
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: _array (read only)
    # ----------------------------------------------------------------
    @property
    def unsafe_array(self):
        '''

A numpy array of the data array.

It is unsafe because it might (or might not) be view to the master
data array - you just don't know. So changing the returned array in
place might have dire consequences.

It is useful because if you are 100% certain that you're not going to
change the returned array in place then this method may be much faster
than the `array` method.

Why not just use the `varray` method? Well, you might not want to
destroy the partition matrix.

The data type of the array is as returned by the `dtype` attribute.

**Examples**

>>> a = d.unsafe_array
>>> isinstance(a, numpy.ndarray)
True

'''      
        partitions = self.partitions

        if partitions.size == 1:
            # If there is only one partition we can speed things up
            conform_args = self.conform_args(revert_to_file=True)
            partition = partitions.matrix.item()
            array_out = partition.dataarray(readonly=True, **conform_args)
            partition.close()
            return array_out
        #--- End: if

        # Still here?
        return self.array
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: varray (read only)
    # ----------------------------------------------------------------
    @property
    def varray(self):
        '''

A numpy array view the data array.

Note that making changes to elements of the returned view changes the
underlying data.

If the data array is stored as datetime-like objects then a numpy
array view of numeric reference times will be returned. In this case,
a numpy array view of datetime-like objects may be created with the
`dtvarray` attribute.

.. seealso:: `array`, `dtarray`, `dtvarray`

**Examples**

>>> a = d.varray
>>> type(a)
<type 'numpy.ndarray'>
>>> a
array([0, 1, 2, 3, 4])
>>> a[0] = 999
>>> d.varray
array([999, 1, 2, 3, 4])

'''
        conform_args = self.conform_args(save=False)
        
        data_type = self.dtype

        if getattr(self, '_dtarray', False):
            del self._dtarray
            if not self._isdt:
                data_type = numpy_dtype(object)
                conform_args['func']   = rt2dt
                # Turn off data type checking and partion updating
                conform_args['dtype']  = None
#        elif getattr(self, '_rtarray', False):
#            del self._rtarray
#            if self._isdt:
#                data_type = numpy_dtype(float)
#                conform_args['func']   = dt2rt
#                # Turn off data type checking and partion updating
#                conform_args['dtype']  = None
        elif self._isdt:
            data_type = numpy_dtype(float)
            conform_args['func']   = dt2rt
            # Turn off data type checking and partion updating
            conform_args['dtype']  = None
        #--- End: if
           
        if self.partitions.size == 1:
            # If there is only one partition, then we can return a
            # view of the partition's data array without having to
            # create an empty array and then filling it up partition
            # by partition.
            partition = self.partitions.matrix.item()
            array = partition.dataarray(revert_to_file=False, **conform_args)
            # Note that there is no need to close the partition here.
            self._dtype = data_type
            return array
        #--- End: if

        # Still here?  
#        conform_args['dtype'] = None ## WHY????? for speed!!!
        shape = self._shape
        array = numpy_empty(shape, dtype=data_type)
        masked = False

        for partition in self.partitions.matrix.flat:
            data = partition.dataarray(readonly=True, **conform_args)

            if not masked and partition.masked:
                array = array.view(numpy_ma_MaskedArray)
                array.set_fill_value(self._fill_value)
                masked = True

            array[partition.indices] = data

            # Note that there is no need to close the partition here
        #--- End: for

        if masked:
            if self._hardmask:
                # Harden the mask of the output array
                array.harden_mask()
        #--- End: if

        matrix = _xxx.copy() #numpy_empty((), dtype=object)

        matrix[()] = Partition(subarray    = array,
                               location    = [(0, n) for n in shape],
                               _dimensions = self.dimensions,
                               _directions = self.directions,
                               shape       = list(shape),
                               Units       = self.Units,
                               part        = []
                               )

        self.partitions = PartitionMatrix(matrix, [])

        self._dtype = data_type

        return array
    #--- End: def

    # ----------------------------------------------------------------
    # Attribute: mask (read only)
    # ----------------------------------------------------------------
    @property
    def mask(self):
        '''

The boolean missing data mask of the data array.

The boolean mask has True where the data array has missing data and
False otherwise.

**Examples**

>>> d.shape
(12, 73, 96)
>>> m = d.mask
>>> m
<CF Data: >
>>> m.dtype
dtype('bool')
>>> m.shape
(12, 73, 96])

'''
        self.to_memory()

        mask = self.copy()

        mask_units = _units_None

        conform_args = mask.conform_args(
            save=self.save_to_disk(numpy_dtype(bool).itemsize))
        
        for partition in mask.partitions.matrix.flat:
            array = partition.dataarray(**conform_args)
            
#            if numpy_ma_is_masked(array):
            if partition.masked:
                # data is a masked array
                partition.subarray = array.mask.copy()
            else:
                # data is not a masked array
                partition.subarray = numpy_zeros(array.shape, dtype=bool)

            partition.Units = mask_units
            partition.close()
        #--- End: for

        mask.Units = mask_units
        mask.dtype = bool

        mask._hardmask = False

        return mask
    #--- End: def

    @mask.setter
    def mask(self, value):
        raise AttributeError(
            "Can't set the 'mask' attribute directly, use the 'setmask' method")
    #--- End: def
    @mask.deleter
    def mask(self):
        raise AttributeError(
"Can't delete the 'mask' attribute directly, use the 'setmask()' method")
    #--- End: def

#    # ----------------------------------------------------------------
#    # Attribute: first_datum (read only)
#    # ----------------------------------------------------------------
#    @property
#    def first_datum(self):
#        ''' 
#
#The first element of the data array.
#
#**Examples**
#
#>>> print d.array
#[[1 2 3 4]]
#>>> d.first_datum
#1
#
#>>> print d.array
#[[-- 2 3 4]]
#>>> d.first_datum
#--
#
#'''
#        array = self[(slice(0, 1),)*self._ndim].unsafe_array
#
#        if not numpy_ma_isMA(array):
#            return array.item()
#        
#        mask = array.mask
#        if mask is numpy_ma_nomask or not mask.item():
#            return array.item()
#        
#        return numpy_ma_masked
#    #--- End: def

#    # ----------------------------------------------------------------
#    # Attribute: last_datum (read only)
#    # ----------------------------------------------------------------
#    @property
#    def last_datum(self):
#        ''' 
#
#The last element of the data array.
#
#**Examples**
#
#>>> print d.array
#[[1 2 3 4]]
#>>> d.last_datum
#4
#
#>>> print d.array
#[[1 2 3 --]]
#>>> d.last_datum
#--
#
#'''
#        array = self[(slice(-1, None),)*self._ndim].unsafe_array
#
#        if not numpy_ma_isMA(array):
#            return array.item()
#        
#        mask = array.mask
#        if mask is numpy_ma_nomask or not mask.item():
#            return array.item()
#        
#        return numpy_ma_masked
#    #--- End: def

    def add_partitions(self,
                       extra_boundaries, 
                       pdim):
        '''

Add partition boundaries.

**Examples**

>>> d.add_partitions(    )

'''            
        self.partitions.add_partitions(self.dimensions,
                                       self.directions,
                                       extra_boundaries, 
                                       pdim)
    #--- End: def

    def all(self):
        '''

Test whether all data array elements evaluate to True.

Performs a logical and over the data array and returns the
result. Masked values are considered as True during computation.

**Examples**

>>> print d.array
[[0 3 0]]
>>> d.all()
False

>>> print d.array
[[1 3 --]]
>>> d.all()
True

'''
        conform_args = self.conform_args(revert_to_file=True,
                                         readonly=True)
              
        for partition in self.partitions.matrix.flat:
            array = partition.dataarray(**conform_args)
            partition.close()
            if not array.all():
                return False
        #--- End: for

        return True
    #--- End: def

    def any(self):
        '''

Test whether any data array elements evaluate to True.

Performs a logical or over the data array and returns the
result. Masked values are considered as False during computation.

**Examples**

>>> print d.array
[[0 0 0]]
>>> d.any()
False

>>> print d.array
[[-- 0 0]]
>>> d.any()
False

>>> print d.array
[[0 3 0]]
>>> d.any()
True

'''         
        conform_args = self.conform_args(revert_to_file=True,
                                         readonly=True)

        for partition in self.partitions.matrix.flat:
            array = partition.dataarray(**conform_args)
            partition.close()
            if array.any():
                return True
        #--- End: for

        return False
    #--- End: def

    def binary_mask(self):
        '''

Return a binary missing data mask of the data array.

The binary mask's data array comprises dimensionless 8-bit integers
and has 0 where the data array has missing data and 1 otherwise.

:Returns:

    out : Data
        The binary mask.

**Examples**

>>> print d.mask.array
[[ True False  True False]]
>>> b = d.binary_mask().array
>>> print b
[[0 1 0 1]]

'''
        self.to_memory()

        binary_mask = self.copy()

        conform_args = binary_mask.conform_args(
            save=self.save_to_disk(numpy_dtype('int32').itemsize),
            dtype=None)

        _units_1 = _units_1

        for partition in binary_mask.partitions.matrix.flat:            
            array = partition.dataarray(**conform_args)        
#            if numpy_ma_is_masked(array):
            if partition.masked:
                # data is a masked array
                partition.subarray = numpy_array(~array.mask, dtype='int32')
            else:
                # data is not a masked array
                partition.subarray = numpy_ones(array.shape, dtype='int32')

            partition.Units = _units_1
            partition.close()
        #--- End: for

        binary_mask.Units = _units_1

        return binary_mask
    #--- End: def

    def clip(self, a_min, a_max, units=None):
        '''

Clip (limit) the values in the data array in place.

Given an interval, values outside the interval are clipped to the
interval edges. For example, if an interval of [0, 1] is specified
then values smaller than 0 become 0 and values larger than 1 become 1.

Parameters :
 
    a_min : scalar

    a_max : scalar

    units : str or Units

:Returns: 

    None

**Examples**

'''
        if units is not None:
            # Convert the limits to the same units as the data array
            units = Units(units)

            self_units = self.Units

            if self_units != units:
                a_min = Units.conform(a_min, units, self_units)
                a_max = Units.conform(a_max, units, self_units)
        #--- End: if
            
        conform_args = self.conform_args()
            
        for partition in self.partitions.matrix.flat:            
            array = partition.dataarray(**conform_args)            
            array.clip(a_min, a_max, out=array)
            partition.close()
        #--- End: if
    #--- End: def

    def close(self):
        '''

Close all files referenced by the data array.

Note that a closed file will be automatically reopened if its contents
are subsequently required.

:Returns:

    None

**Examples**

>>> d.close()

'''    
        for partition in self.partitions.matrix.flat:
            partition.file_close()
    #--- End: def
            
    def copy(self):
        '''

Return a deep copy.

``d.copy()`` is equivalent to ``copy.deepcopy(d)``.

:Returns:

    out : 
        The deep copy.

**Examples**

>>> e = d.copy()

'''
        new = Data.__new__(Data)   ###dch changte

        new.__dict__ = self.__dict__.copy()

        new.partitions = self.partitions.copy()

        return new
    #--- End: def
            
    def cos(self):
        '''

Take the trigonometric cosine of the data array in place.

Units are accounted for in the calculation. If the units are not
equivalent to radians (such as Kelvin) then they are treated as if
they were radians. For example, the the cosine of 90 degrees_east is
0.0, as is the sine of 1.57079632 kg m-2.

The Units are changed to '1' (nondimensionsal).

:Returns:

    None

**Examples**

>>> d.Units
<CF Units: degrees_east>
>>> print d.array
[[-90 0 90 --]]
>>> d.cos()
>>> d.Units
<CF Units: 1>
>>> print d.array
[[0.0 1.0 0.0 --]]

>>> d.Units
<CF Units: m s-1>
>>> print d.array
[[1 2 3 --]]
>>> d.cos()
>>> d.Units
<CF Units: 1>
>>> print d.array
[[0.540302305868 -0.416146836547 -0.9899924966 --]]

'''
        if self.Units.equivalent(_units_radians):
            self.Units = _units_radians

        self.func(numpy_cos)

        self.override_units(_units_1)
    #--- End: def

    def asreftime(self):
        '''

Change the internal representation of data array elements from
datatime-like objects to numeric reference times.

The change is carried out in place.

If the calendar has not been set then the default CF calendar will be
used and the units' and the `calendar` attribute will be updated
accordingly.

If the internal representations are already numeric reference times
then no change occurs.

:Returns:

    None

.. seealso:: `asdatetime`


**Examples**

>>> d.asreftime()

'''
        units = self.Units

        if not self._isdt:
            if units.isreftime:
                return
            else:
                raise ValueError(
                    "Can't convert %s data to numeric reference times" %
                    repr(units))
        #--- End: if

        conform_args = self.conform_args(func=dt2rt, dtype=None)

        for partition in self.partitions.matrix.flat:
            array = partition.dataarray(**conform_args)
            p_units = partition.Units
            partition.Units = Units(p_units.units, p_units._utime.calendar)
            partition.close()
        #--- End: for

        self.Units = Units(units.units, units._utime.calendar)
        
        self._dtype = array.dtype
        self._isdt  = False
    #--- End: def

    def asdatetime(self):
        '''

Change the internal representation of data array elements from numeric
reference times to datatime-like objects.

The change is carried out in place.

If the calendar has not been set then the default CF calendar will be
used and the units' and the `calendar` attribute will be updated
accordingly.

If the internal representations are already datatime-like objects then
no change occurs.

:Returns:

    None

.. seealso:: `asreftime`

**Examples**

>>> f.asdatetime()

'''
        units = self.Units

        if self._isdt:
            return
        elif not units.isreftime:
            raise ValueError("Can't convert %s data to datetime-like objects" %
                             repr(units))

        conform_args = self.conform_args(func=rt2dt, dtype=None)

        for partition in self.partitions.matrix.flat:
            array = partition.dataarray(**conform_args)
            p_units = partition.Units
            partition.Units = Units(p_units.units, p_units._utime.calendar)
            partition.close()
        #--- End: for

        self.Units = Units(units.units, units._utime.calendar)
        
        self._dtype = array.dtype
        self._isdt  = True
    #--- End: def
        
    def _YMDhms(self, attr):
        '''
'''
        def func(array, units_in, dummy0, dummy1):
            '''

    The returned array is always independent.

    :Parameters:

        array : numpy array
        
        units_in : cf.Units
        
        dummy0 :
            Ignored.

        dummy1 :
            Ignored.

    :Returns: 

        out : numpy array

    '''
            if not self._isdt:
                array = rt2dt(array, units_in)

            return _array_getattr(array, attr)
        #--- End: def 

        if not self.Units.isreftime:
            raise ValueError(
                "Can't get %ss from data with %s" % (attr, repr(self.Units)))

        new = self.copy()
 
        new.Units  = _units_None

        conform_args = new.conform_args(func=func, dtype=None)

        for partition in new.partitions.matrix.flat:
            array = partition.dataarray(**conform_args)
            new_dtype = array.dtype
            partition.close()
        #--- End: for

        new._dtype = new_dtype
        new._isdt  = False

        return new
    #--- End: def

    @property
    def year(self):
        '''

The year of each data array element in a new data array.

Only applicable for reference time units.

.. seealso:: `month`, `day`, `hour`, `minute`, `second`

**Examples**

>>> d = cf.Data([[1.93, 5.17]], 'days since 2000-12-29')
>>> d
<CF Data: [[2000-12-30 22:19:12, 2001-01-03 04:04:48]] >
>>> d.year
<CF Data: [[2000, 2001]] >

'''
        return self._YMDhms('year')
    #--- End: def

    @property
    def month(self):
        '''

The month of each data array element in a new data array.

Only applicable for reference time units.

.. seealso:: `year`, `day`, `hour`, `minute`, `second`

**Examples**

>>> d = cf.Data([[1.93, 5.17]], 'days since 2000-12-29')
>>> d
<CF Data: [[2000-12-30 22:19:12, 2001-01-03 04:04:48]] >
>>> d.month
<CF Data: [[12, 1]] >

'''
        return self._YMDhms('month')
    #--- End: def

    @property
    def day(self):
        '''

The day of each data array element in a new data array.

Only applicable for reference time units.

.. seealso:: `year`, `month`, `hour`, `minute`, `second`

**Examples**

>>> d = cf.Data([[1.93, 5.17]], 'days since 2000-12-29')
>>> d
<CF Data: [[2000-12-30 22:19:12, 2001-01-03 04:04:48]] >
>>> d.day
<CF Data: [[30, 3]] >

'''
        return self._YMDhms('day')
    #--- End: def

    @property
    def hour(self):
        '''

The hour of each data array element in a new data array.

Only applicable for reference time units.

.. seealso:: `year`, `month`, `day`, `minute`, `second`

**Examples**

>>> d = cf.Data([[1.93, 5.17]], 'days since 2000-12-29')
>>> d
<CF Data: [[2000-12-30 22:19:12, 2001-01-03 04:04:48]] >
>>> d.hour
<CF Data: [[22, 4]] >

'''
        return self._YMDhms('hour')
    #--- End: def

    @property
    def minute(self):
        '''

The minute of each data array element in a new data array.

Only applicable for reference time units.

.. seealso:: `year`, `month`, `day`, `hour`, `second`

**Examples**

>>> d = cf.Data([[1.93, 5.17]], 'days since 2000-12-29')
>>> d
<CF Data: [[2000-12-30 22:19:12, 2001-01-03 04:04:48]] >
>>> d.minute
<CF Data: [[19, 4]] >

'''
        return self._YMDhms('minute')
    #--- End: def

    @property
    def second(self):
        '''

The second of each data array element in a new data array.

Only applicable for reference time units.

.. seealso:: `year`, `month`, `day`, `hour`, `minute`

>>> d = cf.Data([[1.93, 5.17]], 'days since 2000-12-29')
>>> d
<CF Data: [[2000-12-30 22:19:12, 2001-01-03 04:04:48]] >
>>> d.second
<CF Data: [[12, 48]] >

'''
        return self._YMDhms('second')
    #--- End: def

    @property
    def unique(self):
        '''

The unique elements of the array.

Returns a new object with the sorted unique elements in a one
dimensional array.

**Examples**

>>> d = cf.Data([[4, 2, 1], [1, 2, 3]], 'metre')
>>> d.unique
<CF Data: [1, 2, 3, 4] metre>
>>> d[1, -1] = cf.masked
>>> d.unique
<CF Data: [1, 2, 4] metre>

'''
        conform_args = self.conform_args(revert_to_file=True,
                                         readonly=True)

        u = []
        for partition in self.partitions.matrix.flat:
            array = partition.dataarray(**conform_args)
            array = numpy_unique(array)

            if partition.masked:
                # Note that compressing a masked array may result in
                # an array with zero size
                array = array.compressed()

            size = array.size
            if size > 1:
                u.extend(array)
            elif size == 1:
                u.append(array.item())

            partition.close()
        #--- End: for

        u = numpy_unique(numpy_array(u, dtype=self.dtype))

        return type(self)(u, units=self.Units, dt=self._isdt)
    #--- End: def

    def dump(self, prefix=None):
        '''

Return a string containing a full description of the instance.

:Parameters:

    id : str, optional
       Set the common prefix of component names. By default the
       instance's class name is used.

:Returns:

    out : str
        A string containing the description.

**Examples**

>>> x = d.dump()
>>> print d.dump()
>>> print d.dump(id='data1')

'''
        if prefix is None:
            prefix = self.__class__.__name__
            
        string = ['%s.shape = %s' % (prefix, self._shape)]

        if self._size == 1:
#            string.append('%s.first_datum = %s' % (prefix, self.first_datum))
            string.append('%s.first_datum = %s' % (prefix, self.datum(0)))
        else:
            string.append('%s.first_datum = %s' % (prefix, self.datum(0)))
            string.append('%s.last_datum  = %s' % (prefix, self.datum(-1)))
        #-- End: if
            
        for attr in ('fill_Value', 'Units'):
            string.append('%s.%s = %s' % (prefix, attr,
                                          repr(getattr(self, attr))))
        #--- End: for

        return '\n'.join(string)
    #--- End: def

    def ndindex(self):
        '''

Return an iterator over the N-dimensional indices of the data array.

At each iteration a tuple of indices is returned, the last dimension
is iterated over first.

:Returns:

    out : itertools.product
        An iterator over tuples of indices of the data array.

**Examples**

>>> d.shape
(2, 1, 3)
>>> for i in d.ndindex():
...     print i
...
(0, 0, 0)
(0, 0, 1)
(0, 0, 2)
(1, 0, 0)
(1, 0, 1)
(1, 0, 2)

> d.shape
()
>>> for i in d.ndindex():
...     print i
...
()

'''
#        return numpy_ndindex(self._shape)
        return itertools_product(*[xrange(0, r) for r in self._shape])  
    #return iterindices([(0, n) for n in self._shape])
    #--- End: def

    def equals(self, other, rtol=None, atol=None, traceback=False):
        '''

True if two data arrays are logically equal, False otherwise.

:Parameters:

    other : 
        The object to compare for equality.

    atol : float, optional
        The absolute tolerance for all numerical comparisons, By
        default the value returned by the `ATOL` function is used.

    rtol : float, optional
        The relative tolerance for all numerical comparisons, By
        default the value returned by the `RTOL` function is used.

    traceback : bool, optional
        If True then print a traceback highlighting where the two
        instances differ.

:Returns: 

    out : bool
        Whether or not the two instances are equal.

**Examples**

>>> d.equals(d)
True
>>> d.equals(d + 1)
False

'''
        # Check each instance's id
        if self is other:
            return True
 
        # Check that each instance is the same type
        if self.__class__ != other.__class__:
            return False

        # Check that each instance has the same shape
        if self._shape != other._shape:
            if traceback:
                print("%s: Different shape: %s, %s" %
                      (self.__class__.__name__, self._shape, other._shape))
            return False
        #--- End: if

        # Check that each instance has the same units
        if self.Units != other.Units:
            if traceback:
                print("%s: Different Units: %s, %s" %
                      (self.__class__.__name__,
                       repr(self.Units), repr(other.Units)))
            return False
        #--- End: if

        # Check that each instance has the same fill value
        if self._fill_value != other._fill_value:
            if traceback:
                print("%s: Different array fill_value: %s, %s" %
                      (self.__class__.__name__,
                       self._fill_value, other._fill_value))
            return False
        #--- End: if

        # ------------------------------------------------------------
        # Check that each instance has the array values
        # ------------------------------------------------------------
        # Set default tolerances
        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()        

        conform_args = self.conform_args(readonly=True, revert_to_file=True)

        other.to_memory()

        for partition in self.partitions.matrix.flat:
            array0 = partition.dataarray(**conform_args)
            partition.close()
            array1 = other[partition.indices].varray
            if not _allclose(array0, array1, rtol=rtol, atol=atol):
                if traceback:
                    print("%s: Different data array values" %
                          self.__class__.__name__)
                return False
        #--- End: for

        # ------------------------------------------------------------
        # Still here? Then the two instances are equal
        # ------------------------------------------------------------
        return True            
    #--- End: def

    def expand_dims(self, position=0, axis=None, direction=True):
        '''

Expand the shape of the data array in place.

Insert a new size 1 axis, corresponding to a given position in the
data array shape.

Not to be confused with the `expand_partition_dims` method.

:Parameters:

    position : int, optional
        Position (amongst axes) where new axis is to be inserted. By
        default, insert at position 0.

    axis : str, optional
        
    direction : bool, optional

:Returns:

    None

.. seealso:: `~cf.Data.flip`, `~cf.Data.squeeze`, `~cf.Data.transpose`

**Examples**

>>> d.expand_dims(1)
>>> d.expand_dims(2, direction=True)
>>> d.expand_dims(-1, axis='dim3')
>>> d.expand_dims(0, direction=False, axis=None)

'''
        data_axes = self.dimensions

        if axis is None:
            axis = self.new_axis_identifier()
        elif axis in data_axes:
            raise ValueError("Can't insert a duplicate axis: %s" % repr(axis))
        else:
            existing_axes = self.all_axis_names()
            if axis in existing_axes:
                self.change_axis_names(
                    {axis: self.new_axis_identifier(existing_axes)})
        #--- End: if

        # Expand directions
        if not self._ndim:
            # Data aray is a scalar
            self.directions = {axis: bool(direction)}
        else:
            # Data array has one or more axes
            directions = self.directions.copy()
            directions[axis] = bool(direction)
            self.directions = directions

        # Expand axes
        data_axes = data_axes[:]
        data_axes.insert(position, axis)
        self.dimensions = data_axes

        # Expand shape and ndim
        self._ndim += 1
        shape = list(self._shape)
        shape.insert(position, 1)   
        self._shape = tuple(shape)

        # Expand the locations map
        location = (0, 1)
        for partition in self.partitions.matrix.flat:
            partition.location = partition.location[:]
            partition.shape    = partition.shape[:]
            
            partition.location.insert(position, location)
            partition.shape.insert(position, 1)
    #--- End: def

    def flat(self, ignore_masked=True):
        '''

Return a flat iterator over elements of the data array.

:Parameters:

    ignore_masked : bool, optional
        If False then masked and unmasked elements will be returned. By
        default only unmasked elements are returned

:Returns:

    out : generator
        An iterator over elements of the data array.

**Examples**

>>> print d.array
[[1 -- 3]]
>>> for x in d.flat():
...     print x
...
1
3

>>> for x in d.flat(False):
...     print x
...
1
--
3

'''
        self.to_memory()

        mask = self.mask

        if ignore_masked:
            for index in self.ndindex():
                if not mask[index]:
                    yield self[index].unsafe_array.item()
        else:
            for index in self.ndindex():
                if not mask[index]:
                    yield self[index].unsafe_array.item()
                else:
                    yield cf_masked
    #--- End: def

    def override_units(self, units):
        '''

Override the data array units in place.

Not to be confused with setting the `Units` attribute to units which
are equivalent to the original units. This is different because in
this case the new units need not be equivalent to the original ones
and the data array elements will not be changed to reflect the new
units.

:Parameters:

    units : str or Units
        The new units for the data array.

:Returns:

    None

**Examples**

>>> d = cf.Data(1012.0, 'hPa')
>>> d.override_units('km')
>>> d.Units
<CF Units: km>
>>> d.datum(0)
1012.0
>>> d.override_units(cf.Units('watts'))
>>> d.Units
<CF Units: watts>
>>> d.datum(0)
1012.0

'''
#        if isinstance(units, basestring):
        units = Units(units)
 
        conform_args = self.conform_args()

        for partition in self.partitions.matrix.flat:            
            if partition.Units == units:
                # No need to create the data array if the sub-array
                # units are the same as the master data array units
                partition.Units = units #.copy()
                continue
            #--- End: if

            partition.dataarray(**conform_args)
            partition.Units = units #.copy()
            partition.close()
        #--- End: for

        self.Units = units
    #--- End: def

    def override_calendar(self, calendar):
        '''

Override the data array units in place.

Not to be confused with setting the `Units` attribute to units which
are equivalent to the original units. This is different because in
this case the new units need not be equivalent to the original ones
and the data array elements will not be changed to reflect the new
units.

:Parameters:

    calendar : str

:Returns:

    None

**Examples**

'''
        if not units.isreftime:
            raise ValueError("asdasdasds hhhhhhhh")

        for partition in self.partitions.matrix.flat:            
            partition.Units = Units(partition.Units._units, calendar)
            partition.close()
        #--- End: for

        self.Units = Units(self.Units._units, calendar)
    #--- End: def

    def to_disk(self):
        '''

Store the data array on disk in place.

There is no change to partition's whose sub-arrays are already on
disk.

:Returns:

    None

**Examples**

>>> d.to_disk()

'''
        conform_args = self.conform_args(save=True, revert_to_file=False)

        for partition in self.partitions.matrix.flat:
            if partition.in_memory:
                partition.dataarray(**conform_args)
                partition.close()
    #--- End: def

    def to_memory(self, regardless=False):
        '''

Store each partition's data in memory in place if the master array is
smaller than the chunk size.

There is no change to partitions with data that are already in memory.

:Parameters:
    
    regardless : bool, optional
        If True then store all partitions' data in memory regardless
        of the size of the master array. By default only store all
        partitions' data in memory if the master array is smaller than
        the chunk size.

:Returns:

    None

**Examples**

>>> pa.to_memory()
>>> pa.to_memory(True)

'''
        if not self.save_to_disk(self.dtype.itemsize) or regardless:
            conform_args = self.conform_args(save=False, revert_to_file=False)
            for partition in self.partitions.matrix.flat:
                if partition.on_disk:
                    partition.dataarray(**conform_args)
                    partition.close()
    #--- End: def

    @property
    def in_memory(self):
        '''

:Returns:

**Examples**

>>> d.in_memory

'''
        for partition in self.partitions.matrix.flat:
            if not partition.in_memory:
                return False
        #--- End: for

        return True
    #--- End: def

    def equivalent(self, other, rtol=None, atol=None, traceback=False, 
                   transpose=None, squeeze=True, use_directions=False):
        '''

True if and only if two data arrays are logically equivalent.

Equivalence is defined as both data arrays being the same after
accounting for different but equivalent:

* units

* size one dimensions (if *squeeze* is True), 

* dimension directions (if *use_directions* is True) 

* dimension orders (if *transpose* is set to a dictionary).

:Parameters:

    other : Data

    transpose : dict, optional

    squeeze : bool, optional

    atol : float, optional
        The absolute tolerance for all numerical comparisons, By
        default the value returned by the `ATOL` function is used.

    rtol : float, optional
        The relative tolerance for all numerical comparisons, By
        default the value returned by the `RTOL` function is used.

    use_directions : bool, optional

:Returns:

    out : bool
        Whether or not the two variables are equivalent.

'''     
        if not self.Units.equivalent(other.Units):
            return 
       
        data1 = other.copy()
        
        if squeeze:
            data0 = self.copy()
            data0.squeeze()
            data1.squeeze()
        else:
            data0 = self

        if data0._ndim != data1._ndim:
            return
 
        data0_dimensions = data0.dimensions
        data1_dimensions = data1.dimensions        

        if transpose:
            try:
                axes = [data0_dimensions.index(transpose[dim1])
                        for dim1 in data1_dimensions]
            except ValueError:
                # transpose dictionary is insufficient
                if traceback:
                    print("%s: Can't transpose" % self.__class__.__name__)
                return False

            data1.transpose(axes)
        #--- End: if

        if data0._shape != data1._shape:
            return
 
        if use_directions and self._size > 1:            
            data0_directions = data0.directions
            data1_directions = data1.directions
            
            axes = [dim1
                    for dim1, dim0 in izip(data0_dimensions, data1_dimensions)
                    if data1_directions[dim1] != data0_directions[dim0]]
            data1.flip(axes)
        #--- End: if
        
        data1.Units       = data0.Units #.copy()
        data1._fill_value = data0._fill_value

        return data0.equals(data1, rtol=rtol, atol=atol, traceback=False)
    #--- End: def

    def change_axis_names(self, dim_name_map):
        '''

Change the axis names.

The dimension names are arbitrary, so mapping them to another
arbitrary collection does not change the data array values, units,
dimension directions nor dimension order.

**Examples**

>>> d.dimensions
['dim0', 'dim1', 'dim2']
>>> dim_name_map
{'dim0': 'dim1',
 'dim1': 'dim0',
 'dim2': 'dim2',
 'dim3': 'dim3'}
>>> d.change_axis_names(dim_name_map)
>>> d.dimensions
['dim1', 'dim0', 'dim2']

'''       
        # Check for a null dimension name mapping
        # (e.g. 'dim0'->'dim0', etc.) and return if there is nothing
        # to change. I think that this is worth it, on the grounds
        # that this is often be the case
#        map_is_null = True
#        for dim0, dim1 in dim_name_map.iteritems():
#            if dim0 != dim1:
#                map_is_null = False
#                break
#        #--- End: for       
#        if map_is_null:
#            return

        # Find any partition dimension names which are not master
        # array dimension names. If there are any, then update
        # dim_name_map.
        d = self.all_axis_names().difference(dim_name_map)
      
        if d:
            dim_name_map = dim_name_map.copy()
            existing_axes = set(dim_name_map.itervalues())            
            for dim in d:
                if dim in existing_axes:
                    new_dim = self.new_axis_identifier(existing_axes)
                else:
                    new_dim = dim

                existing_axes.add(new_dim)
                dim_name_map[dim] = new_dim 
        #--- End: if

#        if all([dim0==dim1 for dim0, dim1 in dim_name_map.iteritems()]):
#            return
        
        # Dimensions
        self.dimensions = [dim_name_map[dim] for dim in self.dimensions]

        # Directions
        if self._ndim:
            directions = {}
            for dim, dir in self.directions.iteritems():
                directions[dim_name_map[dim]] = dir

            self.directions = directions    
        #--- End: if           

        # Dimensions and directions of the partitions in the partition
        # matrix
        self.partitions.change_axis_names(dim_name_map)
    #--- End: def

    def override_directions(self, new_directions=None, scalar_dir=True):
        '''

Override the data array directions in place.

:Parameters:

    new_directions : dict, optional
    
    scalar_dir : bool, optional
    
:Returns:

    None

**Examples**


'''        
        orginal_directions = self.directions

        if not self._ndim:
            # Scalar data array
            if orginal_directions == scalar_dir:
                return

            self.directions = scalar_dir
            partition = self.partitions.matrix.item()
            if not partition._dimensions:
                # Scalar sub-array
                partition._directions = scalar_dir

            return
        #--- End: if

        # Still here?
        if orginal_directions == new_directions or new_directions is None:
            return

        directions = {}
        for dim, dir in orginal_directions.iteritems():
            directions[dim] = new_directions.get(dim, dir)

        self.directions = directions

        new_directions = directions

        # Directions of the partitions in the partition matrix
        for partition in self.partitions.matrix.flat:
            if not partition._dimensions:
                # Scalar sub-array
                continue

#            p_directions = partition._directions

            directions = {}
            for p_dim, p_dir in partition._directions.iteritems():
                orginal_dir = orginal_directions.get(p_dim, None)
                if p_dir is orginal_dir:
                    directions[p_dim] = new_directions[p_dim]
                elif orginal_dir is None:
                    directions[p_dim] = p_dir
                else:
                    directions[p_dim] = not new_directions[p_dim]
            #--- End: for
            partition._directions = directions
        #--- End: for
    #--- End: def

    def conform_args(self, save=None, **kwargs):
        '''
        
Return a dictionary of arguments for the `cf.Partition.dataarray`
method.

The values are inferred from the state of the Data object and any
keyword arguments.

:Parameters:

    save : bool, optional
        Set the 'save' key. By default it is set automatically
        depending on the size of te master array, the chunk size and
        the available memory.
    
    kwargs :
    

:Returns:

    out : dict

**Examples**


'''
        dtype = self.dtype

        conform_args = {'dimensions': self.dimensions,
                        'directions': self.directions,
                        'units'     : self.Units,
                        'hardmask'  : self._hardmask,
                        'dtype'     : dtype}

        if save is None:
            save = self.save_to_disk(dtype.itemsize)
            
        conform_args['save'] = save
            
        if kwargs:
            conform_args.update(kwargs)

        return conform_args
    #--- End: def

#    def equivalent(self, other, rtol=None, atol=None):
#        '''
#
#Equivelence is defined as both instances having the same data arrays
#after accounting for different but equivalent units, different
#dimension directions (unless `use_directions` is False) and different
#dimension orders. Equivalent data arrays must have equal numbers of
#dimensions, unless both have one or fewer dimensions.
#
#:Parameters:
#
#    other : Data
#
#    atol : float, optional
#        The absolute tolerance for all numerical comparisons, By
#        default the value returned by the `ATOL` function is used.
#
#    rtol : float, optional
#        The relative tolerance for all numerical comparisons, By
#        default the value returned by the `RTOL` function is used.
#
#:Returns:
#
#    out : bool
#        Whether or not the two variables are equivalent.
#
#'''
#        if self.size != other.size:
#            return False
#        
#        if not self.Units.equivalent(other.Units):
#            return False
#
#        data0 = self.copy()
#        data1 = other.copy()
#
#        data0.squeeze()
#        data1.squeeze()
#        
#        data1.Units = data0.Units.copy()
#
#        data1._FillValue = data0._FillValue
#
#        return data0.equals(data1, rtol=rtol, atol=atol, traceback=False)
#    #--- End: def

    def expand_partition_dims(self, pdim):
        '''

Expand the shape of the partition matrix in place.

Insert a new size partition matrix dimension at position 0. The new
dimension must be a dimension of the data array.

Not to be confused with the `expand_dims` method.

:Parameters:

    axis : str
        The name of the new partition dimension.

:Returns:

    None

**Examples**

>>> d.pmdimensions
['dim0', 'dim1']
>>> d.expand_partition_dims('dim2')
>>> d.pmdimensions
['dim2', 'dim0', 'dim1']

'''
        self.partitions.expand_dims(pdim)
    #--- End: def

    def all_axis_names(self):
        '''

Return a set of all the dimension names in use by the data array.

Note that the output set includes dimensions of individual partitions
which are not dimensions of the master data array.

:Returns:

    out : set
        The axis names.

**Examples**

>>> d.dimensions
['dim1', 'dim0']
>>> d.partitions.info('dimensions')
[['dim0', 'dim0'],
 ['dim1', 'dim0', 'dim2']]
>>> d.all_axis_names()
set(['dim2', dim0', 'dim1'])

'''
        axes_names = []
        for partition in self.partitions.matrix.flat:
            axes_names.extend(partition._dimensions)
#            for dim in partition._dimensions:
#                out[dim] = dim
            
        return set(axes_names)

#        out = self.dimensions[:]
#
#        for partition in self.partitions.matrix.flat:
#            out.extend(partition._dimensions)
#            
 #       return set(out)

#        out = set(self.dimensions)
#
#        for partition in self.partitions.matrix.flat:
#            out.update(partition._dimensions)
#            
#        return out
    #--- End: def

    def new_axis_identifier(self, existing_axes=None):
        '''

Return an axis name not being used by the data array.

The returned axis name will also not be referenced by partitions of
the partition matrix.

:Parameters:

    existing_axes : sequence

:Returns:

    out : str
        The new axis name.

**Examples**

>>> d.all_axis_names()   
['dim1', 'dim0']
>>> d.new_axis_identifier()
'dim2'

>>> d.all_axis_names()   
['dim1', 'dim0', 'dim3']
>>> d.new_axis_identifier()
'dim4'

>>> d.all_axis_names()   
['dim5', 'dim6', 'dim7']
>>> d.new_axis_identifier()
'dim3'

'''
        if existing_axes is None:
            existing_axes = self.all_axis_names()           

        n = len(existing_axes)
        dim = 'dim%d' % n
        while dim in existing_axes:
            n += 1
            dim = 'dim%d' % n
        #--- End: while

        return dim
    #--- End: def

    def partition_boundaries(self):
        '''

Return the partition boundaries for each partition matrix dimension.

:Returns:

    out : dict

**Examples**

'''            
        return self.partitions.partition_boundaries() #self.dimensions)
    #--- End: def

    @property
    def max(self):
        '''

The maximum of the array.

Returns a new object with the maximum in a scalar array.

**Examples**

>>> d = cf.Data([[4, 5, 6], [1, 2, 3]], 'metre')
>>> d.max
<CF Data: 6 metre>
>> d.max.datum()
6

'''
        conform_args = self.conform_args(revert_to_file=True,
                                         readonly=True)
        
        flat = self.partitions.matrix.flat

        partition = flat.next()
        array = partition.dataarray(**conform_args)
        m = numpy_amax(array)
        partition.close()

        for partition in flat:
            array = partition.dataarray(**conform_args)
            m = max(m, numpy_amax(array))
            partition.close()
        #--- End: for

        if m is numpy_ma_masked:
            m = numpy_ma_array(0, mask=True, dtype=array.dtype)

        return type(self)(m, self.Units)
    #--- End: def

    @property
    def min(self):
        '''

The minimum of the array.

Returns a new object with the minumum in a scalar array.

**Examples**

>>> d = cf.Data([[4, 5, 6], [1, 2, 3]], 'metre')
>>> d.min
<CF Data: 1 metre>
>> d.min.datum()
1

'''
        conform_args = self.conform_args(revert_to_file=True,
                                         readonly=True)

        flat = self.partitions.matrix.flat
        
        partition = flat.next()
        array = partition.dataarray(**conform_args)
        m = numpy_amin(array)
        partition.close()

        for partition in flat:
            array = partition.dataarray(**conform_args)
            m = min(m, numpy_amin(array))
            partition.close()
        #--- End: for

        if m is numpy_ma_masked:
            m = numpy_ma_array(0, mask=True, dtype=array.dtype)

        return type(self)(m, self.Units)
    #--- End: def
        
    def datum(self, *index):
        '''

Return an element of the data array as a standard Python scalar.

The first and last elements are always returned with ``d.datum(0)``
and ``d.datum(-1)`` respectively, even if the data array is a scalar
array or has two or more dimensions.

The returned object is of the same type as is stored internally.

:Parameters:

    index : *optional*
        Specify which element to return. When no positional arguments
        are provided, the method only works for data arrays with one
        element (but any number of dimensions), and the single element
        is returned. If positional arguments are given then they must
        be one of the following:

          * An integer. This argument is interpreted as a flat index
            into the array, specifying which element to copy and
            return.
         
              Example: If the data aray shape is ``(2, 3, 6)`` then:
                * ``d.datum(0)`` is equivalent to ``d.datum(0, 0, 0)``.
                * ``d.datum(-1)`` is equivalent to ``d.datum(1, 2, 5)``.
                * ``d.datum(16)`` is equivalent to ``d.datum(0, 2, 4)``.

            If *index* is ``0`` or ``-1`` then the first or last data
            array element respecitively will be returned, even if the
            data array is a scalar array or has two or more
            dimensions.
        ..
         
          * Two or more integers. These arguments are interpreted as a
            multidimensionsal index to the array. There must be the
            same number of integers as data array dimensions.
        ..
         
          * A tuple of integers. This argument is interpreted as a
            multidimensionsal index to the array. There must be the
            same number of integers as data array dimensions.
         
              Example: ``d.datum((0, 2, 4))`` is equivalent to
              ``d.datum(0, 2, 4)``; and ``d.datum(())`` is equivalent
              to ``d.datum()``.

:Returns:

    out :
        A copy of the specified element of the array as a suitable
        Python scalar.

.. seealso:: `array`, `dtarray`

**Examples**

>>> d = cf.Data(2)
>>> d.datum()
2
>>> 2 == d.datum(0) == d.datum(-1) == d.datum(())
True

>>> d = cf.Data([[2]])
>>> 2 == d.datum() == d.datum(0) == d.datum(-1)
True
>>> 2 == d.datum(0, 0) == d.datum((-1, -1)) == d.datum(-1, 0)
True

>>> d = cf.Data([[4, 5, 6], [1, 2, 3]], 'metre')
>>> d[0, 1] = cf.masked
>>> print d
[[4 -- 6]
 [1 2 3]]
>>> d.datum(0)
4
>>> d.datum(-1)
3
>>> d.datum(1)
masked
>>> d.datum(4)
2
>>> d.datum(-2)
2
>>> d.datum(0, 0)
4
>>> d.datum(-2, -1)
6
>>> d.datum(1, 2)
3
>>> d.datum((0, 2))
6


'''
        if index:
            n_index = len(index)
            if n_index == 1:
                index = index[0]
                if index == 0:
                    # This also works for scalar arrays
                    index = (slice(0, 1),) * self._ndim
                elif index == -1:
                    # This also works for scalar arrays
                    index = (slice(-1, None),) * self._ndim
                elif isinstance(index, (int, long)):
                    if index < 0:
                        index += self._size

                    index = numpy_unravel_index(index, self._shape)
                elif len(index) != self._ndim:
                    raise ValueError(
                        "Incorrect number of indices for %s array" %
                        self.__class__.__name__)
                #--- End: if
            elif n_index != self._ndim:
                raise ValueError(
                    "Incorrect number of indices for %s array" %
                    self.__class__.__name__)

            if self._isdt:
                array = self[index].dtarray
            else:
                array = self[index].array
        elif self._size == 1:
            if self._isdt:
                array = self.dtarray            
            else:
                array = self.array
        else:
            raise ValueError(
                "Can only convert a %s array of size 1 to a Python scalar" %
                self.__class__.__name__)

        if not numpy_ma_isMA(array):
            return array.item()
        
        mask = array.mask
        if mask is numpy_ma_nomask or not mask.item():
            return array.item()
        
        return cf_masked
    #--- End: def
  
    def flip(self, axes=None):
        '''

Flip (reverse the direction) axes of the data array in place.

:Parameters:

    axes : (sequence of) int or str, optional
        Select the axes.  By default all axes are flipped. The *axes*
        argument may be one, or a sequence, of:

          * An internal axis identifier. Selects this axis.
        ..

          * An integer. Selects the axis coresponding to the given
            position in the list of axes of the data array.

        No axes are flipped if *axes* is an empty sequence.

:Returns:

    out : list of ints
        The axes which were flipped, in arbitrary dimensions.

.. seealso:: `~cf.Data.expand_dims`, `~cf.Data.squeeze`,
             `~cf.Data.transpose`

**Examples**

>>> d.flip()
>>> d.flip(1)
>>> d.flip([0, 1])

>>> e = d[::-1, :, ::-1]
>>> d.flip([2, 0]).equals(e)
True

'''
        # Scalar data. Allow axis in (None, [], (), etc.) to reverse
        # the direction
        if not self._ndim:
            if not (axes is None or (not axes and axes != 0)):
                raise ValueError(
                    "Can't reverse: Invalid axes for this array: %s" % (axes,))
            self.directions = not self.directions
            return []
        #--- End: if

        if axes is None:
            axes = range(self._ndim)
        else:
            axes = self._parse_axes(axes, 'flip')

        directions = self.directions.copy() # Never update directions in-place
        data_axes  = self.dimensions       
        partitions = self.partitions
        pm_axes = partitions.dimensions

        flip_partition_matrix = False
        if pm_axes:
            indices = [slice(None)] * partitions.ndim

        for i in axes:
            axis = data_axes[i]
            directions[axis] = not directions[axis]                          

            if axis in pm_axes:
                indices[pm_axes.index(axis)] = slice(None, None, -1)
                flip_partition_matrix = True
        #--- End: for
                
        self.directions = directions

        if flip_partition_matrix:
            partitions = partitions[tuple(indices)]  # dch could do inaplce?
            partitions.set_location_map(data_axes)
            self.partitions = partitions
        #--- End: if

        return axes
    #--- End: def

    def save_to_disk(self, itemsize=None):
        '''

Return True if the master array is large enough to be saved to disk.

:Parameters:

    itemsize : int, optional
        The number of bytes per word of the master data array. By
        default it taken from the array's data type.

:Returns:

    out : bool

**Examples**

>>> print d.save_to_disk()
True

>>> print d.save_to_disk(8)
False

'''
        if not itemsize:
            try:
                itemsize = self.dtype.itemsize
            except AttributeError:
                raise ValueError(
                    "save_to_disk: Must set itemsize if there is no dtype")
        #--- End: if

        # ------------------------------------------------------------
        # Note that self._size*(itemsize+1) is the array size in bytes
        # including space for a full boolean mask
        # ------------------------------------------------------------

        return self._size*(itemsize+1) > FREE_MEMORY() - FM_THRESHOLD()
    #--- End: def

    def setdata(self, x, y, condition=True):
        '''

Set data array elements depending on a condition.

Elements are set differently depending on where the condition is True
or False. Two assignment values are given. From one of them, the data
array is set where the condition is True and where the condition is
False, the data array is set from the other.

Each assignment value may either contain a single datum, or is an
array-like object which is broadcastable shape of the data array.

**Missing data**

The treatment of missing data elements depends on the value of the
`hardmask` attribute. If it is True then masked elements will not
unmasked, otherwise masked elements may be set to any value.

In either case, unmasked elements may be set to any value (including
missing data).

Unmasked elements may be set to missing data by assignment to the
`cf.masked` constant or by assignment to a value which contains masked
elements.

:Parameters:

    x, y :

        Specify the assignment value. Where the condition defined by
        the *arg* and *kwargs* arguments is True, set the data array
        from *x* and where the condition is False, set the data array
        from *y*. Arguments *x* and *y* are each one of:

          * ``None``. The appropriate elements of the data array are
            unchanged.
        ..

          * Any object which is broadcastable to the data array's
            shape. The appropriate elements of the data array are set
            to the corresponding values from the object broadcast to
            the data array shape.
        ..
          
    condition : *optional*

        Define the condition which determines how to set the data
        array. The condition is any object which is broadcastable to
        the data array shape. The condition is True where the object
        broadcast to the data array shape is True. If *condition* is
        unset then it defaults to a condition of True everywhere.

:Returns:

    None

.. seealso:: `cf.masked`, `hardmask`, `__setitem__`

**Examples**

Set data array values to 15 everywhere:

>>> d.setdata(15, None)

Set all negative data array values to zero and leave all other
elements unchanged:

>>> d.setdata(0, None, d<0)

Multiply all positive elements by -1 and set other elements to 3.14:

>>> d.setdata(-d, 3.14, d>0)

Set all values less than 280 and greater than 290 to missing data:

>>> d.setdata(cf.masked, None, (d < 280) | (d > 290))

'''     
        def _broadcast(a, shape):
            '''

    Broadcast an array to a given shape.
    
    It is assumed that ``len(array.shape) <= len(shape)`` and that the
    array is broadcastable to the shape by the normal numpy
    boradcasting rules, but neither of these things are checked.
    
    For example, ``d[...] = _broadcast(e, d.shape)`` gives the same
    result as ``d[...] = e``
    
    :Parameters:
    
        a : numpy array-like
    
        shape : tuple
    
    :Returns:
    
        out : numpy array

    '''
            a_shape = numpy_shape(a)
            if a_shape == shape:
                return a

            tile = [(m if n == 1 else 1)
                    for n, m in zip(a_shape[::-1], shape[::-1])]
            tile = shape[0:len(shape)-len(a_shape)] + tuple(tile[::-1])

            return numpy_tile(a, tile)
        #--- End: def

        def _slice_to_partition(data, indices):           
            ''' 
            
     Return a numpy array for the part of the `cf.Data` object which
     spans the given indices.

    :Parameters:
    
        data : cf.Data
    
        indices : tuple
           
    :Returns:
    
        out : numpy array

    '''
            indices2 = [(slice(0, 1) if n == 1 else i)
                        for n, i in zip(data.shape[::-1], indices[::-1])]

            return data[tuple(indices2)[::-1]].unsafe_array
        #--- End: def

        def _is_broadcastable(self, data):
            '''

    Check that the input `cf.Data` object broadcastable to *self* and
    return the data, as a python scalar if possible.
    
    This function updates following lists: do_not_broadcast, is_scalar

    :Parameters:
    
        data : cf.Data
    
    :Returns:
    
        out : cf.Data or scalar
    
    '''
            shape      = self._shape
            data_shape = data._shape
            data_size  = data._size

            if data_shape == shape:
                do_not_broadcast.append(True)
                is_scalar.append(False)

            elif data_size == 1:
                do_not_broadcast.append(False)
                is_scalar.append(True)
                # Replace data with its scalar value
#                data = data.first_datum
                data = data.datum(0)
                
            elif data._ndim <= self._ndim and data_size < self._size:
                do_not_broadcast.append(False)
                is_scalar.append(False)
                for n, m in zip(data_shape[::-1], shape[::-1]):
                    if n != m and n != 1:
                        raise ValueError(
"Can't assign by where: Can't broadcast data with shape %s to shape %s" %
(data_shape, shape))             
            else:
                raise ValueError(
"Can't assign by where: Can't broadcast data with shape %s to shape %s" %
(data_shape, shape))
            #--- End: if

            return data
        #--- End: def
    
        if x is None and y is None:
            return

        do_not_broadcast = []
        is_scalar        = []

        # ------------------------------------------------------------
        # Make sure that the condition is a cf.Data object
        # ------------------------------------------------------------
        if not isinstance(condition, self.__class__):
            condition = type(self)(condition)

        # Check that the condition is broadcastable
        condition = _is_broadcastable(self, condition)

        # ------------------------------------------------------------
        # Parse x and y so that each is one of a) None, b) a scalar or
        # c) a data array with the same shape as the master array
        # ------------------------------------------------------------
        xy = []
        for value in (x, y):
            if value is None or value is cf_masked:
                do_not_broadcast.append(False)
                is_scalar.append(True)

            else:
                # Make sure that the value is a cf.Data object and has
                # compatible units
                if not isinstance(value, self.__class__):
                    value = type(self)(value)
                else:
                    if value.Units.equivalent(self.Units):
                        if not value.Units.equals(self.Units):                    
                            value = value.copy()                    
                            value.Units = self.Units
                    elif value.Units:
                        raise ValueError(" Some bad units %s, %s" %
                                         (repr(self.Units), repr(value.Units)))
                #--- End: if

                # Check that the value is broadcastable
                value = _is_broadcastable(self, value)
            #--- End: if

            xy.append(value)
        #--- End: for
        x, y = xy
        c_is_scalar, x_is_scalar, y_is_scalar = is_scalar

        #-------------------------------------------------------------
        # Try some shortcuts if the condition is a scalar
        #-------------------------------------------------------------
        if c_is_scalar :
            if condition:
                if x is not None:
                    self[...] = x

                return  
            else:
                if y is not None:
                    self[...] = y
                    
                return
        #--- End: if

        # Still here?
        broadcast = not any(do_not_broadcast)

        hardmask = self._hardmask
        conform_args = self.conform_args(readonly=True)
        
        for partition in self.partitions.matrix.flat:
            array = partition.dataarray(**conform_args)

            # --------------------------------------------------------
            # Find the master array indices for this partition
            # --------------------------------------------------------
            shape = array.shape
            indices = partition.indices

            # --------------------------------------------------------
            # Find the condition for this partition
            # --------------------------------------------------------
            if c_is_scalar:
                c = condition
            else:
                c = _slice_to_partition(condition, indices)

            # --------------------------------------------------------
            # Find value to use where condition is True for this
            # partition
            # --------------------------------------------------------
            if x_is_scalar:  
                if x is None:
                    # Use self
                    T = array
                    T_masked = partition.masked
                else:
                    T = x
                    T_masked = x is cf_masked
            else:
                T = _slice_to_partition(x, indices)
                T_masked = numpy_ma_isMA(T) and numpy_ma_is_masked(T)
            #--- End: if

            # --------------------------------------------------------
            # Find value to use where condition is False for this
            # partition
            # --------------------------------------------------------
            if y_is_scalar:  
                if y is None:
                    # Use self
                    F = array
                    F_masked = partition.masked
                else:
                    F = y
                    F_masked = y is cf_masked
            else:
                F = _slice_to_partition(y, indices)
                F_masked = numpy_ma_isMA(F) and numpy_ma_is_masked(F)
            #--- End: if

            # --------------------------------------------------------
            # Make sure that at least one of the arrays is the same
            # shape as the partition
            # --------------------------------------------------------
            if broadcast:
                if x is cf_masked and y is cf_masked:
                    c = _broadcast(c, shape)
                else:
                    max_sizes = max((numpy_size(c), numpy_size(T), numpy_size(F)))
                    if numpy_size(c) == max_sizes:
                        c = _broadcast(c, shape)
                    elif numpy_size(T) == max_sizes:
                        T = _broadcast(T, shape)
                    else:
                        F = _broadcast(F, shape)
            #--- End: if

            # --------------------------------------------------------
            # Create a numpy array which takes vales from T where c
            # is True and from F where c is False
            # --------------------------------------------------------
            if T_masked or F_masked:
                # T and/or F have missing data
                new = numpy_ma_where(c, T, F)
                if partition.masked:
                    if hardmask:
                        # The original partition has missing data and
                        # a hardmask, so apply the original
                        # partition's mask to the new array.
                        new.mask |= array.mask
                    elif not numpy_ma_is_masked(new):
                        # The original partition has missing data and
                        # a softmask and the new array doesn't have
                        # missing data, so turn the new array into an
                        # unmasked array.
                        new = new.data[...]  

                elif not numpy_ma_is_masked(new):
                    # The original partition doesn't have missing data
                    # and neither does the new array
                    new = new.data[...]
           
            else:
                # Neither T nor F have missing data
                new = numpy_where(c, T, F)
                if partition.masked and hardmask:
                    # The original partition has missing data and a
                    # hardmask, so apply the original partition's mask
                    # to the new array.
                    new = numpy_ma_masked_where(array.mask, new, copy=False)
            #--- End: if
                            
            # --------------------------------------------------------
            # Replace the partition's subarray with the new numpy
            # array
            # --------------------------------------------------------
            partition.subarray = new

            partition.close()
        #--- End: for
    #--- End: def

    def setitem(self, *args, **kwargs):
        raise NotImplementedError(
"cf.Data.setitem is dead. Use cf.Data.__setitem__ or cf.Data.setdata instead.")

    def setmask(self, *args, **kwargs):
        raise NotImplementedError(
"cf.Data.setmask is dead. Use cf.Data.__setitem__ or cf.Data.setdata instead.")

    def sin(self):
        '''
Take the trigonometric sine of the data array in place.

Units are accounted for in the calculation. If the units are not
equivalent to radians (such as Kelvin) then they are treated as if
they were radians. For example, the the cosine of 90 degrees_east is
1.0, as is the sine of 1.57079632 radians.

The Units are changed to '1' (nondimensionsal).

:Returns:

    None

**Examples**

>>> d.Units
<CF Units: degrees_north>
>>> print d.array
[[-90 0 90 --]]
>>> d.sin()
>>> d.Units
<CF Units: 1>
>>> print d.array
[[-1.0 0.0 1.0 --]]

>>> d.Units
<CF Units: m s-1>
>>> print d.array
[[1 2 3 --]]
>>> d.sin()
>>> d.Units
<CF Units: 1>
>>> print d.array
[[0.841470984808 0.909297426826 0.14112000806 --]]

'''
        if self.Units.equivalent(_units_radians):
            self.Units = _units_radians

        self.func(numpy_sin)

        self.override_units(_units_1)
    #--- End: def

    def squeeze(self, axes=None):
        '''

Remove size 1 axes from the data array in place.

By default all size 1 axes are removed, but particular axes may be
selected with the keyword arguments.

:Parameters:

    axes : (sequence of) int or str, optional
        Select the axes.  By default all size 1 axes are removed. The
        *axes* argument may be one, or a sequence, of:

          * An internal axis identifier. Selects this axis.
        ..

          * An integer. Selects the axis coresponding to the given
            position in the list of axes of the data array.

        No axes are removed if *axes* is an empty sequence.

:Returns:

    out : list of ints
        The axes which were squeezed as a tuple of their positions.

.. seealso:: `~cf.Data.expand_dims`, `~cf.Data.flip`,
             `~cf.Data.transpose`

**Examples**

>>> v.shape
(1,)
>>> v.squeeze()
>>> v.shape
()

>>> v.shape
(1, 2, 1, 3, 1, 4, 1, 5, 1, 6, 1)
>>> v.squeeze((0,))
>>> v.shape
(2, 1, 3, 1, 4, 1, 5, 1, 6, 1)
>>> v.squeeze(1)
>>> v.shape
(2, 3, 1, 4, 1, 5, 1, 6, 1)
>>> v.squeeze([2, 4])
>>> v.shape
(2, 3, 4, 5, 1, 6, 1)
>>> v.squeeze([])
>>> v.shape
(2, 3, 4, 5, 1, 6, 1)
>>> v.squeeze()
>>> v.shape
(2, 3, 4, 5, 6)

'''
        ndim = self._ndim
        if not ndim:
            if axes or axes == 0:
                raise ValueError(
                    "Can't squeeze: Can't remove an axis from a scalar %s" %
                    self.__class__.__name__)
            return []
        #--- End: if

        shape = list(self._shape)

        if axes is None:
            axes = [i for i, n in enumerate(shape) if n == 1]
        else:
            axes = self._parse_axes(axes, 'squeeze')
            
            # Check the squeeze axes
            for i in axes:
                if shape[i] > 1:
                    raise ValueError(
                        "Can't squeeze %s: Can't remove axis of size > 1" %
                        self.__class__.__name__)
        #--- End: if

        if not axes:
            return axes

        # Still here? Then the data array is not scalar and at least
        # one size 1 axis needs squeezing.
        data_axes  = self.dimensions[:]
        directions = self.directions.copy()

        for axis in [data_axes[i] for i in axes]:
            i = data_axes.index(axis)
            shape.pop(i)
            dir = directions.pop(data_axes.pop(i))

            for partition in self.partitions.matrix.flat:
                p_location = partition.location[:]
                p_shape    = partition.shape[:]

                p_location.pop(i)
                p_shape.pop(i)

                partition.location = p_location
                partition.shape    = p_shape
        #--- End: for

        ndim = len(shape)
        self._ndim  = ndim
        self._shape = tuple(shape)

        # Set the axes
        self.dimensions = data_axes

        # Set the directions
        if not ndim:
            # The data has been squeezed to a scalar
            self.directions = dir
        else:
            self.directions = directions

        # Remove size 1 partition dimensions       
        self.partitions.squeeze()

        return axes
    #--- End: def

    def tan(self):
        '''

Take the trigonometric tangent of the data array in place.

Units are accounted for in the calculation. If the units are not
equivalent to radians (such as Kelvin) then they are treated as if
they were radians. For example, the the tangent of 45 degrees_east is
1.0, as is the tangent of 0.78539816 radians.

The Units are changed to '1' (nondimensionsal).

:Returns:

    None

.. seealso:: `~cf.Data.cos`, `~cf.Data.sin`

**Examples**

>>> d.Units
<CF Units: degrees_north>
>>> print d.array
[[-45 0 45 --]]
>>> d.tan()
>>> d.Units
<CF Units: 1>
>>> print d.array
[[-1.0 0.0 1.0 --]]

>>> d.Units
<CF Units: m s-1>
>>> print d.array
[[1 2 3 --]]
>>> d.tan()
>>> d.Units
<CF Units: 1>
>>> print d.array
[[1.55740772465 -2.18503986326 -0.142546543074 --]]

'''
        if self.Units.equivalent(_units_radians):
            self.Units = _units_radians

        self.func(numpy_tan)

        self.override_units(_units_1)
    #--- End: def

    def transpose(self, axes=None):
        '''
        
Permute the dimensions of the data array in place.

:Parameters:

    axes : (sequence of) int or str, optional
        The new axis order of the data array. By default the order is
        reversed. Each axis of the new order is identified in turn as
        one of:

          * An internal axis identifier. Selects this axis.
        ..

          * An integer. Selects the axis coresponding to the given
            position in the list of axes of the data array.

:Returns:

    out : list of ints
        The axes which were transposed

.. seealso:: `~cf.Data.expand_dims`, `~cf.Data.flip`,
             `~cf.Data.squeeze`

**Examples**

>>> d.ndim
3
>>> d.shape
(19, 73, 96)
>>> d.dimensions
['dim0', 'dim1', 'dim2']
>>> d.transpose()
>>> d.shape
(96, 73, 19)
>>> d.transpose([1, 0, 2])
>>> d.shape
(73, 96, 19)
>>> d.transpose(['dim2', 'dim0', 'dim1'])
>>> d.shape
(96, 19, 73)
>>> d.dimensions
['dim2', 'dim0', 'dim1']
>>> d.transpose((1,  'dim1', 0))
>>> d.shape
(19, 73, 96)

'''
        ndim = self._ndim    
        if ndim <= 1:
            return range(ndim-1, -1, -1)
        
        # Parse the axes. By default, reverse the axes
        if axes is None:
            axes = range(ndim-1, -1, -1)
        else:
            axes = self._parse_axes(axes, 'transpose')

            # Return unchanged if axes are in the same order as the data
            if axes == range(ndim):
                return axes

            if len(axes) != ndim:
                raise ValueError(
                    "Can't tranpose: Axes don't match array: %s" % (axes,))
        #--- End: if

        # Permute the axes
        data_axes = self.dimensions
        self.dimensions = [data_axes[i] for i in axes]

        # Permute the shape
        shape = self._shape
        self._shape = tuple([shape[i] for i in axes])
        
        # Permute the locations map
        for partition in self.partitions.matrix.flat:
            location = partition.location
            shape    = partition.shape

            partition.location = [location[i] for i in axes]
            partition.shape    = [shape[i]    for i in axes]
        #--- End: for

        return axes
    #--- End: def

    def func(self, f, *args, **kwargs):
        '''

Apply an element-wise array operation to the data array in place.

:Parameters:

    f : function
        The function to be applied.

    args, kwargs : 
        Any arguments and keyword arguments passed to the function
        given by *f*.

:Returns:

   None

**Examples**

>>> print d.array
[[ 0.          1.57079633]
 [ 3.14159265  4.71238898]]
>>> import numpy
>>> d.func(numpy.cos)
>>> print d.array
[[ 1.0  0.0]
 [-1.0  0.0]]
>>> def f(x, y, a=0):
...     return x*y + a
...
>>> d.func(f, 2, a=10)
>>> print d.array
[[ 12.0  10.0]
 [-12.0  10.0]]

'''      
        conform_args = self.conform_args()

        for partition in self.partitions.matrix.flat:
            array = partition.dataarray(**conform_args)
            partition.subarray = f(array, *args, **kwargs)
            partition.close()
        #--- End: for
    #--- End: def

#--- End: class
   
def _size_of_index(index, size=None):
    '''

Return the number of elements resulting in applying an index to a
sequence.

:Parameters:

    index : slice or list of ints
        The index being applied to the sequence.

    size : int, optional
        The number of elements in the sequence being indexed. Only
        required if *index* is a slice object.

:Returns:

    out : int
        The length of the sequence resulting from applying the index.

**Examples**

>>> _size_of_index(slice(None, None, -2), 10)
5
>>> _size_of_index([1, 4, 9])
3

'''
    if isinstance(index, slice):
        # Index is a slice object
        start, stop, step = index.indices(size)
        div, mod = divmod(stop-start, step)
        if mod != 0:
            div += 1
        return div
    else:
        # Index is a list of integers
        return len(index)
#--- End: def

def _parse_indices(data, indices):
    '''

:Parameters:

   data : 

   indices : sequence of indices

:Returns:

    parsed_indices, reverse_dimensions : {list, list} 

**Examples**

'''
    parsed_indices = parse_indices(data, indices)
#    parsed_indices = data.parse_indices(indices)

    reverse_dimensions = []

    for i, index in enumerate(parsed_indices):      

        if isinstance(index, slice):
            size = data.shape[i]
            if index.step < 0:              
                # If the slice step is negative, then transform the
                # original slice to a new slice with a positive step
                # such that the result of the new slice is the reverse
                # of the result of the original slice.
                #
                # For example, if the original slice is slice(6,0,-2)
                # then the new slice will be slice(2,7,2):
                #
                # >>> a = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
                # >>> a[slice(6, 0, -2)]
                # [6, 4, 2]
                # >>> a[slice(2, 7, 2)]
                # [2, 4, 6]
                # a[slice(6, 0, -2)] == list(reversed(a[slice(2, 7, 2)]))
                # True
                start, stop, step = index.indices(size)
                step    *= -1
                div, mod = divmod(start-stop-1, step)
                div_step = div*step
                start   -= div_step
                stop     = start + div_step + 1
                
                index = slice(start, stop, step)
                reverse_dimensions.append(i)         
            #--- End: if      

            # If step is greater than one, make sure that index.stop isn't
            # bigger than it needs to be.
            if index.step > 1:
                start, stop, step = index.indices(size)
                div, mod = divmod(stop-start-1, step)
                stop     = start + div*step + 1
                index    = slice(start, stop, step)
            #--- End: if

            parsed_indices[i] = index    

        else:
            # --------------------------------------------------------
            # Check that an integer list is strictly monotonic and if
            # it's descending reverse it so that it's ascending
            # --------------------------------------------------------
            step = index[1] - index[0]

            if step > 0:                        
                if not all(x < y for x, y in izip(index, index[1:])):
                    raise ValueError("Bad slice (not strictly monotonic): %s" % index)

            elif step < 0:
                if not all(x > y for x, y in izip(index, index[1:])):
                    raise ValueError("Bad slice (not strictly monotonic): %s" % index)
                
                # Reverse the list so that it's strictly monotonically
                # increasing and make a note that this dimension will
                # need reversing later
                index.reverse()
                reverse_dimensions.append(i)
            else:
                # Step is 0
                raise ValueError("Bad slice (not strictly monotonic): %s" % index)
        #--- End: if
    #--- End: for
    
    return parsed_indices, reverse_dimensions
#--- End: def

def _overlapping_partitions(partitions, indices, dim2position, 
                            master_directions):
    '''

Return the nested list of (modified) partitions which overlap the
given indices to the master array.

:Parameters:

    partitions : cf.PartitionMatrix

    indices : tuple

    dim2position : dict

    master_directions : dict or bool

:Returns:

    out : list
        A (possibly nested) list of cf.Partition objects.

**Examples**

>>> type f.Data
<class 'cf.data.Data'>
>>> d.dimensions
['dim1', 'dim2', 'dim0']
>>> dim2position = {'dim0': 2, 'dim1': 0, 'dim2' : 1}
>>> indices = (slice(None), slice(5, 1, -2), [1,3,4,8])
>>> x = _overlapping_partitions(d.partitions, indices, dim2position, master_directions)
>>> type(x)
list

'''

    if partitions.size == 1:
        partition = partitions.matrix.item()

        # Find out if this partition overlaps the original slice
        p_indices, shape = partition.overlaps(indices)

        if p_indices is None:
            # This partition is not in the slice out of bounds - raise
            # error?
            return
        
        # Still here? Create a new partition                      
        partition = partition.copy()
#        partition.part = partition.new_part(list(p_indices),
#                                            dim2position,
#                                            directions)
#        partition.new_part(list(p_indices), dim2position, master_directions)
        partition.new_part(p_indices, dim2position, master_directions)
        partition.shape = shape
        
        new_partition_matrix      = numpy_empty(partitions.shape, dtype=object)
        new_partition_matrix[...] = partition
        
        return new_partition_matrix
    #--- End: if

    partitions_list        = []
    partitions_list_append = partitions_list.append

    flat_pm_indices        = []
    flat_pm_indices_append = flat_pm_indices.append

    partitions_flat = partitions.matrix.flat

    i = partitions_flat.index

    for partition in partitions_flat:
        # Find out if this partition overlaps the original slice
#        p_indices, shape = _partition_overlaps(partition, indices)
        p_indices, shape = partition.overlaps(indices)
        
        if p_indices is None:
            # This partition is not in the slice 
            i = partitions_flat.index
            continue

        # Still here? Create a new partition                      
        partition = partition.copy()
#        partition.part = partition.new_part(list(p_indices),
#                                            dim2position,
#                                            master_directions)
#        partition.new_part(list(p_indices), dim2position, master_directions)
        partition.new_part(p_indices, dim2position, master_directions)
        partition.shape = shape
        
        partitions_list_append(partition)

        flat_pm_indices_append(i)

        i = partitions_flat.index 
    #--- End: for

    new_shape = [len(set(s)) 
                 for s in numpy_unravel_index(flat_pm_indices, partitions.shape)]

    new_partition_matrix = numpy_empty((len(flat_pm_indices),), dtype=object)
    new_partition_matrix[...] = partitions_list
    new_partition_matrix.resize(new_shape)

    return new_partition_matrix
#--- End: def

#def subspace_array(array, indices):
#    '''
#
#Subset the input numpy array with the given indices. Indexing is similar to
#that of a numpy array. The differences to numpy array indexing are:
#
#1. An integer index i takes the i-th element but does not reduce the rank of
#   the output array by one.
#
#2. When more than one dimension's slice is a 1-d boolean array or 1-d sequence
#   of integers then these indices work independently along each dimension
#   (similar to the way vector subscripts work in Fortran).
#
#indices must contain an index for each dimension of the input array.
#'''
#    gg = []
#    for i, x in enumerate(indices):
#        if not isinstance(x, slice):
#            gg.append(i)
#    #--- End: for
#
#    len_gg = len(gg)
#    if len_gg > 1:
#        # Slice the dimensions one at a time
#        indices = list(indices)
#        if numpy_ma_isMA(array):
#            for axis in gg:
#                array = numpy_ma_take(array, indices[axis], axis=axis)
#                indices[axis] = slice(None)
#        else:
#            for axis in gg:
#                array = numpy_take(array, indices[axis], axis=axis)
#                indices[axis] = slice(None)
#        #--- End: if
#
#        if len_gg < len(indices):
#            array = array[tuple(indices)]
#    else:
#        # Slice all dimensions at the same time
#        array = array[tuple(indices)]
#
#    return array
##--- End: def

## --------------------------------------------------------------------
##
## --------------------------------------------------------------------
#def _st2dt(timestr):
#    '''
#
#Parse an ISO 8601 datetime string into a datetime object.
#
#:Parameters:
#
#    timestr : str
#
#:Returns:
#
#    out : netCDF4.netcdftime.datetime
#
#'''
#    year,month,day,hour,minute,second,utc_offset = _netCDF4_netcdftime_parse_date(timestr)
#    if utc_offset:
#        raise ValueError("Can't specify a time offset from UTC")
#            
#    return _netCDF4_netcdftime_datetime(year, month, day, hour, minute, second)
##--- End: def
#_v_st2dt = numpy_vectorize(_st2dt, otypes=[object])

# --------------------------------------------------------------------
#
# --------------------------------------------------------------------
def _getattr(x, attr): return getattr(x, attr)
_array_getattr = numpy_vectorize(_getattr)

