from numpy import amax             as numpy_amax
from numpy import amin             as numpy_amin
from numpy import argsort          as numpy_argsort
from numpy import average          as numpy_average
from numpy import broadcast_arrays as numpy_broadcast_arrays
from numpy import copy             as numpy_copy
from numpy import empty            as numpy_empty
from numpy import expand_dims      as numpy_expand_dims
from numpy import multiply         as numpy_multiply
from numpy import ones             as numpy_ones
from numpy import reshape          as numpy_reshape
from numpy import resize           as numpy_resize
from numpy import shape            as numpy_shape
from numpy import sum              as numpy_sum

from numpy.ma import array       as numpy_ma_array
from numpy.ma import average     as numpy_ma_average
from numpy.ma import is_masked   as numpy_ma_is_masked
from numpy.ma import masked_all  as numpy_ma_masked_all

from copy      import deepcopy
from itertools import izip
from operator  import mul as operator_mul

from ..ancillaryvariables import AncillaryVariables
from ..cellmethods        import CellMethods
from ..constants          import masked as cf_masked
from ..coordinate         import Coordinate, DimensionCoordinate
from ..coordinatebounds   import CoordinateBounds
from ..field              import Field
from ..fieldlist          import FieldList
from ..functions          import broadcast_array
from ..variable           import Variable

from ..data.data import Data

#_debug = True
_debug = False

def collapse(fields, method, axes=None, weights=None, unbiased=False,
             ignore_size1=True, squeeze=False, **kwargs):
    '''

Return new fields with dimensions collapsed by statistical operations.

In all cases, the presence of missing data in an input field's data
array is accounted for during the collapse.

The following collapse methods are available over any subset of the
fields' axes, where the methods are defined exactly as for CF cell
methods:

==================  ==============================  =======================
Method              Description                     Options
==================  ==============================  =======================
min                 Minima over the specified axes
                                                                                    
max                 Maxima over the specified axes

mid_range           Means of the minima and maxima  
                    over the specified axes

sum                 Sums over the specified axes     

mean                Means over the specified axes   Weighted or unweighted
                                                                                    
standard_deviation  Standard deviations over the    Weighted or unweighted,
                    specified axes                  biased or unbiased

variance            Variances over the specified    Weighted or unweighted,
                    axes                            biased or unbiased
==================  ==============================  =======================

:Parameters:

    fields : (sequence of) Field or FieldList
        The field or fields to be collapsed.

    method : str or CellMethods
        Define the collapse method (or methods). There are two ways of
        setting the collapse method:

           * A string containing a single method (such as ``'min'``).
           
             - The method is one of the available CF cell methods.

             - All of the dimensions specified by the *axes* parameter
               are collapsed simultaneously with this method.                 
        ..

           * A CF cell methods-like string (such as ``'time: max'`` or
             ``'time: max dim0: mean 2: min'``) or a `cf.CellMethods`
             object equivalent to such a string.
           
             - Specifies both collapse methods and the axes to apply
               them to.

             - Each method is one of the available CF cell
               methods. The axes are interpreted as for the *axes*
               parameter.

             - When multiple cell methods are specified, each collapse
               is carried out independently and in the order that they
               are given. The *axes* parameter must not be set.

    axes : (sequence of) int or str, optional
        The axis or axes which are to be simultaneously collapsed with
        the method specified by the *method* parameter. May be any
        one, or any combination of:

           * A dimension identity, as a defined by an appropriate
             coordinate's `~Coordinate.identity` method (such as
             ``'time'``).
             
           * The integer position of a dimension in the field's data
             array (such as ``2``).
             
           * The internal name of a dimension in the field's domain
             (such as ``'dim0'``).

        By default all dimensions with size greater than 1 are
        collapsed, unless the *method* is a CF cell methods-like
        string or a `cf.CellMethods` object, in which case the
        collapse axes are already specified and the *axes* parameter
        must not be set.

    weights : str or tuple, optional
        Specify the weights to be used in the collapse. One of:

           =============  ===============================================
           weights        Description
           =============  ===============================================
           `None`         Perform an unweighted collapse (this is the
                          default).
 
           ``'equal'``    Perform an unweighted collapse.
           
           `!tuple`       The axes for which non-equal weights are to
                          be computed. Each axis is specified by one
                          of:

                            * The identity of one of its one
                              dimensional coordinate objects (such as
                              ``'time'``)
                            
                            * A coordinate object type (one of
                              ``'T'``, ``'X'``, ``'Y'`` or ``'Z'``)
                            
                            * An axis or coordinate object domain
                              identifier (such as ``'dim1'``).

                          If cell measures are available for any
                          specified axes then they will consitute the
                          weights, otherwise weights are based on each
                          axis's dimension coordinate object.
                          
                          In general, dimension coordinate object
                          weights will be linear, with the exception
                          of a latitude axis, for which linear weights
                          of the sine of the coordinate values are
                          calculated.

                            Example: To specify latitude/longitude area
                            weights you could set ``weights=('X', 'Y')``.

                          If only coordinate object types are
                          specified, then they may be given as a
                          concatenated string instead of a tuple.

                            Example: ``weights=('X', 'T')`` is
                            equivalent to ``weights='XT'``.       
           =============  ===============================================

        .. note:: The interface for the specification of weights is
                  currently limited (version 0.9.8). A fully general
                  API for specifiying weights is under development.

    unbiased : bool, optional
        If True then calculate unbiased statisics where applicable. By
        default biased statistics are calculated.

    ignore_size1 : bool, optional
        If False then raise a ValueError if a dimension to be
        collapsed has size 1, regardless of whether or not it spans
        the data array. By default size 1 collapse dimensions are
        ignored, regardless of whether or not they span the data
        array.

    squeeze : bool, optional
        If True then remove collapsed dimensions from the output data
        array. By default the output data array retains collapsed
        dimensions with size 1 if they were spanned by the original
        data array. It is not possible to remove collapsed dimensions
        when mutliple methods have been specified.

:Returns:

   out : Field or FieldList
        The collapsed fields.

**Examples**

>>> g = cf.collapse(f, 'max')
>>> g = cf.collapse(f, 'min', axes=[2, 1])
>>> g = cf.collapse(f, 'sum', axes='dim2')
>>> g = cf.collapse(f, 'mid_range', axes=('latitude', 0, dim2'))
>>. g = cf.collapse(f, 'mean', axes='latitude', weights=None)
>>> g = cf.collapse(f, 'mean', axes='longitude', weights=None)
>>> g = cf.collapse(f, 'mean', axes=['longitude', 'latitude'], weights=None)
>>> g = cf.collapse(f, 'standard_deviation', weights=None)
>>> g = cf.collapse(f, 'variance', weights=None, unbiased=True)
>>> g = cf.collapse(f, 'mean', axes='latitude')
>>> g = cf.collapse(f, 'mean', axes='longitude')
>>> g = cf.collapse(f, 'mean', axes=['longitude', 'latitude'])
>>> g = cf.collapse(f, 'variance')
>>> g = cf.collapse(f, 'standard_deviation', unbiased=True)
>>> g = cf.collapse(f, 'longitude: mean latitude: max')

>>> weights = {'time': None,
...            (2, 'dim1'): numpy.arange(45).reshape(5, 9))
>>> g = cf.collapse(f, 'mean', weights=weights)

'''
    # ---------------------------------------------------------------
    # Flatten an input list or tuple of fields and field lists into a
    # single field list
    # ---------------------------------------------------------------
    if isinstance(fields, (list, tuple)):
        fields = [f for element in fields for f in element]

    # If method is a cell methods string (such as 'area: mean time:
    # max') then convert it to a CellMethods object
    if ':' in method:
        method = CellMethods(method)

    # ================================================================
    # If method is a CellMethods object then process its methods one
    # by one with recursive calls and then return.
    # ================================================================
    if isinstance(method, CellMethods):
        if squeeze:            
            raise ValueError(
"Can't collapse: Can't squeeze collapsed axes with multiple methods")

        if axes is not None:
            raise ValueError(
                "Can't collapse: Can't set both axes and cell methods")

        methods = [cm['method'] for cm in method]
        axes    = [cm['name']   for cm in method]
        for name, m in zip(axes, methods):
            if None in name:
                raise ValueError(
                    "Can't collapse: Invalid '%s' cell method name: %s" %
                    (m, None))
        #--- End: if

        # Collapse according to each cell method in turn with
        # recursive calls
        new = fields

        while methods:
            new = collapse(new, methods.pop(0), axes=axes.pop(0), weights=weights, 
                           unbiased=unbiased, ignore_size1=ignore_size1,
                           squeeze=False)
        #--- End: while

        return new
    #--- End: if

    # ================================================================
    # Still here? Then method is a string describing a single method
    # (e.g. 'mean')
    # ================================================================
    out = FieldList()

    weights_in = deepcopy(weights)

    for f in fields:
#        print repr(f)
        new = f.copy(_omit_Data=True)

        domain = f.domain
        axes_sizes = domain.dimension_sizes
        data_axes  = domain.dimensions['data']

        #-------------------------------------------------------------
        # Parse the collapse axes
        #-------------------------------------------------------------
        all_collapse_axes = domain.axes(axes, **kwargs)

        collapse_axes = [axis for axis in all_collapse_axes 
                         if axis in data_axes and axes_sizes[axis] > 1]
        
        #axes, collapse_axes = f._parse_axes(axes, 'collapse', dims=True,
        #                                    ignore=ignore_size1,
        #                                    default=range(f.ndim))
        #
        #if ignore_size1:
        #    # Ignore size 1 dimensions
        #    collapse_axes = [dim for dim in collapse_axes if dimension_sizes[dim] > 1]
        #    axes = f._parse_axes(collapse_axes, 'collapse', dims=False)
        #else:
        #    # Fail if any size 1 dimensions are specified
        #    for dim in collapse_axes:
        #        if dimension_sizes[dim] == 1:
        #            raise ValueError(
        #                "Can't collapse: Size 1 dimension has been specified")
        ##--- End: if

        # ------------------------------------------------------------
        # Calculate inferred weights, if required
        # ------------------------------------------------------------
        if weights_in in (None, 'equal'):
            weights = {}

        elif weights_in:

            weights = tuple(weights)
            sss = f.axes(weights)
            sss.update(f.axes(ctype=weights))

            wdims = sss.intersection(collapse_axes)

            weights = {}
            
            # Get weights from cell measures, if possible.
            for key, cm in domain.items(role='c', axes=wdims).iteritems():
                cm_axes = domain.item_axes(key)
                if cm.size == 1: # or wdims.isdisjoint(cm_axes):
                    # Don't use cell measure if it only has one
                    # element or only spans non-collapse axes
                    continue

                if 1 in cm.shape:
                    cm = cm.copy()
                    cm.squeeze()
                    cm_axes = cm.data._dimensions

                weights[tuple(cm_axes)] = cm.data

                wdims.difference_update(cm_axes)
           #--- End: for

            # Get weights for any remaining dimensions from dimension
            # coordinates, if possible.
            for axis in wdims:
 #               coord = domain.item(axis) #role='d', axis=axis)
                w = calc_weights(domain.item(axis), infer_bounds=True)
                if w is not None:
                    weights[(axis,)] = w
            #--- End: for

        #elif isinstance(weights_in, dict):
        #    weights = deepcopy(weights_in)
        #
        #    new_weights = {}
        #    all_dims_with_weights = []
        #
        #    for key, value in weights.iteritems():
        #        if not isinstance(key, tuple):
        #            key = (key,)
        #
        #        if set(key).issubset('TXYZ'):
        #            key = tuple(key)
        #            new_key = [domain.axis(ctype=axis)) for axis in key]
        #        else:
        #            new_key = domain.axes(key, ordered=True)
        #
#       #         new_key = [domain.axis(axis) for axis in key]
        #        if None in new_key or len(key) != weight.ndim:
        #            raise ValueError(
        #                "ERROR: Can't identify weights axes: %s" % repr(key))
        #
        #        if set(new_key).isdisjoint(collapse_axes): 
        #            # Ignore a weight which doesn't span any collapse
        #            # dimensions
        #            continue
        #
        #        if set(new_key).intersection(all_dims_with_weights):
        #            raise ValueError(
        #                "ERROR: Two weights span the same axis: %s" % 
        #                set(new_key).intersection(all_dims_with_weights))
        #
        #        all_dims_with_weights.extend(new_key)
        #
        #        new_weights[tuple(new_key)] = value
        #    #--- End: for            
        #
        #    weights = new_weights
        #
        #    wdims = set(collapse_axes).difference(all_dims_with_weights)
        #    # Get weights from cell measures, if possible.
        #    for key, cm in domain.items(role='c').iteritems():
        #        cm_dims = domain.item_axes(key):
        #        if cm.size == 1 or wdims.isdisjoint(cm_dims):
        #            # Don't use cell measure if it only has one
        #            # element or only spans non-collapse axes
        #            continue
        #
        #        if 1 in cm.shape:
        #            cm = cm.copy()
        #            cm.squeeze()
        #            cm_dims = cm.data._dimensions
        #
        #        weights[tuple(cm_dims)] = cm.data
        #
        #        wdims.difference_update(cm_dims)
        #    #--- End: for
        #
        #    # Get weights for any remaining dimensions from dimension
        #    # coordinates, if possible.
        #    for dim in wdims:
        #        dim_coord = domain.item(dim)                    
        #        if dim_coord is not None and dim_coord.size > 1:
        #            weights[(dim,)] = dim_coord
        #    #--- End: for
        #
        #    
        #
        #
        #
        #    # Sort out 1-d weights
        #    s = domain.analyse()
        #
        #    for dim in collapse_axes:
        #        if dim not in all_dims_with_weights:
        #            weights[(dim,)] = 'infer'
        #        elif dim not in weights:
        #            continue
        #
        #        if weights[dim] == 'infer':
        #            found_cell_measure = False
        #            for key, cm in domain.items(role='c').iteritems():
        #                if cm.ndim == 1 and domain.dimensions[key][0] == dim:
        #                    weights[dim] = cm
        #                    found_cell_measure = True
        #                    continue
        #            #--- End: for
        #            if not found_cell_measure:
        #                if dim in s['dim_to_coord']:
        #                    weights[dim] = s['dim_to_coord'][dim]
        #                else:
        #                    weights[dim] = _equal_weights(axes_sizes[dim])
        #
        #        elif weights[dim] in (None, 'equal'):
        #            weights[dim] = _equal_weights(axes_sizes[dim])
        #    #--- End: if                
        else:
            raise TypeError("Can't collapse: Invalid weights: %s" % repr(weights_in))

        # ------------------------------------------------------------
        # Parse any weights which are Variables into Data objects (or
        # None)
        # ------------------------------------------------------------
        #for key, value in weights.items():
        #    if isinstance(value, (Variable, Data)):
        #        weights[key] = value.data
        #    else:
        #        weights[key] = Data(value)
        ##--- End: for
            

           #if isinstance(value, DimensionCoordinate):
           #    # Convert a coordinate into a weights Data object
           #    weights[key] = calc_weights(value, infer_bounds=True)
           #elif isinstance(value, Coordinate):
           #    raise ValueError("Can't use %s object as weights" % 
           #                     value.__class__.__name__)
           #elif isinstance(value, Variable):
           #    weights[key] = value.data
        #--- End: for

        x = []
        non_partition_dims = set(domain.dimensions['data']).difference(f.Data.pmdimensions)

        for dims in weights:
            if dims in non_partition_dims:
                x.append(dims)
            elif non_partition_dims.issuperset(dims):
                x.append(dims)
        
        if len(x) > 1:
            base = [1]*len(x)
            w=[]
            new_key = []
            for j, i in enumerate(x):
                shape = base[:]
                if isinstance(i, tuple): 
                    shape[j] = weights[i].size
                    w.append(weights[i].unsafe_array.reshape(shape))
                    del weights[i]
                    new_key.extend(i)
                else:
                    shape[j] = weights[i].size
                    w.append(weights[i].unsafe_array.reshape(shape))
                    del weights[i]
                    new_key.append(i)
        
#            weights[tuple(new_key)] = reduce(numpy_multiply, w)
            weights[tuple(new_key)] = Data(reduce(numpy_multiply, w).copy())
            
        # ------------------------------------------------------------
        # Collapse the data array
        # ------------------------------------------------------------
        new.Data = collapse_data(f.data, method, collapse_axes,
                                 weights=weights, unbiased=unbiased)

        # ------------------------------------------------------------
        # Update the cell methods
        # ------------------------------------------------------------
        _update_cell_methods(new, method, collapse_axes)

        # ------------------------------------------------------------
        # Update ancillary variables
        # ------------------------------------------------------------
        _update_ancillary_variables(new, collapse_axes)

        # ------------------------------------------------------------
        # Update transforms
        # ------------------------------------------------------------
        _update_transform_fields(new, collapse_axes)

        #-------------------------------------------------------------
        # Collapse the domain
        #-------------------------------------------------------------
        collapse_domain(new.domain, collapse_axes)

        # ------------------------------------------------------------
        # Squeeze collapsed dimensions
        # ------------------------------------------------------------
        if squeeze:            
            size1_collapse_axes = [axis for axis in all_collapse_axes
                                   if axis in data_axes and axes_sizes[axis] == 1]
            new.squeeze(size1_collapse_axes)

        # We need to finalize the new field because some of the
        # coordinates will have been redefined
        new.finalize()

        out.append(new)
    #--- End: for

    if isinstance(fields, out[0].__class__):
        # Return a Field
        return out[0]
    else:
        # Return a FieldList
        return out
#--- End: def

def collapse_data(data, method, collapse_axes, weights={}, unbiased=False):
    '''

Return a data array collapsed over one or more dimensions.

:Parameters:

    data : cf.Data
        Data to be collapsed.

    method : str

    collapse_axes : sequence of str

    weights : dict, optional

    unbiased : bool, optional

:Returns:

    out : cf.Data

**Examples**

'''
    data_axes = data.dimensions
    data_directions = data.directions

#    collapse_axes = [data_dimensions[i] for i in axes]
    collapse_dirs = [data_directions[dim] for dim in collapse_axes]

    #-----------------------------------------------------------------
    # Parse the collapse axes
    #-----------------------------------------------------------------
#    axes, dims = data._parse_axes(axes, 'collapse', dims=True,
#                                  default=range(data.ndim))
#    axes, dims = zip(*sorted(zip(axes, dims)))
#    axes = list(axes)

    collapse_positions = [data_axes.index(axis) for axis in collapse_axes]
    collapse_positions, collapse_axes = zip(*sorted(zip(collapse_positions,
                                                        collapse_axes)))

    #-----------------------------------------------------------------
    # Initialize the output data array
    #-----------------------------------------------------------------
    indices = [slice(None)] * data.ndim
    for i in collapse_positions:
        indices[i] = 0

    new = data[tuple(indices)]
    
    # Squeeze the collapse axes out of the output data. This is
    # important. But note that we'll put them back at the end.
    new.squeeze(collapse_positions)
    
    new_dimensions = new.dimensions
    new_directions = new.directions

    #-----------------------------------------------------------------
    # Transpose the input data so that the collapse dimensions are all
    # in the fastest varying (outer) positions.
    # ----------------------------------------------------------------
#    non_collapse_axes = [i for i in xrange(data.ndim) if i not in axes]
#    if non_collapse_axes + axes != range(data.ndim):
#        data = data.copy()
#        data.transpose(non_collapse_axes + axes)


    non_collapse_axes = [axis for axis in data.dimensions
                         if axis not in collapse_axes]

    transpose_axes = non_collapse_axes + list(collapse_axes)
    if transpose_axes != data.dimensions:
        data = data.copy()
        data.transpose(transpose_axes)

    n_collapse_axes     = len(collapse_axes)
    n_non_collapse_axes = len(non_collapse_axes)

    #-----------------------------------------------------------------
    # Parse the weights.
    #
    # * Change the keys from dimension names to the integer positions
    #   of the dimensions.
    #
    # * Make sure all non-null weights are Data objects.
    # -----------------------------------------------------------------
    if weights:

        weights = weights.copy()

        # It is important to reset data_axes because we may have
        # transposed data
        data_axes = data.dimensions

        for key in weights.keys():
            weight = weights.pop(key)

            # Ignore undefined weights
            if weight is None:
                continue
            
#            # Make sure the weight is a Data object
#            if not isinstance(weight, data.__class__):
#                weight = type(data)(weight)
            
            # Ignore string-valued weights
            if weight.dtype.char == 'S':
                continue
                
            # Ignore size 1 weights
            if weight.size == 1:
                continue
            
            key = [data_axes.index(k) for k in key]
            if len(key) > 1:
                weight = weight.copy()
                weight.transpose(numpy_argsort(key))
                key.sort()

            weights[tuple(key)] = weight

            # Check that the shape of the weights match the sizes of
            # their corresponding data array dimensions.
            data_shape = data.shape
            weight_shape = weight.shape
            for i, j in enumerate(key):
                if weight_shape[i] != data_shape[j]:
                    raise ValueError(
                        "Can't collapse: Incorrect weights shape: %s" %
                        repr(weight_shape))

            #if key in data_axes:
            #    weights[(data_axes.index(key),)] = weight
            #elif isinstance(key, tuple):
            #    a = [key.index(dim) for dim in collapse_axes if dim in key]
            #    if a not in ([], range(len(key))):
            #        nc = [i for i in xrange(len(key)) if i not in a]
            #        weight = weight.copy()
            #        weight.transpose(nc + a)
            #        
            #        key = [key[i] for i in nc+a]
            #    #--- End: if
            #
            #    if not set(key).issubset(data_axes):
            #        raise ValueError(
            #            "Can't collapse: Incorrect weights dimensions: %s" %
            #            repr(key))
            #
            #    weights[tuple([data_axes.index(k) for k in key])] = weight
        #--- End: for

        ## Check that the shape of the weights match the sizes of their
        ## corresponding data array dimensions.
        #data_shape = data.shape
        #for key, weight in weights.iteritems():
        #    weight_shape = weight.shape
        #    for i, j in enumerate(key):
        #        if weight_shape[i] != data_shape[j]:
        #            raise ValueError(
        #                "Can't collapse: Incorrect weights shape: %s" %
        #                repr(weight_shape))
    #--- End: if

    if not weights:
        weights = None

    units = new.Units

#    #-------------------------------------------------------------
#    # Set the output data type
#    # ------------------------------------------------------------
#    datatype = data.dtype
#    if datatype.kind is 'i' and method not in ['max', 'min', 'mode']:
#        # The output data-type is float if data is of integer type
#        datatype = numpy_dtype(float)

    # ----------------------------------------------------------------
    # Define the collapse function
    # ----------------------------------------------------------------
    func = _collapse_functions[method]

    for partition in new.partitions.flat:
        del partition.subarray

    data.to_memory()

    save = new.save_to_disk()

    for partition in new.partitions.flat:

        # It is very important to set the part attribute to [] because
        # the value inherited from the input partition matrix could be
        # wrong
        partition.part = []

        # We need to reset the dimensions attribute because
        # Data.squeeze() does not change it.
        partition._dimensions = new_dimensions[:]
# dch need to be sure that array matches with dimensions!

        # We need to reset the directions attribute because
        # Data.squeeze() does not change it.
        partition._directions = new._copy_directions(new_directions)

        partition.Units = units

        master_indices = partition.indices + (slice(None),) * n_collapse_axes

        partition.subarray = func(data, master_indices,
                                  n_non_collapse_axes, n_collapse_axes,
                                  weights=weights, unbiased=unbiased)
        
        partition.close(save)
    #--- End: for

    # Clear the datatype in case it has changed
    del new.dtype

    # Put the collapsed dimensions back in with size 1
    for position, axis, direction in zip(collapse_positions, collapse_axes,
                                         collapse_dirs):
        new.expand_dims(position, direction=direction, axis=axis)

    # ----------------------------------------------------------------
    # Return the collapsed data
    # ----------------------------------------------------------------
    return new
#--- End: def

def collapse_domain(domain, collapse_axes):
    '''

Collapse dimensions of a domain to size 1 in place.

A dimension coordinate for a collapse dimension is replaced with a
size 1 coordinate whose bounds are the maximum and minimum of the
original bounds array (of the original data array if there are no
bounds) and whose datum is the average of the new bounds.

All auxiliary coordinates and cell measures which span any of the
collapse dimensions are removed.

:Parameters:

    domain : Domain
        The domain to be collapsed.

    collapse_axes : sequence of strs 
        The dimensions to be collapsed, defined by their domain
        identifiers.

:Returns:

    None

**Examples**

>>> collapse_domain(f.domain, ['dim2', 'dim1'])

'''
    for axis in collapse_axes:
        # Ignore dimensions which are already size 1
        if domain.dimension_sizes[axis] == 1:
            continue
        
        dim_coord = domain.item(axis)
        if dim_coord is None:
            continue

        if dim_coord._isbounded:
            bounds = [dim_coord.bounds.datum(0), dim_coord.bounds.datum(-1)]
        else:
            bounds = [dim_coord.datum(0), dim_coord.datum(-1)]

        units = dim_coord.Units
        data   = Data([(bounds[0] + bounds[1])*0.5], units)
        bounds = Data([bounds], units)

        dim_coord.insert_data(data, bounds=bounds, copy=False)

        # Put the new dimension coordinate into the domain
        domain.insert_axis(1, key=axis, replace=True)
        domain.insert_dim(dim_coord, key=axis, copy=False, replace=True,
                          finalize=False)
        
        # Remove all auxiliary coordinates and cell measures which
        # span this dimension
        domain.remove_items(role=('a', 'c'), axes=axis)
    #--- End: for
#--- End: def

def _update_cell_methods(f, method, collapse_axes):
    '''

:Parameters:

    f : Field

    method : str

    collapse_axes : sequence of strs

:Returns:

    None

**Examples**

'''
    if not hasattr(f, 'cell_methods'):
         f.cell_methods = CellMethods()

    domain = f.domain

    collapse_axes = sorted(collapse_axes)

    name = []
    for axis in collapse_axes:
        if axis in domain.d:
            name.append(domain.d[axis].identity(default=axis))
        else:
            name.append(axis)
    #--- End: for

    string = '%s: %s' % (': '.join(name), method)

    cell_method = CellMethods(string)
    
    cell_method[0]['dim'] = collapse_axes

    f.cell_methods += cell_method
#--- End: def

def _update_ancillary_variables(f, collapse_axes):
    '''

:Parameters:

    f : Field

    collapse_axes : sequence of strs

:Returns:

    None

**Examples**

'''
    if not hasattr(f, 'ancillary_variables'):
        return

    new_av = AncillaryVariables()

    for av in f.ancillary_variables:

        dim_map = av.domain.map_axes(f.domain)

        # Don't keep the ancillary variable if its master data array
        # has an undefined dimension with size > 1.
        keep = True
        for dim1 in set(av.data_axes()).difference(dim_map):
            if av.domain.dimension_sizes[dim1] > 1:
                keep = False
                break
        #--- End: for
        if not keep:
            continue

        # Don't keep the ancillary variable if its master data array
        # has a defined dimension with size > 1 which is also a
        # collapse dimension.
        keep = True
        for dim1, dim0 in dim_map.iteritems():
            if dim0 in collapse_axes and av.domain.dimension_sizes[dim1] > 1:
                keep = False
                break
        #--- End: for
        if not keep:
            continue

        new_av.append(av)
    #--- End: for

    if new_av:
        f.ancillary_variables = new_av
    else:
        del f.ancillary_variables 
#--- End: def

def _update_transform_fields(f, collapse_axes):
    '''

:Parameters:

    f : Field

    collapse_axes : sequence of strs

:Returns:

    None

**Examples**

'''
    for t in f.transforms().itervalues():
        if t.grid_mapping:
            continue
        
        for term, value in t.items():
            if not isinstance(value, Field):
                # Keep the term if it isn't a field
                continue 

            data_axes = set(value.data_axes())

            if data_axes.isdisjoint(collapse_axes):
                # Keep the term if it doesn't span any collapse axes
                continue

            dim_map = value.domain.map_axes(f.domain)
                
            # Don't keep the term if its master data array has an
            # undefined axis with size > 1.
            for dim1 in data_axes.difference(dim_map):
                if av.domain.dimension_sizes[dim1] > 1:
                    t[term] = None
                    break
            #--- End: for
            if t[term] is None:
                continue

            # Don't keep the ancillary variable if its master data
            # array has a defined axis with size > 1 which is also a
            # collapse axis
            for axis, axis1 in dim_map.iteritems():
                if (axis1 in collapse_axes and axis in data_axes and
                    value.domain.dimension_sizes[axis] > 1):
                    t[term] = None
                    break
            #--- End: for
        #--- End: for
    #--- End: for
#--- End: def

def _max(data, master_indices, n_non_collapse_axes, n_collapse_axes,
         **kwargs):
    '''
        
Return the maximum value of a data array over the given indices.

If the entire array is masked then the returned value will be
`cf.masked`.

:Parameters:

    data : cf.Data

    kwargs : *optional*
        Ignored.

:Returns:

    out : scalar
        The maximum value, or `cf.masked` if the entire array is
        masked.

**Examples**

>>> print d
[0 1 2 --]
>>> _max(d)
2
>>> d[...] = cf.masked
>>> print d
[-- -- -- --]
>>> _max(d)
masked

'''
    master_shape = data.shape
    data = data[master_indices]

    out, n = _weighted_map(_numpy_amax, _numpy_amax, data, master_indices, master_shape, 
                         n_non_collapse_axes, n_collapse_axes,
                         weights=None)

    if not out:
        return numpy_ma_masked_all(data.shape[:n_non_collapse_axes], data.dtype)
    
    pmaxs = out[0]

    if pmaxs.ndim == n_non_collapse_axes:
        return pmaxs
    else:
        return numpy_amax(pmaxs, axis=0)

#--- End: def

def _mean(data, master_indices, n_non_collapse_axes, n_collapse_axes,
          weights=None, **kwargs):
    '''
        
Return the weighted mean of a data array over the given indices.

If no weights are specified then equal weights are assumed.

The return type is numpy.float64 if the master data array is of
integer type, otherwise it is of the same type as master data array.

:Parameters:

    data : cf.Data

     master_indices: tuple


    weights : dict, optional

    kwargs : *optional*
        Ignored.

:Returns:

    out : scalar
        The weighted mean.

'''
    master_shape = data.shape
    data = data[master_indices]

    out, n = _weighted_map(numpy_average, numpy_ma_average, data, master_indices, master_shape,
                         n_non_collapse_axes, n_collapse_axes,
                         weights=weights, returned=True)

    if not out:
        return numpy_ma_masked_all(data.shape[:n_non_collapse_axes], data.dtype)
    
    pmeans, psums_of_weights = out

    if n: #pmeans.shape == data.shape[:n_non_collapse_axes]:
        return pmeans
    else:
        if numpy_ma_is_masked(pmeans):
            return numpy_ma_average(pmeans, axis=0, weights=psums_of_weights)
        else:
            return numpy_average(pmeans, axis=0, weights=psums_of_weights)
#--- End: def

def _mid_range(data, master_indices, n_non_collapse_axes, n_collapse_axes,
               **kwargs):
    '''
 
Return the average of the minimum and maximum values of a data array
over the given indices.

If the entire array is masked then the returned value will be
`cf.masked`.

:Parameters:

    data : cf.Data

    kwargs : *optional*
        Ignored.

:Returns:

    out : scalar
        The mid range value, or `cf.masked` if the entire array is
        masked.

**Examples**

>>> print d
[-- 1 2 4]
>>> _mid_range(d)
2.5
>>> d[...] = cf.masked
>>> print d
[-- -- -- --]
>>> _mid_range(d)
masked

'''
    master_shape = data.shape
    data = data[master_indices]

    out, n = _weighted_map(_numpy_amin_amax, _numpy_amin_amax, data, master_indices, master_shape,
                         n_non_collapse_axes, n_collapse_axes,
                         weights=None)

    if not out:
        return numpy_ma_masked_all(data.shape[:n_non_collapse_axes], data.dtype)
    
    pmins, pmaxs = out

    if n: #pmins.shape == data.shape[:n_non_collapse_axes]: 
        gmin = pmins
        gmax = pmaxs
    else:
        gmin = numpy_amin(pmins, axis=0)
        gmax = numpy_amax(pmaxs, axis=0)
    #--- End: if

    gmax -= gmin
    gmax *= 0.5
    return gmax
#--- End: def

def _numpy_amin_amax(array, axis=None):
    '''

Return the minimum and maximum of an array or the minimum and maximum
along an axis.

:Parameters:


    array : numpy array_like
        Input array

    axis : int, optional
        Axis along which to operate. By default, flattened input is
        used.

:Returns:

    out : tuple
        The minimum and maximum values.

'''
    return numpy_amin(array, axis=axis), numpy_amax(array, axis=axis)
#--- End: def

def _numpy_amin(array, axis=None):
    '''

Return the minimum of an array or the minimum along an axis.

:Parameters:

    array : numpy array_like
        Input array

    axis : int, optional
        Axis along which to operate. By default, flattened input is
        used.

:Returns:

    out : tuple
        The minimum in a 1-tuple.

'''
    return (numpy_amin(array, axis=axis),)
#--- End: def

def _numpy_amax(array, axis=None):
    '''

Return the maximum of an array or the maximum along an axis.

:Parameters:


    array : numpy array_like
        Input array

    axis : int, optional
        Axis along which to operate. By default, flattened input is
        used.

:Returns:

    out : tuple
        The maximum in a 1-tuple.

'''
    return (numpy_amax(array, axis=axis),)
#--- End: def

def _min(data, master_indices, n_non_collapse_axes, n_collapse_axes,
         **kwargs):
    '''
        
Return the minimum value of a data array over the given indices.

If the entire array is masked then the returned value will be
`cf.masked`.

The return type is numpy.float64 if the master data array is of
integer type, otherwise it is of the same type as master data array.

:Parameters:

    data : cf.Data

    kwargs : *optional*
        Ignored.

:Returns:

    out : scalar
        The minimum value, or `cf.masked` if the entire array is
        masked.

**Examples**

>>> print d
[-- 1 2 3]
>>> _min(d)
1
>>> d[...] = cf.masked
>>> print d
[-- -- -- --]
>>> _min(d)
masked

'''  
    master_shape = data.shape
    data = data[master_indices]
    
    out, n = _weighted_map(_numpy_amin, _numpy_amin, data, master_indices, master_shape,
                         n_non_collapse_axes, n_collapse_axes,
                         weights=None)

    if not out:
        return numpy_ma_masked_all(data.shape[:n_non_collapse_axes], data.dtype)

    pmins = out[0]

    if n: #pmins.shape == data.shape[:n_non_collapse_axes]:
        return pmins
    else:
        return numpy_amin(pmins, axis=0)
#--- End: def

def _numpy_sum(array, axis=None):
    '''

:Parameters:


    array : numpy array-like
        Input array

    axis : int, optional
        Axis along which to operate. By default, flattened input is
        used.

:Returns:

    out : tuple
        The sum in a 1-tuple.

'''
    return (numpy_sum(array, axis=axis),)
#--- End: def

def _mode(data, master_indices, n_non_collapse_axes, n_collapse_axes,
          **kwargs):
    '''

Return the mode of the master data array.
    
    kwargs : *optional*
        Ignored.

If the entire array is masked then the returned value will be
`cf.masked`.

'''
    raise NotImplementedError("Haven't implement mode, yet. Sorry.")
#--- End: def

def _standard_deviation(data, master_indices, n_non_collapse_axes, n_collapse_axes,
                        weights=None, unbiased=False, **kwargs):
    '''

Return the standard deviation of a data array over the given indices.
    
If the entire array is masked then the returned value will be
`cf.masked`.

The return type is numpy.float64 if the master data array is of
integer type, otherwise it is of the same type as master data array.

:Parameters:

    data : cf.Data

    indices: tuple

    unbiased : bool, optional
        If True then the unbiased sample standard deviation will be
        returned. By default the biased population standard deviation
        will be returned.

    weights : dict, optional        

    unbiased : bool, optional

    kwargs : *optional*
        Ignored.

:Returns:

    out : scalar
        The standard deviation, or `cf.masked` if the entire array is
        masked.

**Examples**

>>> print d
[-- 1 2 3]
>>> _standard_deviation(d)

>>> d[...] = cf.masked
>>> print d
[-- -- -- --]
>>> _standard_deviation(d)
masked

'''
    variance = _variance(data, master_indices,
                         n_non_collapse_axes, n_collapse_axes,
                         weights=weights, unbiased=unbiased)

    variance **= 0.5

    return variance
#--- End: def

def _sum(data, master_indices, n_non_collapse_axes, n_collapse_axes,
         **kwargs):
    '''
        
Return the sum of a data array over the given indices.

If the entire array is masked then the returned value will be
`cf.masked`.

The return type is numpy.float64 if the master data array is of integer type,
otherwise it is of the same type as master data array.

:Parameters:

    data : cf.Data

    kwargs : *optional*
        Ignored.

:Returns:

    out : scalar
        The sum value, or `cf.masked` if the entire array is masked.

**Examples**

>>> print d
[-- 1 2 3]
>>> _sum(d)
6
>>> d[...] = cf.masked
>>> print d
[-- -- -- --]
>>> _sum(d)
masked

'''
    master_shape = data.shape
    data = data[master_indices]
    
    out, n = _weighted_map(_numpy_sum, _numpy_sum, data, master_indices, master_shape,
                         n_non_collapse_axes, n_collapse_axes,
                         weights=None)
    
    if not out:
        return numpy_ma_masked_all(data.shape[:n_non_collapse_axes], data.dtype)

    psums = out[0]

    if n: #psums.ndim == n_non_collapse_axes:
        return psums
    else:
        return numpy_sum(psums, axis=0)
#--- End: def

def _components(array, axis=None, weights=None, unbiased=False):
    '''

Return selected statistics derived from an array:

* Mean of the array
* Biased population variance of the array
* Sum of weights for the array
* Sum of squares of weights for the array (only if unbiased is True)

:Parameters:

    array : array-like
        The array from which the statistics are calculated.

    weights : array-like, optional
        The importance that each element has in the computations. If
        set then the weights array must have the same shape as
        *array*. By default all data in *array* are assumed to have a
        weight equal to one.

:Returns:

    out : tuple
        Either a 3-tuple of the mean, biased population variance and
        sum of weights in that order; or, if *unbiased* is True, a
        4-tuple which has an additional element of the sums of the
        squares of weights.

'''
    # Methodology:
    # http://en.wikipedia.org/wiki/Standard_deviation#Population-based_statistics
    # http://en.wikipedia.org/wiki/Weighted_mean#Weighted_sample_variance

    masked = numpy_ma_is_masked(array)

    if not masked:
        mean, sw = numpy_average(array, weights=weights, axis=axis,
                                 returned=True)

        

    else:
        # Do masked case separately since it is much slower
        mean, sw = numpy_ma_average(array, weights=weights, axis=axis,
                                    returned=True)
        
    if mean.size > 1:
        # We collapsed over axis=-1 and array has 2 or more
        # dimensions, so add an extra size 1 dimension to the mean so
        # that broadcasting works when we calculate sigma2.
        mean = numpy_expand_dims(mean, -1)
        reshape_mean = True
    else:
        reshape_mean = False
        
    sigma2   = array - mean
    sigma2 **= 2

    if not masked:
        sigma2 = numpy_average(sigma2, axis=axis, weights=weights)
    else:
        sigma2 = numpy_ma_average(sigma2, axis=axis, weights=weights)

    if reshape_mean:
        mean = mean.reshape(mean.shape[:-1])
        
    if unbiased:
        # Calculate the sum of the square of weights
        if weights is None:
            sw2 = numpy_copy(sw)
        else:
            if weights.ndim < array.ndim:
                weights = broadcast_array(weights, array.shape)

            if masked:
                # If the array is masked then we have to mask the
                # weights to get the correct sum of squares
                weights = numpy_ma_array(weights, mask=array.mask, copy=False)

            sw2 = numpy_sum(weights*weights, axis=axis) 
        #--- End: if

        return (mean, sigma2, sw, sw2)

    else:
        return (mean, sigma2, sw)
#--- End: def

def _variance(data, master_indices, n_non_collapse_axes, n_collapse_axes,
              weights=None, unbiased=False, **kwargs):
    '''

Return the variance of a data array over the given indices.

The biased population variance or the unbiased sample variance may be
calculated.
    
If the entire array is masked then the returned value will be
`cf.masked`.

The return type is numpy.float64 if the master array is of integer
type, otherwise it is of the same type as master data array.

:Parameters:

    data : cf.Data

    indices: tuple

    unbiased : bool, optional
        If True then the unbiased sample variance will be returned. By
        default the biased population variance will be returned.

    weights : dict, optional

    unbiased : bool, optional

    kwargs : *optional*
        Ignored.

:Returns:

    out : scalar
        The variance, or `cf.masked` if the entire array is masked.

**Examples**

>>> print d
[-- 1 2 3]
>>> _variance(d)

>>> d[...] = cf.masked
>>> print d
[-- -- -- --]
>>> _variance(d)
masked

'''
    # Methodology:
    # http://en.wikipedia.org/wiki/Standard_deviation#Population-based_statistics
    # http://en.wikipedia.org/wiki/Weighted_mean#Weighted_sample_variance

    master_shape = data.shape
    data = data[master_indices]
    
    out, n = _weighted_map(_components, _components, data, master_indices, master_shape,
                           n_non_collapse_axes, n_collapse_axes,
                           weights=weights, unbiased=unbiased)
                         
    if _debug:
        print '_variance: out =', out

    if not out:
        return numpy_ma_masked_all(data.shape[:n_non_collapse_axes], data.dtype)

    if unbiased:        
        pmeans, pvariances, psums_of_weights, psums_of_weights2 = out
    else:
        pmeans, pvariances, psums_of_weights = out

    # -----------------------------------------------------------------------------
    # Calculate the biased variance
    # (http://en.wikipedia.org/wiki/Standard_deviation#Population-based_statistics)
    #
    # global biased variance = {[SUM(psw(pv+pm**2)]/sw} - m**2
    # 
    # where psw = partial sum of weights  (psums_of_weights)
    #       pv  = partial biased variance (pvariances)
    #       pm  = partial mean            (pmeans)
    #       sw  = global sum of weights   (gsum_of_weights)
    #       m   = global mean             (gmean)
    # -----------------------------------------------------------------------------
      
    if n: #pmeans.ndim == n_non_collapse_axes: 
        gvariance       = pvariances
        gsum_of_weights = psums_of_weights
        if unbiased:
            gsum_of_weights2 = psums_of_weights2
    
        if _debug:
            print '_variance: gmean              =', pmeans
            print '_variance: gsum_of_weights    =', gsum_of_weights
            print '_variance: gvariance (biased) =', gvariance

    else:
        if not numpy_ma_is_masked(pmeans):
            gmean, gsum_of_weights = numpy_average(pmeans, axis=0,
                                                   weights=psums_of_weights,
                                                   returned=True)
        else:
            gmean, gsum_of_weights = numpy_ma_average(pmeans, axis=0,
                                                      weights=psums_of_weights,
                                                      returned=True)

        if _debug:
            print '_variance: gmean              =', gmean
            print '_variance: gsum_of_weights    =', gsum_of_weights

        pmeans *= pmeans
        pvariances += pmeans
        pvariances *= psums_of_weights
        gvariance = numpy_sum(pvariances, axis=0)

        # Divide by the global sum of weights
        gvariance /= gsum_of_weights
        # Subtract the square of the global mean
        gmean *= gmean
        gvariance -= gmean

        if _debug:
            print '_variance: gvariance (biased) =', gvariance

        if unbiased:
            gsum_of_weights2 = numpy_sum(psums_of_weights2, axis=0)
    #--- End: if

    if unbiased:
        # ---------------------------------------------------------------------
        # Convert the biased population variance to the unbiased
        # sample variance
        # (http://en.wikipedia.org/wiki/Weighted_mean#Weighted_sample_variance)
        #
        # global unbiased variance = [sw**2/(sw**2 - sw2)] * sigma2
        # 
        # where sw     = global sum of weights            (gsum_of_weights)
        #       sw2    = global sum of squares of weights (gsum_of_weights2)
        #       sigma2 = global biased variance           (gvariance)
        # ---------------------------------------------------------------------
        gsum_of_weights *= gsum_of_weights
        gvariance *= gsum_of_weights
        gsum_of_weights -= gsum_of_weights2
        gvariance /= gsum_of_weights

        if _debug:
            print '_variance: gsum_of_weights2   =', gsum_of_weights2
       #--- End: if

    if _debug:
        print '_variance: gvariance          =', gvariance
   
    # ------------------------------------------------------------
    # Return the global variance
    # ------------------------------------------------------------
    return gvariance
#--- End: def

#def _create_weights(partition, master_indices, master_weights, one_axis=False):
#    '''
#
#:Parameters:
#
#    partition : cf.Partition
#
#    master_indices : tuple
#
#    master_weights : dict
#
#    one_axis : bool, optional
#        If True then the collapse iver exactly one axis
#:Returns:
#
#    out : numpy.ma.array
#
#**Examples**
#
#'''
#    if not master_weights:
#        return None
#
#    partition_indices = partition.indices
#    partition_shape   = tuple(partition.shape)
#    base_shape = [1] * len(partition_shape)
#
#    master_indices += partition_indices[len(master_indices):]
#
#    weights = []
#
#    for key, weight in master_weights.iteritems():
#        shape = base_shape[:]
#        indices = []
#        for i in key:
#            shape[i] = partition_shape[i]
#            indices.append(master_indices[i])
#        #--- End: for
#
#        weight = weight[tuple(indices)].unsafe_array
#
#        if not one_axis:
#            if weight.shape != partition_shape:
#                weight = weight.reshape(shape)
#            
#        if not (weight.size == 1 and weight.item() == 1):
#            weights.append(weight)
#    #--- End: for
#
#    # Create the outer product of the weights - i.e. create a weights
#    # array which has the same shape as the partition's data array.
#    if len(weights) > 1:
#        if one_axis:
#            raise ValueError("aaaaaaaaarrrrrrrrrrrrrrrrrrrrrgggggggggggggghhhh")
#        weights = reduce(numpy_multiply, weights)
#    elif weights:
#        weights = weights[0]
#    else:
#        weights = None
#
##    # Broadcast the weights across all dimensions of the
##    # partition, and return them
##    if weights.shape != partition_shape:
##        if not numpy_ma_is_masked(weights):
##            weights, j = numpy_broadcast_arrays(weights, partition.subarray)
##    #        return numpy_resize(weights, partition_shape)
##        else:
##            weights = numpy_ma_resize(weights, partition_shape)
##    #--- End: if
#
#    return weights
##--- End: def

def _create_weights(array, indices, master_indices, master_shape, master_weights,
                    n_non_collapse_axes, n_collapse_axes):
    '''

:Parameters:

    array : numpy array

    master_indices : tuple

    master_weights : dict

s    axes : 

:Returns:

    out : numpy.ma.array

**Examples**

'''
    array_shape = array.shape
    array_ndim  = array.ndim

    weights_indices = []
    for master_index, index, size in izip(master_indices,
                                          indices,
                                          master_shape):
        start , stop , step = master_index.indices(size)

        size1, mod = divmod(stop-start-1, step)            
        
        start1, stop1, step1 = index.indices(size1+1)
        
        size2, mod = divmod(stop1-start1, step1)

        if mod != 0:
            size2 += 1
            
        start += start1 * step
        step  *= step1
        stop   = start + (size2-1)*step + 1
#        if step > 0:
#            stop += 1
#        else:
#            stop -= 1

#        if stop < 0:
#            stop = None

        weights_indices.append(slice(start, stop, step))
    #--- End: for

    base_shape = (1,) * array_ndim

    weights = [] 
    for key, weight in master_weights.iteritems():
        shape = list(base_shape)
        index = []
        for i in key:
            shape[i] = array_shape[i]
            index.append(weights_indices[i])
        #--- End: for

        weight = weight[tuple(index)].unsafe_array

        if weight.ndim != array_ndim:
            # Make sure that the weight has the same number of
            # dimensions as the array
            weight = weight.reshape(shape)

        weights.append(weight)
    #--- End: for

    weights_out = weights[0]

    if len(weights) > 1:
        # There are two or more weights, so create their product
        for w in weights[1:]:
            # Can't do this in-place because of broadcasting woe
            weights_out = weights_out * w
   
    if weights_out.shape[:n_non_collapse_axes] == base_shape[:n_non_collapse_axes]:
        # The input weights only span collapse axes
        if weights_out.shape[n_non_collapse_axes:] != array_shape[n_non_collapse_axes:]:
            # The input weights span some, but not all, of the
            # collapse axes, so broadcast the weights over all
            # collapse axes.
            weights_out = broadcast_array(weights_out, array_shape[n_non_collapse_axes:])
        #--- End: if
        # Get rid of the non-collapse axes
        weights_out = weights_out.reshape(array_shape[n_non_collapse_axes:])
    else:
        # The input weights span at least one non-collapse axis, so
        # broadcast the weights over all axes.
        weights_out = broadcast_array(weights_out, array_shape)   

    return weights_out
#--- End: def

#def _weighted_map(data, function, ma_function, indices, dim2size, datatype,
#                  **kwargs):
#    '''
#
#Return a list of the results of applying the function to the
#partitions of the argument Data object.
#
#All keyword arguments are passed into the function call. If there is a
#`weights` keyword argument then this should either evaluate to False
#or be a dictionary of weights for at least one of the data dimensions.
#
#:Parameters:
#
#    data : cf.Data
#
#    function : function
#
#    ma_function : function
#
#    indices: tuple or None
#
#    dim2size : int
#
#    datatype : numpy.dtype
#
#    kwargs : *optional*
#
#:Returns:
#
#    out : numpy array
#
#**Examples**
#
#'''    
#    weights = kwargs.get('weights', None)
#
#    if not dim2size:
#        
#        
#        
#        array = data[indices].unsafe_array
#        
#        if len(axis) == 1:
#            kwargs['axis'] = axis[0]
#        else:
#            new_shape = array.shape[:array.ndim-len(axis)]
#            new_shape += (reduce(operator_mul, array.shape[array.ndim-len(axis):]),)
#            array = array.reshape(new_shape)
#            kwargs['axis'] = -1
##        else:
#        if weights is not None:
##            kwargs['weights'] = _create_weights(partition, indices, weights)
#            kwargs['weights'] = _create_weights(False, indices, weights)

#        if not numpy_ma_is_masked(array):
#            # Array has no masked values
#            return function(array, **kwargs)
#        else:
#            # Array has at least one masked value, so use the masked
#            # array function
#            return ma_function(array, **kwargs)
#    #--- End: for
#
#    data = data[indices]
#
#    if dim2size == 1:
#        shape = data.pmsize
#    else:
#        shape = (data.pmsize, dim2size)
#
#    out = numpy_ma_empty(shape, dtype=datatype)
#
##    if weights is not None:
##        kwargs = kwargs.copy()
#    conform_args = data.conform_args(revert_to_file=True, readonly=True)
#
#    for i, partition in enumerate(data.partitions.flat):
#        array = partition.dataarray(**conform_args)
#        partition.close()
#        
#        if partition.masked and array.mask.all():
#            out[i] = numpy.ma.masked
#            continue
#
#        # Still here?
#        if weights is not None:
#            kwargs['weights'] = _create_weights(partition, indices, weights)
#            
#        if not partition.masked: #numpy_ma_is_masked(array):
#            # Array has no masked values
#            out[i] = function(array, **kwargs)
#        else:
#            # Array has at least one masked value, so use the masked
#            # array function
#            out[i] = ma_function(array, **kwargs)
#    #--- End: for
#
#    return out
##--- End: def

def _weighted_map(function, ma_function, data, master_indices, master_shape, 
                  n_non_collapse_axes, n_collapse_axes,
                  weights=None, **kwargs):
    '''

Return a list of the results of applying the function to the
partitions of the argument Data object.

All keyword arguments are passed into the function call. If there is a
`weights` keyword argument then this should either evaluate to False
or be a dictionary of weights for at least one of the data dimensions.

:Parameters:

    function : function

    ma_function : function

    data : cf.Data
        A subspace of the master array.

    master_indices: tuple
        The indices of the master array corresponding to *data*.    

    n_non_collapse_axes : int

    n_collapse_axes : int

    kwargs : *optional*

:Returns:

    out : list

**Examples**

'''    
    out = []
    one_partition = False
                
    conform_args = data.conform_args(revert_to_file=True, readonly=True)

    for i, partition in enumerate(data.partitions.matrix.flat):
        array = partition.dataarray(**conform_args)

        masked = partition.masked

        if masked and array.mask.all():
            continue

        # Still here?
        if weights:
            kwargs['weights'] = _create_weights(
                array, partition.indices, master_indices, master_shape, weights,
                n_non_collapse_axes, n_collapse_axes)

        if n_collapse_axes == 1:
            # Exactly one axis is to be collapsed
            kwargs['axis'] = -1
            new_shape = None
        elif not n_non_collapse_axes:
            # All axes are to be collapsed
            kwargs.pop('axis', None)
            new_shape = None
        else:
            # At least two, but not all, axes are to be
            # collapsed. Therefor we need to reshape the array and the
            # weights so that.
            kwargs['axis'] = -1
            shape = array.shape 
            new_shape  = shape[:n_non_collapse_axes]
            new_shape += (reduce(operator_mul, shape[n_non_collapse_axes:]),)            
            array = numpy_reshape(array.copy(), new_shape)
            if weights:
                w = kwargs['weights']
                if w.ndim != array.ndim:
                    new_shape = (w.size,)

                kwargs['weights'] = numpy_reshape(w, new_shape)
        #--- End: if

        if not masked: 
            # Array has no masked values
            p_out = function(array, **kwargs)
        else:
            # Array has at least one masked value
            p_out = ma_function(array, **kwargs)

        partition.close()

        if not out:
            size = data.pmsize - i
            if size == 1:
                # There is exactly one partition with some valid data,
                # so we are done.
                out = p_out
                one_partition = True
                break
            else:
                # There is at least one partition with some valid data
                j = i
                datatype = p_out[0].dtype
                for y in p_out:
                    out.append(numpy_ma_masked_all((size,) + numpy_shape(y),
                                                   datatype))
        #--- End: if

        k = i - j
        for x, y in izip(out, p_out):
            x[k] = y
    #--- End: for

    return out, one_partition
#--- End: def

def calc_weights(coord, infer_bounds=False):
    '''

Calculate weights for collapsing. See def collapse for help

:Parameters:

    coord : cf.DimensionCoordinate

:Returns:

    out : cf.Data or None

'''

    if coord is None or coord.size == 1:
        return None
    elif coord.Y or coord.identity() == 'grid_latitude':
        return _latitude_area_weights(coord, infer_bounds=infer_bounds)    
    else:
        return _linear_weights(coord, infer_bounds=infer_bounds)
#--- End: def

def _equal_weights(size):
    '''

Return equal weights

:Parameters:

    size : int

:Returns:

    out : cf.Data

'''
    return Data(numpy_ones(size))
#--- End: def

def _linear_weights(coord, infer_bounds):
    '''

Calculate linear weights based on cell bounds.

:Parameters:

    coord : cf.DimensionCoordinate

:Returns:

    out : cf.Data

'''
#    if coord.size == 1:
#        return None

    bounds = coord.get_bounds(create=infer_bounds).data
    bounds = abs(bounds[..., 1] - bounds[..., 0])
    bounds.squeeze(bounds.dimensions[-1])

    if (bounds == bounds[0]).all():
        # The weights are equal
        return _equal_weights(bounds.size) #Data(numpy_ones(bounds.size))

    return bounds
#--- End: def

def _latitude_area_weights(coord, infer_bounds):
    '''

Calculate latitude-area weights based on cell bounds

:Parameters:

    coord : cf.DimensionCoordinate

:Returns:

    out : cf.Data

'''
#    if coord.size == 1:
#        return None

    bounds = coord.get_bounds(create=infer_bounds).data
    bounds.sin()
    bounds = abs(bounds[..., 1] - bounds[..., 0])
    bounds.squeeze(bounds.dimensions[-1])

    if (bounds == bounds[0]).all():
        # The weights are equal
        return _equal_weights(bounds.size) #Data(numpy_ones(bounds.size))

    return bounds
#--- End: def

_collapse_functions = {'mean': _mean,
                       'max': _max,
                       'min': _min,
                       'mid_range': _mid_range,
                       'variance': _variance,
                       'standard_deviation': _standard_deviation,
                       'sum': _sum,
                       'mode': _mode,
                       }
