from .utils     import Dict
from .functions import RTOL, ATOL, equals, equivalent
from .variable  import Variable


# ====================================================================
#
# Transform object
#
# ====================================================================

class Transform(Dict):
    '''

A CF transform construct.

A transform construct defines a mapping from one set of coordinates
which can not geo-locate the field construct's data to another set of
coordinates that can geo-locate the field construct's data.

Transform constructs correspond to the functions of the CF-netCDF
attributes formula_terms, which describes how to compute a vertical
auxiliary coordinate variable from components and grid_mapping, which
describes how to compute true latitude and longitude auxiliary
coordinate variables from horizontal projection dimension coordinates,
or describes the figure of earth for true latitude and longitude
coordinate variables.

A transform is a dictionary-like object for which each key/value pair
comprises the name of a term of the transform and its corresponding
variable. A term that is omitted is assumed to be zero.

The each variable is one of the following:

=====================  ====================================================
Variable               Notes
=====================  ====================================================
Pointer to coordinate  A domain's coordinate identifier; or a
                       coordinate's identity, as a defined by the
                       coordinate's `~cf.Coordinate.identity` method.

Python numeric scalar  

`cf.Data`              Must have an array size of 1

`cf.Field`             

``None``               Indicates that the term is unset. Note that
                       this is different than the term being missing,
                       as in this case there is no assumption of a
                       value of zero.
=====================  ====================================================


**Attributes**

===============  ========  ===================================================
Attribute        Type      Description
===============  ========  ===================================================
`!coordinates`   ``set``   A set of pointers to coordinates (may be empty)
`!grid_mapping`  ``bool``  True if and only if the transform is a grid_mapping
`!name`          ``str``   The identity of the transform (such as 
                           ``'ocean_sigma_coordinate'`` or
                           ``'rotated_latitude_longitude'``)
===============  ========  ===================================================


**Examples**

>>> t
<CF Transform: rotated_latitude_longitude>
>>> print t.dump()
Transform: rotated_latitude_longitude
    grid_north_pole_latitude = 38.0
    grid_north_pole_longitude = 190.0

>>> print t.dump()
Transform: atmosphere_hybrid_height_coordinate
    a = 'aux0'
    b = 'aux1'
    orog = <CF Field: surface_altitude(grid_longitude(96), grid_latitude(73)) m>

'''

    def __init__(self, **kwargs):
        '''

**Initialization**

:Parameters:

    kwargs :
        Keys and values are initialized exactly as keyword arguments
        for a built-in dict. In general, keys are transform parameter
        names (such as ``'grid_north_pole_latitude'`` or ``'sigma'``)
        with appropriate values, with three exceptions:

          * If the key ``'standard_name'`` then its value is the
            identity for a formula_terms-type transform and no term is
            created.
          
          * If the key ``'grid_mapping_name'`` is specified then its
            value is the identity for a grid_mapping-type transform
            and no term is created.
          
          * If the key ``'inputs'`` is specified then its value is a
            sequence of coordinate pointers for *all* of the inputs to
            the transform which do not have dedicated terms.

        One of the keys ``'standard_name'`` and
        ``'grid_mapping_name'`` must be specified, but ``'inputs'`` is
        optional.

**Examples**

>>> t = cf.Transform(grid_mapping_name='rotated_latitude_longitude',
...                  grid_north_pole_latitude=38.0, 
...                  grid_north_pole_longitude=190)

>>> t = cf.Transform(standard_name='atmosphere_ln_pressure_coordinate',
...                  p0=1000, lev='dim1', 

>>> t = cf.Transform(standard_name='atmosphere_ln_pressure_coordinate',
...                  p0=cf.Data(100000, 'Pa'),
...                  lev='atmosphere_ln_pressure_coordinate')

'''
        super(Transform, self).__init__(**kwargs)

        if 'grid_mapping_name' in self:
            self.grid_mapping = True
            name = self.pop('grid_mapping_name')
        elif 'standard_name' in self:
            self.grid_mapping = False
            name = self.pop('standard_name')
        else:
            raise ValueError(
"ERROR: Transform not initialized with a grid_mapping_name or standard_name")
        
        coordinates = set([value for value in self.itervalues()
                           if isinstance(value, basestring)])

        if 'inputs' in self:
            coordinates.update(self.pop('inputs'))
        elif not self.grid_mapping:
            coordinates.add(name)
        elif name == 'rotated_latitude_longitude':
            coordinates.add('grid_latitude')
            coordinates.add('grid_longitude')
        elif name == 'latitude_longitude':
            coordinates.add('latitude')
            coordinates.add('longitude')
        else:
            coordinates.add('projection_x_coordinate')
            coordinates.add('projection_y_coordinate')
        #--- End: if

        self.coordinates = coordinates
        self.name        = name
    #--- End: def

    def __eq__(self, other):
        '''

The rich comparison operator ``==``

x.__eq__(y) <==> x==y

'''
        return self.equals(other)
    #--- End: def

    def __ne__(self, other):
        '''

The rich comparison operator ``!=``

x.__ne__(y) <==> x!=y

'''
        return not self.equals(other)
    #--- End: def

    def __hash__(self):
        '''

x.__hash__() <==> hash(x)

:Raises:

    ValueError:
        If the transform is not a grid mapping.

'''
        if not self.grid_mapping:
            raise ValueError("Can't hash a formula_terms transform")

        h = sorted(self.items())
        h.append(self.identity())

        return hash(tuple(h))
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)

''' 
        try:
            return '<CF %s: %s>' % (self.__class__.__name__, self.identity(''))
        except AttributeError:
            return '<CF %s: >' % self.__class__.__name__
    #--- End: def

    def __setitem__(self, item, value):
        '''
x.__setitem__(key, value) <==> x[key]=value

'''
        if item in self:
            old_value = self.pop(item)
            if isinstance(old_value, basestring) and old_value not in self.values():
                self.coordinates.discard(old_value)
        #-- End: if

        super(Transform, self).__setitem__(item, value)

        # Add string values to the coordinates set
        if isinstance(value, basestring):
            self.coordinates.add(value)
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)

'''    
        return 'Transform       : %s' % repr(self)
    #--- End: def

    def close(self):
        '''

Close all files referenced by transform values.

:Returns:

    None

**Examples**

>>> t.close()

'''
        for value in self.itervalues():
            if hasattr(value, 'close'):
                value.close()
    #--- End: def

    def copy(self):
        '''

Return a deep copy.

``t.copy()`` is equivalent to ``copy.deepcopy(t)``.

.. note:: This method is *not* like the built-in `dict.copy` method,
          which creates a shallow copy.

:Returns:

    out :
        The deep copy.

**Examples**

>>> u = t.copy()

'''       
        X = type(self)
        new = X.__new__(X)

        coordinates = self.coordinates.copy()

        new.coordinates  = coordinates
        new.name         = self.name
        new.grid_mapping = self.grid_mapping

        _dict = {}
        for term, value in self.iteritems():
            if (value in coordinates or value is None or
                not hasattr(value, 'copy')):
                _dict[term] = value
            else:
                _dict[term] = value.copy()
        #--- End: for
        new._dict = _dict

        return new
    #---End: def

    def dump(self, level=0, complete=False, domain=None):
        '''

Return a string containing a full description of the transform.

:Parameters:

    level : int, optional

    complete : bool, optional

    domain : cf.Domain, optional

:Returns:

    out : str
        A string containing the description.

**Examples**

>>> print t.dump()

'''          
        indent0 = '    ' * level
        indent1 = '    ' * (level+1)

        try:
            string = ['%sTransform: %s' % (indent0, self.identity(''))]
        except AttributeError:
            string = ['%sTransform: ' % indent0]

        if domain:
            coord_keys = domain.items('coord')
            for key, value in sorted(self.items()):
#                value = self[key]
                if value in coord_keys:
                    string.append("%s%s = %s" % (indent1, key, repr(domain.get(value))))
                else:
                    if complete and hasattr(value, 'domain'):
                        string.append("%s%s = \n%s" % 
                                      (indent1, key, 
                                       value.dump(level=level+2, complete=False,
                                                  title='Transform field', q='-')))
                    else:
                        string.append("%s%s = %s" % (indent1, key, repr(value)))
        else:
            for key, value in sorted(self.items()):
#                value = self[key]
                if complete and hasattr(value, 'domain'):
                    string.append("%s%s = \n%s" % 
                                  (indent1, key, 
                                   value.dump(level=level+2, complete=False,
                                              title='Transform field', q='-')))
                else:
                    string.append("%s%s = %s" % (indent1, key, repr(value)))
                
        return '\n'.join(string)
    #--- End: def

    def equals(self, other, rtol=None, atol=None, traceback=False):
        '''

True if two instances are equal, False otherwise.

:Parameters:

    other : 
        The object to compare for equality.

    atol : float, optional
        The absolute tolerance for all numerical comparisons, By
        default the value returned by the `ATOL` function is used.

    rtol : float, optional
        The relative tolerance for all numerical comparisons, By
        default the value returned by the `RTOL` function is used.

    traceback : bool, optional
        If True then print a traceback highlighting where the two
        instances differ.

:Returns: 

    out : bool
        Whether or not the two instances are equal.

**Examples**

'''
        if self is other:
            return True
        
        # Check that each instance is the same type
        if self.__class__ != other.__class__:
            if traceback:
                print("%s: Different types: %s, %s" %
                      (self.__class__.__name__,
                       self.__class__.__name__,
                       other.__class__.__name__))
            return False
        #--- End: if
   
        # Check the name
        attr = 'name'
        if self.identity() != other.identity():
            if traceback:
                print("%s: Different identities: %s, %s" %
                      (self.__class__.__name__, attr, 
                       self.identity(), other.identity()))
            return False
        #--- End: if
                
        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()

        # Check that the term values are equal.
        #
        # If the values for a particular key are both undefined or
        # both pointers to coordinates, then they are considered
        # equal.
            
        if set(self) != set(other):
            if traceback:
                print("%s: Different collections of terms:" % 
                      self.__class__.__name__)
            return False
        #--- End: if

        coordinates0 = self.coordinates
        coordinates1 = other.coordinates

        for term, value0 in self.iteritems():            
            value1 = other[term]  

            if value0 is None and value1 is None:
                continue
                
            if value0 in coordinates0 and value1 in coordinates1:
                continue
                
            if equals(value0, value1, rtol=rtol, atol=atol,
                          traceback=traceback):
                continue

            if traceback:
                print("%s: Unequal '%s' term: %s, %s" % 
                      (self.__class__.__name__, 
                       term, repr(value0), repr(value1)))
                return False

#            if (value0 is None != value1 is None or
#                value0 in coordinates0 != value1 in coordinates1 or
#                (value0 not in coordinates0 and 
#                 not equals(value0, value1, rtol=rtol, atol=atol,
#                            traceback=traceback))):
#                if traceback:
#                    print("%s: Unequal '%s' term: %s, %s" % 
#                          (self.__class__.__name__, 
#                           term, repr(value0), repr(value1)))
#                return False
        #--- End: for

        # Still here?
        return True
    #--- End: def

    def equivalent(self, other, rtol=None, atol=None, traceback=False):
        '''

True if two transforms are logically equal, False otherwise.

:Parameters:

    other : cf.Transform
        The object to compare for equality.

    atol : float, optional
        The absolute tolerance for all numerical comparisons, By
        default the value returned by the `cf.ATOL` function is used.

    rtol : float, optional
        The relative tolerance for all numerical comparisons, By
        default the value returned by the `cf.RTOL` function is used.

    traceback : bool, optional
        If True then print a traceback highlighting where the two
        instances differ.

:Returns:

    out : bool
        Whether or not the two objects are equivalent.

**Examples**

>>>

'''
        if self is other:
            return True
        
        # Check that each instance is the same type
        if self.__class__ != other.__class__:
            if traceback:
                print("%s: Different types: %s, %s" %
                      (self.__class__.__name__,
                       self.__class__.__name__,
                       other.__class__.__name__))
            return False
        #--- End: if
   
        # Check the name
        attr = 'name'
        if getattr(self, attr, None) != getattr(other, attr, None):
            if traceback:
                print("%s: Different '%s' attributes: %s, %s" %
                      (self.__class__.__name__, attr, x, y))
            return False
        #--- End: if
                
        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()

        # Check that the term values are equal.
        #
        # If the values for a particular key are both undefined or
        # pointers to coordinates, then they are considered equal.
        coordinates0 = self.coordinates
        coordinates1 = other.coordinates
        for term in set(self.keys() + other.keys()):
            
            if term not in other:
                value0 = self[term]                
                if value0 not in coordinates0 and _nonzero_value(value0):
                    if traceback:
                        print("%s: Unequivalent '%s' term" %
                              (self.__class__.__name__,  term))
                    return False
                else:
                    continue
            #--- End: if

            # Still here?
            if term not in self:
                value1 = other[term]                
                if value1 not in coordinates1 and _nonzero_value(value1):
                    if traceback:
                        print("%s: Unequivalent '%s' term" %
                              (self.__class__.__name__,  term))
                    return False
                else:
                    continue
            #--- End: if

            # Still here?
            value0 = self[term]
            value1 = other[term]
            if ((value0 is None != value1 is None) or
                (value0 not in coordinates0 and value1 not in coordinates1 and 
                 not equivalent(value0, value1, rtol=rtol, atol=atol,
                                traceback=traceback))):
                if traceback:
                    print("%s: Unequivalent '%s' term: %s, %s" % 
                          (self.__class__.__name__, 
                           term, repr(value0), repr(value1)))
                return False
        #--- End: for

        # Still here?
        return True
    #--- End: def

    def identity(self, default=None):
        '''

Return the identity of the transform.

The identity is the standard_name of a formula_terms-type transform or
the grid_mapping_name of grid_mapping-type transform.

:Parameters:

    default : optional
        If the transform has no identity then return *default*. By
        default, *default* is None.

:Returns:

    out :
        The identity.

**Examples**

>>> t.identity()
'rotated_latitude_longitude'
>>> t.identity()
'atmosphere_hybrid_height_coordinate'

'''
        return getattr(self, 'name', default)
    #--- End: def

    @property
    def inspect(self):
        '''

Inspect the object for debugging.

:Returns: 

    None

**Examples**

>>> print t.inspect
<CF Transform: rotated_latitude_longitude>
------------------------------------------
_dict: {'grid_north_pole_latitude': 38.0, 'grid_north_pole_longitude': 190.0}
coordinates: set(['grid_latitude', 'grid_longitude'])
grid_mapping: True
name: 'rotated_latitude_longitude'

'''
        name = repr(self)
        out = [name, ''.ljust(len(name), '-')]

        for key, value in sorted(self.__dict__.items()):
            out.append('%s: %s' % (key, repr(value)))

        return '\n'.join(out)
    #--- End: def

    def match(self, match=None, exact=False, match_all=True, regex=False):
        '''

Test whether or not the transform satisfies the given conditions.

:Returns:

    out : bool
        True if the transform satisfies the given criteria, False
        otherwise.

**Examples**

'''
        if not match:
            return True

        found_match = False
         
        if not isinstance(match, dict):
            match = {None: match}

        for prop, value in match.iteritems():           
            if prop is not None:
                # If, in the future, other tests are allowed then this
                # will have to change
                continue
            
            x = self.identity()

            if x is None:
                var_matches = False
            elif isinstance(x, basestring) and isinstance(value, basestring):
                if exact:
                    var_matches = (value == x)
                elif not regex:
                    var_matches = x.startswith(value)
                else:
                    var_matches = re_search(value, x)
            else:	
                var_matches = (value == x)
                if var_matches is not True:
                    var_matches = False
            #--- End: if

            if match_all:
                if var_matches:
                    found_match = True
                else:
                    return var_matches
            elif var_matches:
                return var_matches
        #--- End: for

        return found_match
    #--- End: def

    def remove_pointer(self, value):
        '''

Remove a pointer to a coordinate.

All terms referencing the coordinate are set to None.

:Returns:

    None

**Examples**

>>> t.remove_pointer('aux1')

''' 
        coordinates = self.coordinates
        if value not in coordinates:
            raise ValueError(
"Error: Can't remove a transform pointer which doesn't exist: %s" % repr(value))

        for term, v in self.iteritems():
            if equals(v, value):
                self[term] = None
        #---End: for
     
        coordinates.discard(value)
    #---End: def

    def remove_all_pointers(self):
        '''

Remove all all pointers to coordinates.

All terms referencing coordinates are set to None.

:Returns:

    None

**Examples**

>>> t.remove_all_pointers()

'''
        coordinates = self.coordinates

        for term, value in self.iteritems():
            if value in coordinates:
                self[term] = None
                    
        coordinates.clear()
    #---End: def

    def rename_pointer(self, value, new_value):
        '''

Rename a coordinate pointer.

:Parameters:

    value : str
        The coordinate pointer to be renamed.

    new_value : str or None
        The new coordinate pointer. May be None to signify an unset
        coordinate.

:Returns:

    None

**Examples**

>>> t.rename_pointer('aux0', 'aux3')
>>> t.rename_pointer('dim1', 'dim2')
>>> t.rename_pointer('dim0', 'aux0')
>>> t.rename_pointer('aux2', None)

'''
        coordinates = self.coordinates
        if value not in coordinates:
            raise ValueError(
"Error: Can't rename a coordinate pointer which doesn't exist: %s" % 
repr(value))
    
        coordinates.remove(value)

        for term, v in self.items():
            if equals(v, value):
                self[term] = new_value
    #---End: def

#--- End: class


def _nonzero_value(value):
    try:
        return value.data.any()
    except AttributeError:
        return value != 0
